@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _apu_h_
 #define _apu_h_
 
@@ -43,52 +45,52 @@
 
 struct SIAPU
 {
-    uint8 *PC;
-    uint8 *RAM;
-    uint8 *DirectPage;
-    bool8 APUExecuting;
-    uint8 Bit;
+    uint8  *PC;
+    uint8  *RAM;
+    uint8  *DirectPage;
+    bool8  APUExecuting;
+    uint8  Bit;
     uint32 Address;
-    uint8 *WaitAddress1;
-    uint8 *WaitAddress2;
+    uint8  *WaitAddress1;
+    uint8  *WaitAddress2;
     uint32 WaitCounter;
-    uint8 *ShadowRAM;
-    uint8 *CachedSamples;
-    uint8 _Carry;
-    uint8 _Zero;
-    uint8 _Overflow;
+    uint8  *ShadowRAM;
+    uint8  *CachedSamples;
+    uint8  _Carry;
+    uint8  _Zero;
+    uint8  _Overflow;
     uint32 TimerErrorCounter;
     uint32 Scanline;
-    uint32 OneCycle;
-    uint32 TwoCycles;
+    int32  OneCycle;
+    int32  TwoCycles;
 };
 
 struct SAPU
 {
-    int32 Cycles;
-    bool8 ShowROM;
-    uint8 Flags;
-    uint8 KeyedChannels;
-    uint8 OutPorts [4];
-    uint8 DSP [0x80];
-    uint8 ExtraRAM [64];
+    int32  Cycles;
+    bool8  ShowROM;
+    uint8  Flags;
+    uint8  KeyedChannels;
+    uint8  OutPorts [4];
+    uint8  DSP [0x80];
+    uint8  ExtraRAM [64];
     uint16 Timer [3];
     uint16 TimerTarget [3];
-    bool8 TimerEnabled [3];
-    bool8 TimerValueWritten [3];
+    bool8  TimerEnabled [3];
+    bool8  TimerValueWritten [3];
 };
 
 EXTERN_C struct SAPU APU;
 EXTERN_C struct SIAPU IAPU;
 
-inline void S9xAPUUnpackStatus()
+STATIC inline void S9xAPUUnpackStatus()
 {
     IAPU._Zero = ((APURegisters.P & Zero) == 0) | (APURegisters.P & Negative);
     IAPU._Carry = (APURegisters.P & Carry);
     IAPU._Overflow = (APURegisters.P & Overflow) >> 6;
 }
 
-inline void S9xAPUPackStatus()
+STATIC inline void S9xAPUPackStatus()
 {
     APURegisters.P &= ~(Zero | Negative | Carry | Overflow);
     APURegisters.P |= IAPU._Carry | ((IAPU._Zero == 0) << 1) |
@@ -115,10 +117,10 @@ extern void (*S9xApuOpcodes [256]) (void);
 END_EXTERN_C
 
 
-#define APU_VOL_LEFT    0x00
-#define APU_VOL_RIGHT   0x01
-#define APU_P_LOW   0x02
-#define APU_P_HIGH  0x03
+#define APU_VOL_LEFT 0x00
+#define APU_VOL_RIGHT 0x01
+#define APU_P_LOW 0x02
+#define APU_P_HIGH 0x03
 #define APU_SRCN 0x04
 #define APU_ADSR1 0x05
 #define APU_ADSR2 0x06
