@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include <stdlib.h>
 
 #include "snes9x.h"
@@ -87,18 +89,19 @@ void ComputeClipWindows ()
 {
     struct ClipData *pClip = &IPPU.Clip [0];
 
+    // Loop around the main screen then the sub-screen.
     for (int c = 0; c < 2; c++, pClip++)
     {
+        // Loop around the colour window then a clip window for each of the
+        // background layers.
 	for (int w = 5; w >= 0; w--)
 	{
 	    pClip->Count[w] = 0;
 
-	    if (w == 5)
+	    if (w == 5) // The colour window...
 	    {
-		// The colour window...
-		if (c == 0)
+		if (c == 0) // ... on the main screen
 		{
-		    // ... on the main screen.
 		    if ((Memory.FillRAM [0x2130] & 0xc0) == 0xc0)
 		    {
 			// The whole of the main screen is switched off,
@@ -149,74 +152,77 @@ void ComputeClipWindows ()
 				    ((c == 1 && (Memory.FillRAM [0x2130] & 0x30) == 0x10) ||
 				     (c == 0 && (Memory.FillRAM [0x2130] & 0xc0) == 0x40)));
 
-		    if (PPU.ClipWindow1Enable [w])
+		    if (w == 5 ||
+			(Memory.FillRAM [0x212c + c] & Memory.FillRAM [0x212e + c] & (1 << w)))
 		    {
-			if (!PPU.ClipWindow1Inside [w])
+			if (PPU.ClipWindow1Enable [w])
 			{
-			    Win1[Window1Enabled].Left = PPU.Window1Left;
-			    Win1[Window1Enabled++].Right = PPU.Window1Right + 1;
-			}
-			else
-			{
-			    if (PPU.Window1Left <= PPU.Window1Right)
+			    if (!PPU.ClipWindow1Inside [w])
 			    {
-				if (PPU.Window1Left > 0)
+				Win1[Window1Enabled].Left = PPU.Window1Left;
+				Win1[Window1Enabled++].Right = PPU.Window1Right + 1;
+			    }
+			    else
+			    {
+				if (PPU.Window1Left <= PPU.Window1Right)
 				{
-				    Win1[Window1Enabled].Left = 0;
-				    Win1[Window1Enabled++].Right = PPU.Window1Left;
+				    if (PPU.Window1Left > 0)
+				    {
+					Win1[Window1Enabled].Left = 0;
+					Win1[Window1Enabled++].Right = PPU.Window1Left;
+				    }
+				    if (PPU.Window1Right < 255)
+				    {
+					Win1[Window1Enabled].Left = PPU.Window1Right + 1;
+					Win1[Window1Enabled++].Right = 256;
+				    }
+				    if (Window1Enabled == 0)
+				    {
+					Win1[Window1Enabled].Left = 1;
+					Win1[Window1Enabled++].Right = 0;
+				    }
 				}
-				if (PPU.Window1Right < 255)
+				else
 				{
-				    Win1[Window1Enabled].Left = PPU.Window1Right + 1;
+				    // 'outside' a window with no range - 
+				    // appears to be the whole screen.
+				    Win1[Window1Enabled].Left = 0;
 				    Win1[Window1Enabled++].Right = 256;
 				}
-				if (Window1Enabled == 0)
-				{
-				    Win1[Window1Enabled].Left = 1;
-				    Win1[Window1Enabled++].Right = 0;
-				}
+			    }
+			}
+			if (PPU.ClipWindow2Enable [w])
+			{
+			    if (!PPU.ClipWindow2Inside [w])
+			    {
+				Win2[Window2Enabled].Left = PPU.Window2Left;
+				Win2[Window2Enabled++].Right = PPU.Window2Right + 1;
 			    }
 			    else
 			    {
-				// XXX: What about 'outside' a window with no
-				// range - is that the whole screen or is this
-				// special cased as well?
-				Win1[Window1Enabled].Left = 0;
-				Win1[Window1Enabled++].Right = 256;
-			    }
-			}
-		    }
-		    if (PPU.ClipWindow2Enable [w])
-		    {
-			if (!PPU.ClipWindow2Inside [w])
-			{
-			    Win2[Window2Enabled].Left = PPU.Window2Left;
-			    Win2[Window2Enabled++].Right = PPU.Window2Right + 1;
-			}
-			else
-			{
-			    if (PPU.Window2Left <= PPU.Window2Right)
-			    {
-				if (PPU.Window2Left > 0)
+				if (PPU.Window2Left <= PPU.Window2Right)
+				{
+				    if (PPU.Window2Left > 0)
+				    {
+					Win2[Window2Enabled].Left = 0;
+					Win2[Window2Enabled++].Right = PPU.Window2Left;
+				    }
+				    if (PPU.Window2Right < 255)
+				    {
+					Win2[Window2Enabled].Left = PPU.Window2Right + 1;
+					Win2[Window2Enabled++].Right = 256;
+				    }
+				    if (Window2Enabled == 0)
+				    {
+					Win2[Window2Enabled].Left = 1;
+					Win2[Window2Enabled++].Right = 0;
+				    }
+				}
+				else
 				{
 				    Win2[Window2Enabled].Left = 0;
-				    Win2[Window2Enabled++].Right = PPU.Window2Left + 1;
-				}
-				if (PPU.Window2Right < 255)
-				{
-				    Win2[Window2Enabled].Left = PPU.Window2Right + 1;
 				    Win2[Window2Enabled++].Right = 256;
 				}
-				if (Window2Enabled == 0)
-				{
-				    Win2[Window2Enabled].Left = 1;
-				    Win2[Window2Enabled++].Right = 0;
-				}
-			    }
-			    else
-			    {
-				Win2[Window2Enabled].Left = 0;
-				Win2[Window2Enabled++].Right = 256;
 			    }
 			}
 		    }
@@ -430,7 +436,7 @@ void ComputeClipWindows ()
 			    {
 				uint32 p = 0;
 				uint32 points [10];
-				int i;
+				uint32 i;
 
 				invert = !invert;
 				// Build an array of points (window edges)
@@ -606,7 +612,7 @@ void ComputeClipWindows ()
 			    }
 			    else
 			    {
-				for (int j = 0; j < Window1Enabled; j++)
+				for (uint32 j = 0; j < Window1Enabled; j++)
 				{
 				    pClip->Left [j][w] = Win1[j].Left;
 				    pClip->Right [j][w] = Win1[j].Right;
@@ -655,7 +661,7 @@ void ComputeClipWindows ()
 			    }
 			    else
 			    {
-				for (int j = 0; j < Window2Enabled; j++)
+				for (uint32 j = 0; j < Window2Enabled; j++)
 				{
 				    pClip->Left [j][w] = Win2[j].Left;
 				    pClip->Right [j][w] = Win2[j].Right;
@@ -672,11 +678,10 @@ void ComputeClipWindows ()
 			    // Colour window enabled. Set the
 			    // clip windows for all remaining backgrounds to be
 			    // the same as the colour window.
-			    // XXX: should really intersect them instead.
 			    if (pClip->Count [w] == 0)
 			    {
 				pClip->Count [w] = pClip->Count [5];
-				for (int i = 0; i < pClip->Count [w]; i++)
+				for (uint32 i = 0; i < pClip->Count [w]; i++)
 				{
 				    pClip->Left [i][w] = pClip->Left [i][5];
 				    pClip->Right [i][w] = pClip->Right [i][5];
@@ -684,30 +689,32 @@ void ComputeClipWindows ()
 			    }
 			    else
 			    {
-			    pClip->Count [w] = pClip->Count [5];
-			    for (int i = 0; i < pClip->Count [w]; i++)
-			    {
-				if (pClip->Left [i][5] >= pClip->Right [i][5])
+				// Intersect the colour window with the bg's
+				// own clip window.
+				pClip->Count [w] = pClip->Count [5];
+				for (uint32 i = 0; i < pClip->Count [w]; i++)
 				{
-				    // Colour window has no range, so other
-				    // window will have no range as well.
-				    pClip->Left [i][w] = pClip->Left [i][5];
-				    pClip->Right [i][w] = pClip->Right [i][5];
-				}
-				else
-				if (pClip->Left [i][w] <= pClip->Right [i][w])
-				{
-				    // Intersect the bands of the two windows.
-				    if (pClip->Count [5] == pClip->Count [w])
+				    if (pClip->Left [i][5] >= pClip->Right [i][5])
 				    {
-					pClip->Left [i][w] = MAX(pClip->Left [i][w], 
-								 pClip->Left [i][5]);
-					pClip->Right [i][w] = MIN(pClip->Right [i][w], 
-								  pClip->Right [i][5]);
+					// Colour window has no range, so other
+					// window will have no range as well.
+					pClip->Left [i][w] = pClip->Left [i][5];
+					pClip->Right [i][w] = pClip->Right [i][5];
+				    }
+				    else
+				    if (pClip->Left [i][w] <= pClip->Right [i][w])
+				    {
+					// Intersect the bands of the two windows.
+					if (pClip->Count [5] == pClip->Count [w])
+					{
+					    pClip->Left [i][w] = MAX(pClip->Left [i][w], 
+								     pClip->Left [i][5]);
+					    pClip->Right [i][w] = MIN(pClip->Right [i][w], 
+								      pClip->Right [i][5]);
+					}
 				    }
 				}
 			    }
-			    }
 			}
 		    }
 		} // if (w == 5 | ...
