@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "memmap.h"
 #include "ppu.h"
@@ -45,6 +47,9 @@
 #include "apu.h"
 #include "dma.h"
 #include "sa1.h"
+#include "cheats.h"
+#include "srtc.h"
+#include "sdd1.h"
 
 #ifndef ZSNES_FX
 #include "fxemu.h"
@@ -75,10 +80,7 @@ void S9xResetCPU ()
     SetFlags (MemoryFlag | IndexFlag | IRQ | Emulation);
     ClearFlags (Decimal);
 
-    CPU.Flags = 0;
-    CPU.Cycles = 0;
-    CPU.NextEvent = Settings.HBlankStart;
-    CPU.V_Counter = 0;
+    CPU.Flags = CPU.Flags & (DEBUG_MODE_FLAG | TRACE_FLAG);
     CPU.BranchSkip = FALSE;
     CPU.NMIActive = FALSE;
     CPU.IRQActive = FALSE;
@@ -87,12 +89,22 @@ void S9xResetCPU ()
     CPU.WhichEvent = HBLANK_START_EVENT;
     CPU.PC = NULL;
     CPU.PCBase = NULL;
-    CPU.MemSpeed = 8;
-    CPU.MemSpeedx2 = 16;
-    CPU.FastROMSpeed = 8;
     CPU.PCAtOpcodeStart = NULL;
     CPU.WaitAddress = NULL;
     CPU.WaitCounter = 0;
+    CPU.Cycles = 0;
+    CPU.NextEvent = Settings.HBlankStart;
+    CPU.V_Counter = 0;
+    CPU.MemSpeed = SLOW_ONE_CYCLE;
+    CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
+    CPU.FastROMSpeed = SLOW_ONE_CYCLE;
+    CPU.AutoSaveTimer = 0;
+    CPU.SRAMModified = FALSE;
+    // CPU.NMITriggerPoint = 4; // Set when ROM image loaded
+    CPU.BRKTriggered = FALSE;
+    //CPU.TriedInterleavedMode2 = FALSE; // Reset when ROM image loaded
+    CPU.NMICycleCount = 0;
+    CPU.IRQCycleCount = 0;
     S9xSetPCBase (Registers.PC);
 
 #ifndef VAR_CYCLES
@@ -105,29 +117,38 @@ void S9xResetCPU ()
 }
 
 #ifdef ZSNES_FX
-EXTERN_C void S9xResetSuperFX ();
+START_EXTERN_C
+void S9xResetSuperFX ();
 bool8 WinterGold = 0;
+extern uint8 *C4Ram;
+END_EXTERN_C
 #endif
 
 void S9xReset (void)
 {
     if (Settings.SuperFX)
-	S9xResetSuperFX ();
+        S9xResetSuperFX ();
 
 #ifdef ZSNES_FX
     WinterGold = Settings.WinterGold;
 #endif
+    ZeroMemory (Memory.FillRAM, 0x8000);
+    memset (Memory.VRAM, 0x00, 0x10000);
+    memset (Memory.RAM, 0x55, 0x20000);
+
     S9xResetCPU ();
     S9xResetPPU ();
+    S9xResetSRTC ();
+    if (Settings.SDD1)
+        S9xResetSDD1 ();
+
     S9xResetDMA ();
     S9xResetAPU ();
     S9xResetDSP1 ();
     S9xSA1Init ();
-    
-    ZeroMemory (Memory.FillRAM, 0x8000);
-
-    memset (Memory.VRAM, 0x00, 0x10000);
-    memset (Memory.RAM, 0x55, 0x20000);
+    if (Settings.C4)
+        S9xInitC4 ();
+    S9xInitCheatData ();
 
     Settings.Paused = FALSE;
 }
