@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "memmap.h"
 #include "cpuops.h"
@@ -59,23 +61,16 @@ void S9xMainLoop (void)
 	{
 	    if (CPU.Flags & NMI_FLAG)
 	    {
-		CPU.Flags &= ~NMI_FLAG;
-		if (CPU.WaitingForInterrupt)
+		if (--CPU.NMICycleCount == 0)
 		{
-		    CPU.WaitingForInterrupt = FALSE;
-		    CPU.PC++;
+		    CPU.Flags &= ~NMI_FLAG;
+		    if (CPU.WaitingForInterrupt)
+		    {
+			CPU.WaitingForInterrupt = FALSE;
+			CPU.PC++;
+		    }
+		    S9xOpcode_NMI ();
 		}
-		S9xOpcode_NMI ();
-	    }
-	    if (CPU.Flags & DELAYED_NMI_FLAG2)
-	    {
-		CPU.Flags &= ~DELAYED_NMI_FLAG2;
-		CPU.Flags |= NMI_FLAG;
-	    }
-	    if (CPU.Flags & DELAYED_NMI_FLAG)
-	    {
-		CPU.Flags &= ~DELAYED_NMI_FLAG;
-		CPU.Flags |= DELAYED_NMI_FLAG2;
 	    }
 
 #ifdef DEBUGGER
@@ -100,13 +95,23 @@ void S9xMainLoop (void)
 
 	    if (CPU.Flags & IRQ_PENDING_FLAG)
 	    {
-		if (CPU.IRQActive && !Settings.DisableIRQ)
+		if (CPU.IRQCycleCount == 0)
 		{
-		    if (!CheckFlag (IRQ))
-			S9xOpcode_IRQ ();
+		    if (CPU.WaitingForInterrupt)
+		    {
+			CPU.WaitingForInterrupt = FALSE;
+			CPU.PC++;
+		    }
+		    if (CPU.IRQActive && !Settings.DisableIRQ)
+		    {
+			if (!CheckFlag (IRQ))
+			    S9xOpcode_IRQ ();
+		    }
+		    else
+			CPU.Flags &= ~IRQ_PENDING_FLAG;
 		}
 		else
-		    CPU.Flags &= ~IRQ_PENDING_FLAG;
+		    CPU.IRQCycleCount--;
 	    }
 #ifdef DEBUGGER
 	    if (CPU.Flags & DEBUG_MODE_FLAG)
@@ -127,7 +132,7 @@ void S9xMainLoop (void)
 	}
 #ifdef CPU_SHUTDOWN
 	CPU.PCAtOpcodeStart = CPU.PC;
-#endif	
+#endif
 #ifdef VAR_CYCLES
 	CPU.Cycles += CPU.MemSpeed;
 #else
@@ -135,9 +140,9 @@ void S9xMainLoop (void)
 #endif
 	(*ICPU.S9xOpcodes [*CPU.PC++].S9xOpcode) ();
 	
-	DO_HBLANK_CHECK();
 	if (SA1.Executing)
 	    S9xSA1MainLoop ();
+	DO_HBLANK_CHECK();
     }
     Registers.PC = CPU.PC - CPU.PCBase;
     S9xPackStatus ();
@@ -151,19 +156,27 @@ void S9xMainLoop (void)
 	    S9xSyncSpeed ();
 	CPU.Flags &= ~SCAN_KEYS_FLAG;
     }
+    if (CPU.BRKTriggered && Settings.SuperFX && !CPU.TriedInterleavedMode2)
+    {
+	CPU.TriedInterleavedMode2 = TRUE;
+	CPU.BRKTriggered = FALSE;
+	S9xDeinterleaveMode2 ();
+    }
 }
 
 void S9xSetIRQ (uint32 source)
 {
     CPU.IRQActive |= source;
-
+    CPU.Flags |= IRQ_PENDING_FLAG;
+    CPU.IRQCycleCount = 3;
     if (CPU.WaitingForInterrupt)
     {
+	// Force IRQ to trigger immediately after WAI - 
+	// Final Fantasy Mystic Quest crashes without this.
+	CPU.IRQCycleCount = 0;
 	CPU.WaitingForInterrupt = FALSE;
 	CPU.PC++;
     }
-    if (!Settings.DisableIRQ && !CheckIRQ ())
-	S9xOpcode_IRQ ();
 }
 
 void S9xClearIRQ (uint32 source)
@@ -181,7 +194,9 @@ void S9xDoHBlankProcessing ()
     case HBLANK_START_EVENT:
 	if (IPPU.HDMA && CPU.V_Counter <= PPU.ScreenHeight)
 	    IPPU.HDMA = S9xDoHDMA (IPPU.HDMA);
+
 	break;
+
     case HBLANK_END_EVENT:
 	S9xSuperFXExec ();
 
@@ -204,6 +219,7 @@ void S9xDoHBlankProcessing ()
 	    PPU.OAMAddr = PPU.SavedOAMAddr;
 	    PPU.OAMFlip = 0;
 	    CPU.V_Counter = 0;
+	    CPU.NMIActive = FALSE;
 	    ICPU.Frame++;
 	    PPU.HVBeamCounterLatched = 0;
 	    CPU.Flags |= SCAN_KEYS_FLAG;
@@ -213,15 +229,7 @@ void S9xDoHBlankProcessing ()
 	if (PPU.VTimerEnabled && !PPU.HTimerEnabled &&
 	    CPU.V_Counter == PPU.IRQVBeamPos)
 	{
-	    CPU.IRQActive |= PPU_V_BEAM_IRQ_SOURCE;
-	    // Skip a WAI instruction
-	    if (CPU.WaitingForInterrupt)
-	    {
-		CPU.WaitingForInterrupt = FALSE;
-		CPU.PC++;
-	    }
-	    if (!Settings.DisableIRQ && !CheckIRQ ())
-		S9xOpcode_IRQ ();
+	    S9xSetIRQ (PPU_V_BEAM_IRQ_SOURCE);
 	}
 
 	if (CPU.V_Counter == PPU.ScreenHeight + FIRST_VISIBLE_LINE)
@@ -234,20 +242,15 @@ void S9xDoHBlankProcessing ()
 	    missing.dma_this_frame = 0;
 	    IPPU.MaxBrightness = PPU.Brightness;
 	    PPU.ForcedBlanking = (Memory.FillRAM [0x2100] >> 7) & 1;
-	    Memory.FillRAM[0x4210] = 0x80;
 
+	    Memory.FillRAM[0x4210] = 0x80;
 	    if (Memory.FillRAM[0x4200] & 0x80)
 	    {
 		CPU.NMIActive = TRUE;
-		if (CPU.WaitingForInterrupt)
-		{
-		    CPU.WaitingForInterrupt = FALSE;
-		    CPU.PC++;
-		    CPU.Flags |= NMI_FLAG;
-		}
-		else
-		    CPU.Flags |= DELAYED_NMI_FLAG;
+		CPU.Flags |= NMI_FLAG;
+		CPU.NMICycleCount = CPU.NMITriggerPoint;
 	    }
+
 #ifdef OLD_SNAPSHOT_CODE
 	    if (CPU.Flags & SAVE_SNAPSHOT_FLAG)
 	    {
@@ -266,6 +269,7 @@ void S9xDoHBlankProcessing ()
 	if (CPU.V_Counter == FIRST_VISIBLE_LINE)
 	{
 	    Memory.FillRAM[0x4210] = 0;
+	    CPU.Flags &= ~NMI_FLAG;
 	    S9xStartScreenRefresh ();
 	}
 	if (CPU.V_Counter >= FIRST_VISIBLE_LINE &&
@@ -327,19 +331,12 @@ void S9xDoHBlankProcessing ()
 	    }
 	}
 	break;
-    case HTIMER_EVENT:
-	if (!PPU.VTimerEnabled ||
-	    (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos))
+    case HTIMER_BEFORE_EVENT:
+    case HTIMER_AFTER_EVENT:
+	if (PPU.HTimerEnabled &&
+	    (!PPU.VTimerEnabled || CPU.V_Counter == PPU.IRQVBeamPos))
 	{
-	    CPU.IRQActive |= PPU_H_BEAM_IRQ_SOURCE;
-	    // Skip a WAI instruction
-	    if (CPU.WaitingForInterrupt)
-	    {
-		CPU.WaitingForInterrupt = FALSE;
-		CPU.PC++;
-	    }
-	    if (!Settings.DisableIRQ && !CheckIRQ ())
-		S9xOpcode_IRQ ();
+	    S9xSetIRQ (PPU_H_BEAM_IRQ_SOURCE);
 	}
 	break;
     }
