@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,10 +38,10 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _CPUEXEC_H_
 #define _CPUEXEC_H_
 #include "ppu.h"
+#include "memmap.h"
 #include "65c816.h"
 
 #define DO_HBLANK_CHECK() \
@@ -46,7 +49,7 @@
 	S9xDoHBlankProcessing ();
 
 struct SOpcodes {
-#ifdef WINDOWS
+#ifdef __WIN32__
 	void (__cdecl *S9xOpcode)( void);
 #else
 	void (*S9xOpcode)( void);
@@ -66,6 +69,7 @@ struct SICPU
     uint32 ShiftedDB;
     uint32 Frame;
     uint32 Scanline;
+    uint32 FrameAdvanceCount;
 };
 
 START_EXTERN_C
@@ -164,7 +168,8 @@ STATIC inline void S9xReschedule ()
     uint8 which;
     long max;
     
-    if (CPU.NextEvent >= Settings.HBlankStart)
+    if (CPU.WhichEvent == HBLANK_START_EVENT ||
+	CPU.WhichEvent == HTIMER_AFTER_EVENT)
     {
 	which = HBLANK_END_EVENT;
 	max = Settings.H_Max;
@@ -175,12 +180,14 @@ STATIC inline void S9xReschedule ()
 	max = Settings.HBlankStart;
     }
 
-    if ((long) PPU.HTimerPosition < max &&
+    if (PPU.HTimerEnabled &&
+        (long) PPU.HTimerPosition < max &&
 	(long) PPU.HTimerPosition > CPU.NextEvent &&
 	(!PPU.VTimerEnabled ||
 	 (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos)))
     {
-	which = HTIMER_EVENT;
+	which = (long) PPU.HTimerPosition < Settings.HBlankStart ?
+			HTIMER_BEFORE_EVENT : HTIMER_AFTER_EVENT;
 	max = PPU.HTimerPosition;
     }
     CPU.NextEvent = max;
