@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _CPUMACRO_H_
 #define _CPUMACRO_H_
 
@@ -303,6 +305,10 @@ STATIC inline void A_DEC16 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.A.W--;
     SetZN16 (Registers.A.W);
 }
@@ -312,6 +318,10 @@ STATIC inline void A_DEC8 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.AL--;
     SetZN8 (Registers.AL);
 }
@@ -321,6 +331,10 @@ STATIC inline void DEC16 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Work16 = S9xGetWord (OpAddress) - 1;
     S9xSetWord (Work16, OpAddress);
     SetZN16 (Work16);
@@ -331,6 +345,10 @@ STATIC inline void DEC8 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Work8 = S9xGetByte (OpAddress) - 1;
     S9xSetByte (Work8, OpAddress);
     SetZN8 (Work8);
@@ -353,6 +371,10 @@ STATIC inline void A_INC16 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.A.W++;
     SetZN16 (Registers.A.W);
 }
@@ -362,6 +384,10 @@ STATIC inline void A_INC8 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.AL++;
     SetZN8 (Registers.AL);
 }
@@ -371,6 +397,10 @@ STATIC inline void INC16 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Work16 = S9xGetWord (OpAddress) + 1;
     S9xSetWord (Work16, OpAddress);
     SetZN16 (Work16);
@@ -381,6 +411,10 @@ STATIC inline void INC8 ()
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Work8 = S9xGetByte (OpAddress) + 1;
     S9xSetByte (Work8, OpAddress);
     SetZN8 (Work8);
