@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 /**********************************************************************************************/
 /* CPU-S9xOpcodes.CPP                                                                            */
 /* This file contains all the opcodes                                                         */
@@ -2503,44 +2505,29 @@ static void Op0CM0 (void)
 #ifndef SA1_OPCODES
 inline void CPUShutdown()
 {
-    if (!CPU.BranchSkip)
+    if (Settings.Shutdown && CPU.PC == CPU.WaitAddress)
     {
-	if (Settings.Shutdown && CPU.PCBase + OpAddress == CPU.WaitAddress)
-	{
-	    if (CPU.WaitCounter == 0)
-	    {
-		CPU.WaitAddress = NULL;
-		if (Settings.SA1)
-		    S9xSA1ExecuteDuringSleep ();
-		CPU.Cycles = CPU.NextEvent;
-		if (IAPU.APUExecuting)
-		{
-		    ICPU.CPUExecuting = FALSE;
-		    do
-		    {
-			APU_EXECUTE1();
-		    } while (APU.Cycles < CPU.NextEvent);
-		    ICPU.CPUExecuting = TRUE;
-		}
-	    }
-	    else
-	    if (CPU.WaitCounter >= 2)
-		CPU.WaitCounter = 1;
-	    else
-		CPU.WaitCounter--;
-	}
-    }
-}
-#else
-inline void CPUShutdown()
-{
-    if (Settings.Shutdown && CPU.PCBase + OpAddress == CPU.WaitAddress)
-    {
-	if (CPU.WaitCounter == 0)
+	// Don't skip cycles with a pending NMI or IRQ - could cause delayed
+	// interrupt. Interrupts are delayed for a few cycles already, but
+	// the delay could allow the shutdown code to cycle skip again.
+	// Was causing screen flashing on Top Gear 3000.
+
+	if (CPU.WaitCounter == 0 && 
+	    !(CPU.Flags & (IRQ_PENDING_FLAG | NMI_FLAG)))
 	{
 	    CPU.WaitAddress = NULL;
-	    SA1.Executing = FALSE;
-	    SA1.CPUExecuting = FALSE;
+	    if (Settings.SA1)
+		S9xSA1ExecuteDuringSleep ();
+	    CPU.Cycles = CPU.NextEvent;
+	    if (IAPU.APUExecuting)
+	    {
+		ICPU.CPUExecuting = FALSE;
+		do
+		{
+		    APU_EXECUTE1();
+		} while (APU.Cycles < CPU.NextEvent);
+		ICPU.CPUExecuting = TRUE;
+	    }
 	}
 	else
 	if (CPU.WaitCounter >= 2)
@@ -2549,11 +2536,26 @@ inline void CPUShutdown()
 	    CPU.WaitCounter--;
     }
 }
+#else
+inline void CPUShutdown()
+{
+    if (Settings.Shutdown && CPU.PC == CPU.WaitAddress)
+    {
+	if (CPU.WaitCounter >= 1)
+	{
+	    SA1.Executing = FALSE;
+	    SA1.CPUExecuting = FALSE;
+	}
+	else
+	    CPU.WaitCounter++;
+    }
+}
 #endif
 #else
 #define CPUShutdown()
 #endif
 
+/* BCC */
 static void Op90 (void)
 {
     Relative ();
@@ -2568,9 +2570,11 @@ static void Op90 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BCS */
 static void OpB0 (void)
 {
     Relative ();
@@ -2585,16 +2589,17 @@ static void OpB0 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BEQ */
 static void OpF0 (void)
 {
     Relative ();
     BranchCheck2 ();
     if (CheckZero ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
@@ -2603,16 +2608,17 @@ static void OpF0 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BMI */
 static void Op30 (void)
 {
     Relative ();
     BranchCheck1 ();
     if (CheckNegative ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
@@ -2621,17 +2627,19 @@ static void Op30 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BNE */
 static void OpD0 (void)
 {
     Relative ();
     BranchCheck1 ();
     if (!CheckZero ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
+
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
 #else
@@ -2639,16 +2647,17 @@ static void OpD0 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BPL */
 static void Op10 (void)
 {
     Relative ();
     BranchCheck1 ();
     if (!CheckNegative ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
@@ -2657,9 +2666,11 @@ static void Op10 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BRA */
 static void Op80 (void)
 {
     Relative ();
@@ -2671,15 +2682,16 @@ static void Op80 (void)
     CPU.Cycles++;
 #endif
 #endif
+    CPUShutdown ();
 }
 
+/* BVC */
 static void Op50 (void)
 {
     Relative ();
     BranchCheck0 ();
     if (!CheckOverflow ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
@@ -2688,16 +2700,17 @@ static void Op50 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 
+/* BVS */
 static void Op70 (void)
 {
     Relative ();
     BranchCheck0 ();
     if (CheckOverflow ())
     {
-	CPUShutdown ();
 	CPU.PC = CPU.PCBase + OpAddress;
 #ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
@@ -2706,11 +2719,13 @@ static void Op70 (void)
 	CPU.Cycles++;
 #endif
 #endif
+	CPUShutdown ();
     }
 }
 /**********************************************************************************************/
 
 /* ClearFlag Instructions ******************************************************************** */
+/* CLC */
 static void Op18 (void)
 {
     ClearCarry ();
@@ -2719,6 +2734,7 @@ static void Op18 (void)
 #endif
 }
 
+/* CLD */
 static void OpD8 (void)
 {
     ClearDecimal ();
@@ -2727,15 +2743,17 @@ static void OpD8 (void)
 #endif
 }
 
+/* CLI */
 static void Op58 (void)
 {
     ClearIRQ ();
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
-    CHECK_FOR_IRQ();
+/*    CHECK_FOR_IRQ(); */
 }
 
+/* CLV */
 static void OpB8 (void)
 {
     ClearOverflow ();
@@ -2751,6 +2769,10 @@ static void OpCAX1 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.XL--;
     SetZN8 (Registers.XL);
 }
@@ -2760,6 +2782,9 @@ static void OpCAX0 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
 
     Registers.X.W--;
     SetZN16 (Registers.X.W);
@@ -2770,6 +2795,10 @@ static void Op88X1 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.YL--;
     SetZN8 (Registers.YL);
 }
@@ -2779,6 +2808,10 @@ static void Op88X0 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.Y.W--;
     SetZN16 (Registers.Y.W);
 }
@@ -2790,6 +2823,10 @@ static void OpE8X1 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.XL++;
     SetZN8 (Registers.XL);
 }
@@ -2799,6 +2836,10 @@ static void OpE8X0 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
+
     Registers.X.W++;
     SetZN16 (Registers.X.W);
 }
@@ -2808,6 +2849,9 @@ static void OpC8X1 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
 
     Registers.YL++;
     SetZN8 (Registers.YL);
@@ -2818,6 +2862,9 @@ static void OpC8X0 (void)
 #ifdef VAR_CYCLES
     CPU.Cycles += ONE_CYCLE;
 #endif
+#ifdef CPU_SHUTDOWN
+    CPU.WaitAddress = NULL;
+#endif
 
     Registers.Y.W++;
     SetZN16 (Registers.Y.W);
@@ -2978,6 +3025,7 @@ static void OpAB (void)
     ICPU.ShiftedDB = Registers.DB << 16;
 }
 
+/* PHP */
 static void Op2B (void)
 {
 #ifdef VAR_CYCLES
@@ -2987,6 +3035,7 @@ static void Op2B (void)
     SetZN16 (Registers.D.W);
 }
 
+/* PLP */
 static void Op28 (void)
 {
 #ifdef VAR_CYCLES
@@ -3001,7 +3050,7 @@ static void Op28 (void)
 	Registers.YH = 0;
     }
     S9xFixCycles();
-    CHECK_FOR_IRQ();
+/*     CHECK_FOR_IRQ();*/
 }
 
 static void OpFAX1 (void)
@@ -3043,6 +3092,7 @@ static void Op7AX0 (void)
 /**********************************************************************************************/
 
 /* SetFlag Instructions ********************************************************************** */
+/* SEC */
 static void Op38 (void)
 {
     SetCarry ();
@@ -3051,6 +3101,7 @@ static void Op38 (void)
 #endif
 }
 
+/* SED */
 static void OpF8 (void)
 {
     SetDecimal ();
@@ -3060,6 +3111,7 @@ static void OpF8 (void)
     missing.decimal_mode = 1;
 }
 
+/* SEI */
 static void Op78 (void)
 {
     SetIRQ ();
@@ -3070,6 +3122,7 @@ static void Op78 (void)
 /**********************************************************************************************/
 
 /* Transfer Instructions ********************************************************************* */
+/* TAX8 */
 static void OpAAX1 (void)
 {
 #ifdef VAR_CYCLES
@@ -3079,6 +3132,7 @@ static void OpAAX1 (void)
     SetZN8 (Registers.XL);
 }
 
+/* TAX16 */
 static void OpAAX0 (void)
 {
 #ifdef VAR_CYCLES
@@ -3088,6 +3142,7 @@ static void OpAAX0 (void)
     SetZN16 (Registers.X.W);
 }
 
+/* TAY8 */
 static void OpA8X1 (void)
 {
 #ifdef VAR_CYCLES
@@ -3097,6 +3152,7 @@ static void OpA8X1 (void)
     SetZN8 (Registers.YL);
 }
 
+/* TAY16 */
 static void OpA8X0 (void)
 {
 #ifdef VAR_CYCLES
@@ -3280,6 +3336,10 @@ static void Op00 (void)
 	S9xTraceMessage ("*** BRK");
 #endif
 
+#ifndef SA1_OPCODES
+    CPU.BRKTriggered = TRUE;
+#endif
+
     if (!CheckEmulation())
     {
 	PushB (Registers.PB);
@@ -3411,9 +3471,6 @@ void S9xOpcode_NMI (void)
 	PushW (CPU.PC - CPU.PCBase);
 	S9xPackStatus ();
 	PushB (Registers.PL);
-	// For battle toads: BATTLEMANIACS
-	if (Settings.BattleToadsNMIHack)
-	    ClearFlags (MemoryFlag | IndexFlag | Decimal);
 	ClearDecimal ();
 	SetIRQ ();
 
@@ -3543,6 +3600,9 @@ static void Op4C (void)
 {
     Absolute ();
     S9xSetPCBase (ICPU.ShiftedPB + (OpAddress & 0xffff));
+#if defined(CPU_SHUTDOWN) && defined(SA1_OPCODES)
+    CPUShutdown ();
+#endif
 }
 
 static void Op6C (void)
@@ -3727,7 +3787,7 @@ static void OpC2 (void)
 	Registers.YH = 0;
     }
     S9xFixCycles();
-    CHECK_FOR_IRQ();
+/*    CHECK_FOR_IRQ(); */
 }
 
 static void OpE2 (void)
@@ -3797,13 +3857,13 @@ static void Op40 (void)
     CPU.Cycles += TWO_CYCLES;
 #endif
     S9xFixCycles();
-    CHECK_FOR_IRQ();
+/*    CHECK_FOR_IRQ(); */
 }
 
 /**********************************************************************************************/
 
 /* STP/WAI/DB ******************************************************************************** */
-// Wait for interrupt
+// WAI
 static void OpCB (void)
 {
     if (CPU.IRQActive)
@@ -3824,8 +3884,6 @@ static void OpCB (void)
 #ifndef SA1_OPCODES
 	if (Settings.Shutdown)
 	{
-	    if (Settings.SA1)
-		S9xSA1ExecuteDuringSleep ();
 	    CPU.Cycles = CPU.NextEvent;
 	    if (IAPU.APUExecuting)
 	    {
@@ -3848,7 +3906,7 @@ static void OpCB (void)
     }
 }
 
-// Halt Emulation
+// STP
 static void OpDB (void)
 {
     CPU.PC--;
