@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,17 +38,12 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
-/**********************************************************************************************/
-/* About.CPP                                                                                  */
-/* This file contains all the code of the aboutbox for Snes96                                 */
-/**********************************************************************************************/
-
 #include <string.h>
 
 #include "snes9x.h"
 #include "memmap.h"
 #include "cpuops.h"
+#include "cheats.h"
 #include "ppu.h"
 #include "cpuexec.h"
 #include "debug.h"
@@ -54,17 +52,20 @@
 #include "apu.h"
 #include "sa1.h"
 
-/**********************************************************************************************/
-/* Debug Globals                                                                              */
-/**********************************************************************************************/
-
 #ifdef DEBUGGER
 static void WhatsMissing ();
 static void WhatsUsed ();
 EXTERN_C SDMA DMA[8];
-extern int OBJList[129];
+extern struct SCheatData Cheat;
+
+#ifdef SPCTOOL
+#include "spctool/spc700.h"
+extern "C" void TraceSPC (unsigned char *PC, unsigned short YA, unsigned char X,
+			  SPCFlags PS, unsigned char *SP);
+#endif
 
 FILE *trace = NULL;
+FILE *trace2 = NULL;
 struct SBreakPoint S9xBreakpoint[6];
 struct SDebug
 {
@@ -210,7 +211,8 @@ uint8 S9xOPrint (char *Line, uint8 Bank, uint16 Address)
 
 #else
     uint32 Cycles = CPU.Cycles;
-    
+    uint8 *WaitAddress = CPU.WaitAddress;
+
     S9xOpcode = S9xGetByte ((Bank << 16) + Address);
     sprintf (Line, "$%02X:%04X %02X ", Bank, Address, S9xOpcode);
     Operant[0] = S9xGetByte ((Bank << 16) + Address + 1);
@@ -579,7 +581,8 @@ uint8 S9xOPrint (char *Line, uint8 Bank, uint16 Address)
 	Size = 3;
 	break;
     }
-    sprintf (Line, "%-44s A:%04X X:%04X Y:%04X D:%04X DB:%02X S:%04X P:%c%c%c%c%c%c%c%c%c HC:%03d VC:%03ld",
+// XXX:
+    sprintf (Line, "%-44s A:%04X X:%04X Y:%04X D:%04X DB:%02X S:%04X P:%c%c%c%c%c%c%c%c%c HC:%03d VC:%03ld %02x",
 	     Line, Registers.A.W, Registers.X.W, Registers.Y.W,
 	     Registers.D.W, Registers.DB, Registers.S.W,
 	     CheckEmulation () ? 'E' : 'e',
@@ -592,9 +595,11 @@ uint8 S9xOPrint (char *Line, uint8 Bank, uint16 Address)
 	     CheckZero () ? 'Z' : 'z',
 	     CheckCarry () ? 'C' : 'c',
 	     Cycles,
-	     CPU.V_Counter);
+	     CPU.V_Counter,
+	     CPU.IRQActive);
 
     CPU.Cycles = Cycles;
+    CPU.WaitAddress = WaitAddress;
     return Size;
 #endif
 }
@@ -995,7 +1000,7 @@ uint8 S9xSA1OPrint (char *Line, uint8 Bank, uint16 Address)
 	Size = 3;
 	break;
     }
-    sprintf (Line, "%-44s A:%04X X:%04X Y:%04X D:%04X DB:%02X S:%04X P:%c%c%c%c%c%c%c%c%c HC:%03d VC:%03ld",
+    sprintf (Line, "%-44s A:%04X X:%04X Y:%04X D:%04X DB:%02X S:%04X P:%c%c%c%c%c%c%c%c%c HC:%03ld VC:%03ld",
 	     Line, SA1Registers.A.W, SA1Registers.X.W, SA1Registers.Y.W,
 	     SA1Registers.D.W, SA1Registers.DB, SA1Registers.S.W,
 	     SA1CheckEmulation () ? 'E' : 'e',
@@ -1040,7 +1045,7 @@ int GetNumber (char *Line, uint16 * Number)
 /* GetStartAddress()                                                                          */
 /* This function gets a starting address from a debug command                                 */
 /**********************************************************************************************/
-short GetStartAddress (char *Line, uint8 * Bank, uint16 * Address)
+short GetStartAddress (char *Line, uint8 *Bank, uint32 *Address)
 {
     int a, b;
     if (sscanf (Line + 1, " $%x:%x", &b, &a) != 2)
@@ -1056,11 +1061,61 @@ short GetStartAddress (char *Line, uint8 * Bank, uint16 * Address)
 void ProcessDebugCommand (char *Line)
 {
     uint8 Bank = Registers.PB;
-    uint16 Address = CPU.PC - CPU.PCBase;
+    uint32 Address = CPU.PC - CPU.PCBase;
     uint16 Hold;
     uint16 Number;
     char String [512];
     short ErrorCode;
+
+    if (strcasecmp (Line, "cheat") == 0)
+    {
+	S9xStartCheatSearch (&Cheat);
+	printf ("Cheat Search Started\n");
+	return;
+    }
+    if (strcasecmp (Line, "less") == 0)
+    {
+	S9xSearchForChange (&Cheat, S9X_LESS_THAN, S9X_8_BITS, FALSE, TRUE);
+	printf ("Recorded all values that have decreased\n");
+	return;
+    }
+    if (strcasecmp (Line, "print") == 0)
+    {
+	printf ("Cheat search results:\n");
+	S9xOutputCheatSearchResults (&Cheat);
+	return;
+    }
+
+    if (strncasecmp (Line, "constant", 8) == 0)
+    {
+	uint32 Byte;
+	if (sscanf (&Line [8], "%x %x", &Address, &Byte) == 2)
+	    S9xAddCheat (TRUE, TRUE, Address, Byte);
+	return;
+    }
+    
+    if (strncasecmp (Line, "dump", 4) == 0)
+    {
+	int Count;
+	if (sscanf (&Line [4], "%x %d", &Address, &Count) == 2)
+	{
+	    sprintf (String, "%06x%05d.sd2", Address, Count);
+	    FILE *fs = fopen (String, "wb");
+	    if (fs)
+	    {
+		int i;
+		for (i = 0; i < Count; i++)
+		    putc (S9xGetByte (Address + i), fs);
+
+		fclose (fs);
+	    }
+	    else
+		printf ("Can't open %s for writing\n", String);
+	}
+	else
+	    printf ("Usage: dump start_address_in_hex count_in_decimal\n");
+	return;
+    }
     if (Line[0] == 'i')
     {
 	printf ("Vectors:\n");
@@ -1158,9 +1213,9 @@ void ProcessDebugCommand (char *Line)
 	{
 	    int c = 0;
 	    int i;
-	    for (i = 0; OBJList [i] >= 0; i++)
+	    for (i = 0; GFX.OBJList [i] >= 0; i++)
 	    {
-		if (PPU.OBJ[OBJList [i]].Priority == p)
+		if (PPU.OBJ[GFX.OBJList [i]].Priority == p)
 		    c++;
 	    }
 	    printf ("Priority %d: %03d, ", p, c);
@@ -1187,14 +1242,14 @@ void ProcessDebugCommand (char *Line)
 	    if (SA1.Flags & TRACE_FLAG)
 	    {
 		printf ("SA1 CPU instruction tracing enabled.\n");
-		if (trace == NULL)
-		    trace = fopen ("trace.log", "wb");
+		if (trace2 == NULL)
+		    trace2 = fopen ("trace_sa1.log", "wb");
 	    }
 	    else
 	    {
 		printf ("SA1 CPU instruction tracing disabled.\n");
-		fclose (trace);
-		trace = NULL;
+		fclose (trace2);
+		trace2 = NULL;
 	    }
 	}
 	else
@@ -1220,11 +1275,18 @@ void ProcessDebugCommand (char *Line)
 	extern FILE *apu_trace;
 	if (APU.Flags & TRACE_FLAG)
 	{
+#ifdef SPCTOOL
+	    printf ("ENABLED\n");
+	    _SetSPCDbg (TraceSPC);                   //Install debug handler
+#endif
 	    if (apu_trace == NULL)
 		apu_trace = fopen ("aputrace.log", "wb");
 	}
 	else
 	{
+#ifdef SPCTOOL
+	    _SetSPCDbg (NULL);
+#endif
 	    if (apu_trace)
 	    {
 		fclose (apu_trace);
@@ -1289,8 +1351,13 @@ void ProcessDebugCommand (char *Line)
     {
 	printf ("APU in-ports: %02X %02X %02X %02X\n",
 		IAPU.RAM [0xF4], IAPU.RAM [0xF5], IAPU.RAM [0xF6], IAPU.RAM [0xF7]);
+#ifdef SPCTOOL
+	printf ("APU out-ports: %02X %02X %02X %02X\n",
+		_SPCOutP [0], _SPCOutP [1], _SPCOutP [2], _SPCOutP [3]);
+#else
 	printf ("APU out-ports: %02X %02X %02X %02X\n",
 		APU.OutPorts [0], APU.OutPorts [1], APU.OutPorts [2], APU.OutPorts [3]);
+#endif
 	printf ("ROM/RAM switch: %s\n", (IAPU.RAM [0xf1] & 0x80) ? "ROM" : "RAM");
 	for (int i = 0; i < 3; i++)
 	    if (APU.TimerEnabled [i])
@@ -1390,6 +1457,8 @@ void ProcessDebugCommand (char *Line)
 	// Render this frame
 	IPPU.RenderThisFrame = TRUE;
 	IPPU.FrameSkip = 0;
+	if (sscanf (&Line [1], "%d", &ICPU.FrameAdvanceCount) != 1)
+	    ICPU.Frame = 0;
     }
 
     if (Line[0] == 'g')
@@ -1678,8 +1747,9 @@ static void WhatsUsed ()
     }
     printf ("VRAM write address: 0x%04x(%s), Full Graphic: %d, Address inc: %d\n",
 	    PPU.VMA.Address,
-	    (Memory.FillRAM[0x2115] & 0x80) ? "Byte" : "Word",
-	    (Memory.FillRAM[0x2115] >> 2) & 3, PPU.VMA.Increment);
+	    PPU.VMA.High  ? "Byte" : "Word",
+	    PPU.VMA.FullGraphicCount, PPU.VMA.Increment);
+
     for (i = 0; i < 4; i++)
     {
 	printf ("BG%d: VOffset:%d, HOffset:%d, W:%d, H:%d, TS:%d, BA:0x%04x, TA:0x%04X\n",
@@ -2029,27 +2099,48 @@ void S9xSA1Trace ()
 {
     char String [512];
     S9xSA1OPrint (String, SA1Registers.PB, SA1.PC - SA1.PCBase);
-    fprintf (trace, "%s\n", String);
-    fflush (trace);
+    fprintf (trace2, "%s\n", String);
+    fflush (trace2);
 }
 
 void S9xTraceMessage (const char *s)
 {
-    fprintf (trace, "%s\n", s);
+    if (trace)
+	fprintf (trace, "%s\n", s);
+    else
+    if (trace2)
+	fprintf (trace2, "%s\n", s);
 }
 
 extern "C" void TraceSA1 ()
 {
-    CPU.Flags ^= TRACE_FLAG;
-    if (CPU.Flags & TRACE_FLAG)
+    SA1.Flags ^= TRACE_FLAG;
+    if (SA1.Flags & TRACE_FLAG)
     {
 	printf ("SA1 CPU instruction tracing enabled.\n");
-	if (trace == NULL)
-	    trace = fopen ("trace.log", "wb");
+	if (trace2 == NULL)
+	    trace2 = fopen ("trace_sa1.log", "wb");
     }
     else
     {
 	printf ("SA1 CPU instruction tracing disabled.\n");
+	fclose (trace2);
+	trace2 = NULL;
+    }
+}
+
+extern "C" void Trace ()
+{
+    CPU.Flags ^= TRACE_FLAG;
+    if (CPU.Flags & TRACE_FLAG)
+    {
+	if (trace == NULL)
+	    trace = fopen ("trace.log", "wb");
+	printf ("CPU instruction tracing enabled.\n");
+    }
+    else
+    {
+	printf ("CPU instruction tracing disabled.\n");
 	fclose (trace);
 	trace = NULL;
     }
