@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _DISPLAY_H_
 #define _DISPLAY_H_
 
@@ -63,6 +65,7 @@ void S9xToggleSoundChannel (int channel);
 void S9xSetInfoString (const char *string);
 int S9xMinCommandLineArgs ();
 void S9xNextController ();
+bool8 S9xLoadROMImage (const char *string);
 const char *S9xSelectFilename (const char *def, const char *dir,
 			       const char *ext, const char *title);
 
@@ -70,15 +73,14 @@ const char *S9xChooseFilename (bool8 read_only);
 bool8 S9xOpenSnapshotFile (const char *base, bool8 read_only, STREAM *file);
 void S9xCloseSnapshotFile (STREAM file);
 
-#if !defined(__GLIBC__) || (__GLIBC__ < 2)
-const char *basename (const char *filename);
-#endif
+const char *S9xBasename (const char *filename);
 
 int S9xFStrcmp (FILE *, const char *);
 const char *S9xGetHomeDirectory ();
 const char *S9xGetSnapshotDirectory ();
 const char *S9xGetROMDirectory ();
 const char *S9xGetSRAMFilename ();
+const char *S9xGetFilename (const char *extension);
 END_EXTERN_C
 
 #endif
