@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 
 #include "memmap.h"
@@ -51,6 +53,13 @@ extern int HDMA_ModeByteCounts [8];
 extern uint8 *HDMAMemPointers [8];
 extern uint8 *HDMABasePointers [8];
 
+#if defined(__linux__) || defined(__WIN32__)
+static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
+{
+    return (*(uint32 *) p1 - *(uint32 *) p2);
+}
+#endif
+
 /**********************************************************************************************/
 /* S9xDoDMA()                                                                                   */
 /* This function preforms the general dma transfer                                            */
@@ -64,6 +73,7 @@ void S9xDoDMA (uint8 Channel)
 
     CPU.InDMA = TRUE;
     bool8 in_sa1_dma = FALSE;
+    uint8 *in_sdd1_dma = NULL;
     SDMA *d = &DMA[Channel];
 
     int count = d->TransferBytes;
@@ -82,6 +92,70 @@ void S9xDoDMA (uint8 Channel)
 	break;
     }
 
+    if (Settings.SDD1)
+    {
+	if (d->AAddressFixed && Memory.FillRAM [0x4801] > 0)
+	{
+	    // Hacky support for pre-decompressed S-DD1 data
+	    inc = !d->AAddressDecrement ? 1 : -1;
+	    uint32 address = (((d->ABank << 16) | d->AAddress) & 0xfffff) << 4;
+
+	    address |= Memory.FillRAM [0x4804 + ((d->ABank - 0xc0) >> 4)];
+#if defined(__linux__) || defined (__WIN32__)
+	    void *ptr = bsearch (&address, Memory.SDD1Index, 
+				 Memory.SDD1Entries, 12, S9xCompareSDD1IndexEntries);
+	    if (ptr)
+		in_sdd1_dma = *(uint32 *) ((uint8 *) ptr + 4) + Memory.SDD1Data;
+#else
+	    uint8 *ptr = Memory.SDD1Index;
+
+	    for (uint32 e = 0; e < Memory.SDD1Entries; e++, ptr += 12)
+	    {
+		if (address == *(uint32 *) ptr)
+		{
+		    in_sdd1_dma = *(uint32 *) (ptr + 4) + Memory.SDD1Data;
+		    break;
+		}
+	    }
+#endif
+
+	    if (!in_sdd1_dma)
+	    {
+		// No matching decompressed data found. Must be some new 
+		// graphics not encountered before. Log it if it hasn't been
+		// already.
+		uint8 *p = Memory.SDD1LoggedData;
+		bool8 found = FALSE;
+		uint8 SDD1Bank = Memory.FillRAM [0x4804 + ((d->ABank - 0xc0) >> 4)] | 0xf0;
+
+		for (uint32 i = 0; i < Memory.SDD1LoggedDataCount; i++, p += 8)
+		{
+		    if (*p == d->ABank ||
+			*(p + 1) == (d->AAddress >> 8) &&
+			*(p + 2) == (d->AAddress & 0xff) &&
+			*(p + 3) == (count >> 8) &&
+			*(p + 4) == (count & 0xff) &&
+			*(p + 7) == SDD1Bank)
+		    {
+			found = TRUE;
+			break;
+		    }
+		}
+		if (!found && Memory.SDD1LoggedDataCount < MEMMAP_MAX_SDD1_LOGGED_ENTRIES)
+		{
+		    *p = d->ABank;
+		    *(p + 1) = d->AAddress >> 8;
+		    *(p + 2) = d->AAddress & 0xff;
+		    *(p + 3) = count >> 8;
+		    *(p + 4) = count & 0xff;
+		    *(p + 7) = SDD1Bank;
+		    Memory.SDD1LoggedDataCount += 1;
+		}
+	    }
+	}
+	Memory.FillRAM [0x4801] = 0;
+    }
+
     if (d->BAddress == 0x18 && SA1.in_char_dma && (d->ABank & 0xf0) == 0x40)
     {
 	// Perform packed bitmap to PPU character format conversion on the
@@ -90,74 +164,110 @@ void S9xDoDMA (uint8 Channel)
 	int depth = (Memory.FillRAM [0x2231] & 3) == 0 ? 8 :
 		    (Memory.FillRAM [0x2231] & 3) == 1 ? 4 : 2;
 
-	uint8 *base = GetBasePointer ((d->ABank << 16) + d->AAddress) + d->AAddress;
-	uint8 *buffer = &Memory.ROM [CMemory::MAX_ROM_SIZE - 0x10000];
 	int bytes_per_char = 8 * depth;
 	int bytes_per_line = depth * num_chars;
 	int char_line_bytes = bytes_per_char * num_chars;
+	uint32 addr = (d->AAddress / char_line_bytes) * char_line_bytes;
+	uint8 *base = GetBasePointer ((d->ABank << 16) + addr) + addr;
+	uint8 *buffer = &Memory.ROM [CMemory::MAX_ROM_SIZE - 0x10000];
 	uint8 *p = buffer;
-	in_sa1_dma = TRUE;
+	uint32 inc = char_line_bytes - (d->AAddress % char_line_bytes);
+	uint32 char_count = inc / bytes_per_char;
 
+	in_sa1_dma = TRUE;
+	
+//printf ("%08x,", base); fflush (stdout);
+//printf ("depth = %d, count = %d, bytes_per_char = %d, bytes_per_line = %d, num_chars = %d, char_line_bytes = %d\n",
+//depth, count, bytes_per_char, bytes_per_line, num_chars, char_line_bytes);
 	int i;
 
 	switch (depth)
 	{
 	case 2:
-	    for (i = 0; i < count; i += char_line_bytes, 
-		 base += char_line_bytes)
+	    for (i = 0; i < count; i += inc, base += char_line_bytes, 
+		 inc = char_line_bytes, char_count = num_chars)
 	    {
-		uint8 *line = base;
-		for (int j = 0; j < num_chars; j++, line += 2)
+		uint8 *line = base + (num_chars - char_count) * 2;
+		for (uint32 j = 0; j < char_count && p - buffer < count; 
+		     j++, line += 2)
 		{
 		    uint8 *q = line;
 		    for (int l = 0; l < 8; l++, q += bytes_per_line)
 		    {
-			uint32 byte = *(q + 0) + (*(q + 1) << 8);
-			uint8 p1 = 0;
-			uint8 p2 = 0;
-			for (int b = 0; b < 8; b++)
+			for (int b = 0; b < 2; b++)
 			{
-			    p1 |= ((byte >> (b << 1)) & 1) << (7 - b);
-			    p2 |= ((byte >> ((b << 1) + 1)) & 1) << (7 - b);
+			    uint8 r = *(q + b);
+			    *(p + 0) = (*(p + 0) << 1) | ((r >> 0) & 1);
+			    *(p + 1) = (*(p + 1) << 1) | ((r >> 1) & 1);
+			    *(p + 0) = (*(p + 0) << 1) | ((r >> 2) & 1);
+			    *(p + 1) = (*(p + 1) << 1) | ((r >> 3) & 1);
+			    *(p + 0) = (*(p + 0) << 1) | ((r >> 4) & 1);
+			    *(p + 1) = (*(p + 1) << 1) | ((r >> 5) & 1);
+			    *(p + 0) = (*(p + 0) << 1) | ((r >> 6) & 1);
+			    *(p + 1) = (*(p + 1) << 1) | ((r >> 7) & 1);
 			}
-			*p++ = p1;
-			*p++ = p2;
+			p += 2;
 		    }
 		}
 	    }
 	    break;
 	case 4:
-	    for (i = 0; i < count; i += char_line_bytes, 
-		 base += char_line_bytes)
+	    for (i = 0; i < count; i += inc, base += char_line_bytes, 
+		 inc = char_line_bytes, char_count = num_chars)
 	    {
-		uint8 *line = base;
-		for (int j = 0; j < num_chars; j++, line += 4)
+		uint8 *line = base + (num_chars - char_count) * 4;
+		for (uint32 j = 0; j < char_count && p - buffer < count; 
+		     j++, line += 4)
 		{
 		    uint8 *q = line;
 		    for (int l = 0; l < 8; l++, q += bytes_per_line)
 		    {
-			uint32 byte = *(q + 0) + (*(q + 1) << 8) +
-				      (*(q + 2) << 16) + (*(q + 3) << 24);
-			uint8 p1 = 0;
-			uint8 p2 = 0;
-			uint8 p3 = 0;
-			uint8 p4 = 0;
-			for (int b = 0; b < 8; b++)
+			for (int b = 0; b < 4; b++)
 			{
-			    p1 |= ((byte >> (b << 2)) & 1) << (7 - b);
-			    p2 |= ((byte >> ((b << 2) + 1)) & 1) << (7 - b);
-			    p3 |= ((byte >> ((b << 2) + 2)) & 1) << (7 - b);
-			    p4 |= ((byte >> ((b << 2) + 3)) & 1) << (7 - b);
+			    uint8 r = *(q + b);
+			    *(p +  0) = (*(p +  0) << 1) | ((r >> 0) & 1);
+			    *(p +  1) = (*(p +  1) << 1) | ((r >> 1) & 1);
+			    *(p + 16) = (*(p + 16) << 1) | ((r >> 2) & 1);
+			    *(p + 17) = (*(p + 17) << 1) | ((r >> 3) & 1);
+			    *(p +  0) = (*(p +  0) << 1) | ((r >> 4) & 1);
+			    *(p +  1) = (*(p +  1) << 1) | ((r >> 5) & 1);
+			    *(p + 16) = (*(p + 16) << 1) | ((r >> 6) & 1);
+			    *(p + 17) = (*(p + 17) << 1) | ((r >> 7) & 1);
 			}
-			*p++ = p1;
-			*p++ = p2;
-			*p++ = p3;
-			*p++ = p4;
+			p += 2;
 		    }
+		    p += 32 - 16;
 		}
 	    }
 	    break;
 	case 8:
+	    for (i = 0; i < count; i += inc, base += char_line_bytes, 
+		 inc = char_line_bytes, char_count = num_chars)
+	    {
+		uint8 *line = base + (num_chars - char_count) * 8;
+		for (uint32 j = 0; j < char_count && p - buffer < count; 
+		     j++, line += 8)
+		{
+		    uint8 *q = line;
+		    for (int l = 0; l < 8; l++, q += bytes_per_line)
+		    {
+			for (int b = 0; b < 8; b++)
+			{
+			    uint8 r = *(q + b);
+			    *(p +  0) = (*(p +  0) << 1) | ((r >> 0) & 1);
+			    *(p +  1) = (*(p +  1) << 1) | ((r >> 1) & 1);
+			    *(p + 16) = (*(p + 16) << 1) | ((r >> 2) & 1);
+			    *(p + 17) = (*(p + 17) << 1) | ((r >> 3) & 1);
+			    *(p + 32) = (*(p + 32) << 1) | ((r >> 4) & 1);
+			    *(p + 33) = (*(p + 33) << 1) | ((r >> 5) & 1);
+			    *(p + 48) = (*(p + 48) << 1) | ((r >> 6) & 1);
+			    *(p + 49) = (*(p + 49) << 1) | ((r >> 7) & 1);
+			}
+			p += 2;
+		    }
+		    p += 64 - 16;
+		}
+	    }
 	    break;
 	}
     }
@@ -207,6 +317,12 @@ void S9xDoDMA (uint8 Channel)
 	    p = 0;
 	}
 
+	if (in_sdd1_dma)
+	{
+	    base = in_sdd1_dma;
+	    p = 0;
+	}
+
 	if (inc > 0)
 	    d->AAddress += count;
 	else
@@ -227,6 +343,7 @@ void S9xDoDMA (uint8 Channel)
 		} while (--count > 0);
 		break;
 	    case 0x18:
+		IPPU.FirstVRAMRead = TRUE;
 		if (!PPU.VMA.FullGraphicCount)
 		{
 		    do
@@ -249,6 +366,7 @@ void S9xDoDMA (uint8 Channel)
 		}
 		break;
 	    case 0x19:
+		IPPU.FirstVRAMRead = TRUE;
 		if (!PPU.VMA.FullGraphicCount)
 		{
 		    do
@@ -300,63 +418,76 @@ void S9xDoDMA (uint8 Channel)
 	    }
 	}
 	else
-	if (d->TransferMode == 1)
+	if (d->TransferMode == 1 || d->TransferMode == 5)
 	{
 	    if (d->BAddress == 0x18)
 	    {
 		// Write to V-RAM
+		IPPU.FirstVRAMRead = TRUE;
 		if (!PPU.VMA.FullGraphicCount)
 		{
-		    do
+		    while (count > 1)
 		    {
 			Work = *(base + p);
 			REGISTER_2118_linear(Work);
 			p += inc;
-			if (count <= 1)
-			    break;
 
 			Work = *(base + p);
 			REGISTER_2119_linear(Work);
 			p += inc;
 			CHECK_SOUND();
 			count -= 2;
-		    } while (count > 0);
+		    }
+		    if (count == 1)
+		    {
+			Work = *(base + p);
+			REGISTER_2118_linear(Work);
+			p += inc;
+		    }
 		}
 		else
 		{
-		    do
+		    while (count > 1)
 		    {
 			Work = *(base + p);
 			REGISTER_2118_tile(Work);
 			p += inc;
-			if (count <= 1)
-			    break;
 
 			Work = *(base + p);
 			REGISTER_2119_tile(Work);
 			p += inc;
 			CHECK_SOUND();
 			count -= 2;
-		    } while (count > 0);
+		    }
+		    if (count == 1)
+		    {
+			Work = *(base + p);
+			REGISTER_2118_tile(Work);
+			p += inc;
+		    }
 		}
 	    }
 	    else
 	    {
 		// DMA mode 1 general case
-		do
+		while (count > 1)
 		{
 		    Work = *(base + p);
 		    S9xSetPPU (Work, 0x2100 + d->BAddress);
 		    p += inc;
-		    if (count <= 1)
-			break;
 
 		    Work = *(base + p);
 		    S9xSetPPU (Work, 0x2101 + d->BAddress);
 		    p += inc;
 		    CHECK_SOUND();
 		    count -= 2;
-		} while (count > 0);
+		}
+		if (count == 1)
+		{
+		    Work = *(base + p);
+		    S9xSetPPU (Work, 0x2100 + d->BAddress);
+		    p += inc;
+		}
 	    }
 	}
 	else
@@ -422,7 +553,7 @@ void S9xDoDMA (uint8 Channel)
 	else
 	{
 #ifdef DEBUGGER
-	    if (Settings.TraceDMA)
+//	    if (Settings.TraceDMA)
 	    {
 		sprintf (String, "Unknown DMA transfer mode: %d on channel %d\n",
 			 d->TransferMode, Channel);
@@ -430,17 +561,6 @@ void S9xDoDMA (uint8 Channel)
 	    }
 #endif
 	}
-
-#ifdef SPC700_C
-	APU_EXECUTE ();
-#endif
-	while (CPU.Cycles > CPU.NextEvent)
-	    S9xDoHBlankProcessing ();
-
-	// Check for reaching the start of v-blank then skipping on right
-	// past. Rise of the Robots needs this.
-//	if (!(Memory.FillRAM [0x4210] & 0x80))
-//	    CPU.Flags &= ~DELAYED_NMI_FLAG;
     }
     else
     {
@@ -460,6 +580,7 @@ void S9xDoDMA (uint8 Channel)
 		break;
 
 	    case 1:
+	    case 5:
 #ifndef VAR_CYCLES
 		CPU.Cycles += 3;
 #endif
@@ -544,14 +665,26 @@ void S9xDoDMA (uint8 Channel)
 		break;
 	    }
 	    CHECK_SOUND();
-	    DO_HBLANK_CHECK();
 	} while (count);
     }
     
+#ifdef SPC700_C
+    IAPU.APUExecuting = Settings.APUEnabled;
+    APU_EXECUTE ();
+#endif
+    while (CPU.Cycles > CPU.NextEvent)
+	S9xDoHBlankProcessing ();
+
+    // Super Punch-Out requires that the A-BUS address be updated after the
+    // DMA transfer.
     Memory.FillRAM[0x4302 + (Channel << 4)] = (uint8) d->AAddress;
     Memory.FillRAM[0x4303 + (Channel << 4)] = d->AAddress >> 8;
-    Memory.FillRAM[0x4305 + (Channel << 4)] = 0;
-    Memory.FillRAM[0x4306 + (Channel << 4)] = 0;
+
+    // Secret of the Mana requires that the DMA bytes transfer count be set to
+    // zero when DMA has completed.
+    Memory.FillRAM [0x4305 + (Channel << 4)] = 0;
+    Memory.FillRAM [0x4306 + (Channel << 4)] = 0;
+
     DMA[Channel].IndirectAddress = 0;
     d->TransferBytes = 0;
     
@@ -579,6 +712,12 @@ void S9xStartHDMA ()
     }
 }
 
+#ifdef DEBUGGER
+void S9xTraceSoundDSP (const char *s, int i1 = 0, int i2 = 0, int i3 = 0,
+		       int i4 = 0, int i5 = 0, int i6 = 0, int i7 = 0);
+#endif
+
+
 uint8 S9xDoHDMA (uint8 byte)
 {
     struct SDMA *p = &DMA [0];
@@ -603,7 +742,8 @@ uint8 S9xDoHDMA (uint8 byte)
 		    p->LineCount = line & 0x7f;
 		}
 		
-		if (!p->LineCount)
+		// Disable H-DMA'ing into V-RAM (register 2118) for Hook
+		if (!p->LineCount || p->BAddress == 0x18)
 		{
 		    byte &= ~mask;
 		    p->IndirectAddress += HDMAMemPointers [d] - HDMABasePointers [d];
@@ -645,7 +785,12 @@ uint8 S9xDoHDMA (uint8 byte)
 		p->LineCount--;
 		continue;
 	    }
+
 #ifdef DEBUGGER
+	    if (Settings.TraceSoundDSP && p->FirstLine && 
+		p->BAddress >= 0x40 && p->BAddress <= 0x43)
+		S9xTraceSoundDSP ("Spooling data!!!\n");
+
 	    if (Settings.TraceHDMA && p->FirstLine)
 	    {
 		sprintf (String, "H-DMA[%d] (%d) 0x%02X%04X->0x21%02X %s, Count: %3d, Rep: %s, V-LINE: %3ld %02X%04X",
@@ -659,7 +804,6 @@ uint8 S9xDoHDMA (uint8 byte)
 		S9xMessage (S9X_TRACE, S9X_HDMA_TRACE, String);
 	    }
 #endif
-
 	    switch (p->TransferMode)
 	    {
 	    case 0:
