@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,12 +38,13 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _GETSET_H_
 #define _GETSET_H_
 
 #include "ppu.h"
 #include "dsp1.h"
+#include "cpuexec.h"
+#include "sa1.h"
 
 INLINE uint8 S9xGetByte (uint32 Address)
 {
@@ -71,25 +75,25 @@ INLINE uint8 S9xGetByte (uint32 Address)
 #endif	
 	return (S9xGetPPU (Address & 0xffff));
     case CMemory::MAP_CPU:
-#ifdef VAR_CYCLES   
+#ifdef VAR_CYCLES
 	CPU.Cycles += ONE_CYCLE;
 #endif
 	return (S9xGetCPU (Address & 0xffff));
     case CMemory::MAP_DSP:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif	
 	return (S9xGetDSP (Address & 0xffff));
     case CMemory::MAP_SA1RAM:
     case CMemory::MAP_LOROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	return (*(Memory.SRAM + ((Address & Memory.SRAMMask))));
 
     case CMemory::MAP_HIROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	return (*(Memory.SRAM + (((Address & 0x7fff) - 0x6000 +
 				  ((Address & 0xf0000) >> 3)) & Memory.SRAMMask)));
@@ -101,16 +105,22 @@ INLINE uint8 S9xGetByte (uint32 Address)
 
     case CMemory::MAP_BWRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	return (*(Memory.BWRAM + ((Address & 0x7fff) - 0x6000)));
-	
+
+    case CMemory::MAP_C4:
+	return (S9xGetC4 (Address & 0xffff));
+    
     default:
     case CMemory::MAP_NONE:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
-	return (0);
+#ifdef DEBUGGER
+	printf ("R(B) %06x\n", Address);
+#endif
+	return ((Address >> 8) & 0xff);
     }
 }
 
@@ -160,21 +170,21 @@ INLINE uint16 S9xGetWord (uint32 Address)
 		(S9xGetCPU ((Address + 1) & 0xffff) << 8));
     case CMemory::MAP_DSP:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif	
 	return (S9xGetDSP (Address & 0xffff) |
 		(S9xGetDSP ((Address + 1) & 0xffff) << 8));
     case CMemory::MAP_SA1RAM:
     case CMemory::MAP_LOROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	return (*(Memory.SRAM + (Address & Memory.SRAMMask)) |
 		(*(Memory.SRAM + ((Address + 1) & Memory.SRAMMask)) << 8));
 
     case CMemory::MAP_HIROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	return (*(Memory.SRAM +
 		  (((Address & 0x7fff) - 0x6000 +
@@ -185,7 +195,7 @@ INLINE uint16 S9xGetWord (uint32 Address)
 
     case CMemory::MAP_BWRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	return (*(Memory.BWRAM + ((Address & 0x7fff) - 0x6000)) |
 		(*(Memory.BWRAM + (((Address + 1) & 0x7fff) - 0x6000)) << 8));
@@ -195,28 +205,51 @@ INLINE uint16 S9xGetWord (uint32 Address)
 	printf ("R(W) %06x\n", Address);
 #endif
 
+    case CMemory::MAP_C4:
+	return (S9xGetC4 (Address & 0xffff) |
+		(S9xGetC4 ((Address + 1) & 0xffff) << 8));
+    
     default:
     case CMemory::MAP_NONE:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
-	return (0);
+#ifdef DEBUGGER
+	printf ("R(W) %06x\n", Address);
+#endif
+	return (((Address >> 8) | (Address & 0xff00)) & 0xffff);
     }
 }
 
 INLINE void S9xSetByte (uint8 Byte, uint32 Address)
 {
-#ifdef CPU_SHUTDOWN
+#if defined(CPU_SHUTDOWN)
     CPU.WaitAddress = NULL;
 #endif
+#if defined(VAR_CYCLES)
+    int block;
+    uint8 *SetAddress = Memory.WriteMap [block = ((Address >> MEMMAP_SHIFT) & MEMMAP_MASK)];
+#else
     uint8 *SetAddress = Memory.WriteMap [(Address >> MEMMAP_SHIFT) & MEMMAP_MASK];
+#endif
 
     if (SetAddress >= (uint8 *) CMemory::MAP_LAST)
     {
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += Memory.MemorySpeed [block];
 #endif
+#ifdef CPU_SHUTDOWN
+	SetAddress += Address & 0xffff;
+	if (SetAddress == SA1.WaitByteAddress1 ||
+	    SetAddress == SA1.WaitByteAddress2)
+	{
+	    SA1.Executing = SA1.S9xOpcodes != NULL;
+	    SA1.WaitCounter = 0;
+	}
+	*SetAddress = Byte;
+#else
 	*(SetAddress + (Address & 0xffff)) = Byte;
+#endif
 	return;
     }
 
@@ -239,33 +272,40 @@ INLINE void S9xSetByte (uint8 Byte, uint32 Address)
 
     case CMemory::MAP_DSP:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif	
 	S9xSetDSP (Byte, Address & 0xffff);
 	return;
 
     case CMemory::MAP_LOROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	if (Memory.SRAMMask)
+	{
 	    *(Memory.SRAM + (Address & Memory.SRAMMask)) = Byte;
+	    CPU.SRAMModified = TRUE;
+	}
 	return;
 
     case CMemory::MAP_HIROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	if (Memory.SRAMMask)
+	{
 	    *(Memory.SRAM + (((Address & 0x7fff) - 0x6000 +
 			      ((Address & 0xf0000) >> 3)) & Memory.SRAMMask)) = Byte;
+	    CPU.SRAMModified = TRUE;
+	}
 	return;
 
     case CMemory::MAP_BWRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	*(Memory.BWRAM + ((Address & 0x7fff) - 0x6000)) = Byte;
+	CPU.SRAMModified = TRUE;
 	return;
 
     case CMemory::MAP_DEBUG:
@@ -275,44 +315,66 @@ INLINE void S9xSetByte (uint8 Byte, uint32 Address)
 
     case CMemory::MAP_SA1RAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	*(Memory.SRAM + (Address & 0xffff)) = Byte;
 	SA1.Executing = !SA1.Waiting;
 	break;
 
+    case CMemory::MAP_C4:
+	S9xSetC4 (Byte, Address & 0xffff);
+	return;
+	
     default:
     case CMemory::MAP_NONE:
 #ifdef VAR_CYCLES    
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif	
+#ifdef DEBUGGER
+	printf ("W(B) %06x\n", Address);
+#endif
 	return;
     }
 }
 
 INLINE void S9xSetWord (uint16 Word, uint32 Address)
 {
-    if ((Address & 0x1fff) == 0x1fff)
-    {
-	S9xSetByte ((uint8) Word, Address);
-	S9xSetByte ((uint8) (Word >> 8), Address + 1);
-	return;
-    }
-#ifdef CPU_SHUTDOWN
+#if defined(CPU_SHUTDOWN)
     CPU.WaitAddress = NULL;
 #endif
+#if defined (VAR_CYCLES)
+    int block;
+    uint8 *SetAddress = Memory.WriteMap [block = ((Address >> MEMMAP_SHIFT) & MEMMAP_MASK)];
+#else
     uint8 *SetAddress = Memory.WriteMap [(Address >> MEMMAP_SHIFT) & MEMMAP_MASK];
+#endif
 
     if (SetAddress >= (uint8 *) CMemory::MAP_LAST)
     {
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += Memory.MemorySpeed [block] << 1;
 #endif
+#ifdef CPU_SHUTDOWN
+	SetAddress += Address & 0xffff;
+	if (SetAddress == SA1.WaitByteAddress1 ||
+	    SetAddress == SA1.WaitByteAddress2)
+	{
+	    SA1.Executing = SA1.S9xOpcodes != NULL;
+	    SA1.WaitCounter = 0;
+	}
+#ifdef FAST_LSB_WORD_ACCESS
+	*(uint16 *) SetAddress = Word;
+#else
+	*SetAddress = (uint8) Word;
+	*(SetAddress + 1) = Word >> 8;
+#endif
+#else
 #ifdef FAST_LSB_WORD_ACCESS
 	*(uint16 *) (SetAddress + (Address & 0xffff)) = Word;
 #else
 	*(SetAddress + (Address & 0xffff)) = (uint8) Word;
 	*(SetAddress + ((Address + 1) & 0xffff)) = Word >> 8;
+#endif
 #endif
 	return;
     }
@@ -338,7 +400,7 @@ INLINE void S9xSetWord (uint16 Word, uint32 Address)
 
     case CMemory::MAP_DSP:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif	
 	S9xSetDSP ((uint8) Word, (Address & 0xffff));
 	S9xSetDSP (Word >> 8, (Address & 0xffff) + 1);
@@ -346,18 +408,19 @@ INLINE void S9xSetWord (uint16 Word, uint32 Address)
 
     case CMemory::MAP_LOROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	if (Memory.SRAMMask)
 	{
 	    *(Memory.SRAM + (Address & Memory.SRAMMask)) = (uint8) Word;
 	    *(Memory.SRAM + ((Address + 1) & Memory.SRAMMask)) = Word >> 8;
+	    CPU.SRAMModified = TRUE;
 	}
 	return;
 
     case CMemory::MAP_HIROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	if (Memory.SRAMMask)
 	{
@@ -367,15 +430,17 @@ INLINE void S9xSetWord (uint16 Word, uint32 Address)
 	    *(Memory.SRAM + 
 	      ((((Address + 1) & 0x7fff) - 0x6000 +
 		(((Address + 1) & 0xf0000) >> MEMMAP_SHIFT) & Memory.SRAMMask))) = (uint8) (Word >> 8);
+	    CPU.SRAMModified = TRUE;
 	}
 	return;
 
     case CMemory::MAP_BWRAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
 #endif
 	*(Memory.BWRAM + ((Address & 0x7fff) - 0x6000)) = (uint8) Word;
 	*(Memory.BWRAM + (((Address + 1) & 0x7fff) - 0x6000)) = (uint8) (Word >> 8);
+	CPU.SRAMModified = TRUE;
 	return;
 
     case CMemory::MAP_DEBUG:
@@ -385,17 +450,25 @@ INLINE void S9xSetWord (uint16 Word, uint32 Address)
 
     case CMemory::MAP_SA1RAM:
 #ifdef VAR_CYCLES
-	CPU.Cycles += 8;
+	CPU.Cycles += SLOW_ONE_CYCLE;
 #endif
 	*(Memory.SRAM + (Address & 0xffff)) = (uint8) Word;
 	*(Memory.SRAM + ((Address + 1) & 0xffff)) = (uint8) (Word >> 8);
 	SA1.Executing = !SA1.Waiting;
 	break;
 
+    case CMemory::MAP_C4:
+	S9xSetC4 (Word & 0xff, Address & 0xffff);
+	S9xSetC4 ((uint8) (Word >> 8), (Address + 1) & 0xffff);
+	return;
+	
     default:
     case CMemory::MAP_NONE:
 #ifdef VAR_CYCLES    
-	CPU.Cycles += 16;
+	CPU.Cycles += SLOW_ONE_CYCLE * 2;
+#endif
+#ifdef DEBUGGER
+	printf ("W(W) %06x\n", Address);
 #endif
 	return;
     }
@@ -422,6 +495,8 @@ INLINE uint8 *GetBasePointer (uint32 Address)
 	return (Memory.BWRAM - 0x6000);
     case CMemory::MAP_HIROM_SRAM:
 	return (Memory.SRAM - 0x6000);
+    case CMemory::MAP_C4:
+	return (Memory.C4RAM - 0x6000);
     case CMemory::MAP_DEBUG:
 #ifdef DEBUGGER
 	printf ("GBP %06x\n", Address);
@@ -429,6 +504,9 @@ INLINE uint8 *GetBasePointer (uint32 Address)
 
     default:
     case CMemory::MAP_NONE:
+#ifdef DEBUGGER
+	printf ("GBP %06x\n", Address);
+#endif
 	return (0);
     }
 }
@@ -454,6 +532,8 @@ INLINE uint8 *S9xGetMemPointer (uint32 Address)
 	return (Memory.BWRAM - 0x6000 + (Address & 0xffff));
     case CMemory::MAP_HIROM_SRAM:
 	return (Memory.SRAM - 0x6000 + (Address & 0xffff));
+    case CMemory::MAP_C4:
+	return (Memory.C4RAM - 0x6000 + (Address & 0xffff));
 
     case CMemory::MAP_DEBUG:
 #ifdef DEBUGGER
@@ -461,6 +541,9 @@ INLINE uint8 *S9xGetMemPointer (uint32 Address)
 #endif
     default:
     case CMemory::MAP_NONE:
+#ifdef DEBUGGER
+	printf ("GMP %06x\n", Address);
+#endif
 	return (0);
     }
 }
@@ -506,8 +589,8 @@ INLINE void S9xSetPCBase (uint32 Address)
 	
     case CMemory::MAP_DSP:
 #ifdef VAR_CYCLES
-	CPU.MemSpeed = 8;
-	CPU.MemSpeedx2 = 16;
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
 #endif	
 	CPU.PCBase = Memory.FillRAM - 0x6000;
 	CPU.PC = CPU.PCBase + (Address & 0xffff);
@@ -516,8 +599,8 @@ INLINE void S9xSetPCBase (uint32 Address)
     case CMemory::MAP_SA1RAM:
     case CMemory::MAP_LOROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.MemSpeed = 8;
-	CPU.MemSpeedx2 = 16;
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
 #endif
 	CPU.PCBase = Memory.SRAM;
 	CPU.PC = CPU.PCBase + (Address & 0xffff);
@@ -525,21 +608,30 @@ INLINE void S9xSetPCBase (uint32 Address)
 
     case CMemory::MAP_BWRAM:
 #ifdef VAR_CYCLES
-	CPU.MemSpeed = 8;
-	CPU.MemSpeedx2 = 16;
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
 #endif
 	CPU.PCBase = Memory.BWRAM - 0x6000;
 	CPU.PC = CPU.PCBase + (Address & 0xffff);
 	return;
     case CMemory::MAP_HIROM_SRAM:
 #ifdef VAR_CYCLES
-	CPU.MemSpeed = 8;
-	CPU.MemSpeedx2 = 16;
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
 #endif
 	CPU.PCBase = Memory.SRAM - 0x6000;
 	CPU.PC = CPU.PCBase + (Address & 0xffff);
 	return;
 
+    case CMemory::MAP_C4:
+#ifdef VAR_CYCLES
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
+#endif
+	CPU.PCBase = Memory.C4RAM - 0x6000;
+	CPU.PC = CPU.PCBase + (Address & 0xffff);
+	return;
+
     case CMemory::MAP_DEBUG:
 #ifdef DEBUGGER
 	printf ("SBP %06x\n", Address);
@@ -548,11 +640,14 @@ INLINE void S9xSetPCBase (uint32 Address)
     default:
     case CMemory::MAP_NONE:
 #ifdef VAR_CYCLES
-	CPU.MemSpeed = 8;
-	CPU.MemSpeedx2 = 16;
+	CPU.MemSpeed = SLOW_ONE_CYCLE;
+	CPU.MemSpeedx2 = SLOW_ONE_CYCLE * 2;
 #endif
-	CPU.PCBase = Memory.RAM;
-	CPU.PC = Memory.RAM + (Address & 0xffff);
+#ifdef DEBUGGER
+	printf ("SBP %06x\n", Address);
+#endif
+	CPU.PCBase = Memory.SRAM;
+	CPU.PC = Memory.SRAM + (Address & 0xffff);
 	return;
     }
 }
