@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _GFX_H_
 #define _GFX_H_
 
@@ -45,6 +47,8 @@ struct SGFX{
     // Initialize these variables
     uint8  *Screen;
     uint8  *SubScreen;
+    uint8  *ZBuffer;
+    uint8  *SubZBuffer;
     uint32 Pitch;
 
     // Setup in call to S9xGraphicsInit()
@@ -52,21 +56,35 @@ struct SGFX{
     uint16 *X2;
     uint16 *ZERO_OR_X2;
     uint16 *ZERO;
-    uint32 RealPitch;
-    uint32 Pitchx2;
-    uint32 PPL;
+    uint32 RealPitch; // True pitch of Screen buffer.
+    uint32 Pitch2;    // Same as RealPitch except while using speed up hack for Glide.
+    uint32 ZPitch;    // Pitch of ZBuffer
+    uint32 PPL;	      // Number of pixels on each of Screen buffer
     uint32 PPLx2;
+    uint32 PixSize;
     uint8  *S;
     uint8  *DB;
     uint16 *ScreenColors;
-    uint8  *ZBuffer;
-    uint8  *SubZBuffer;
     uint32 DepthDelta;
     uint8  Z1;
     uint8  Z2;
     uint32 FixedColour;
     const char *InfoString;
     uint32 InfoStringTimeout;
+    uint32 StartY;
+    uint32 EndY;
+    struct ClipData *pCurrentClip;
+    uint32 Mode7Mask;
+    uint32 Mode7PriorityMask;
+    int	   OBJList [129];
+    uint32 Sizes [129];
+    int    VPositions [129];
+
+    uint8  r212c;
+    uint8  r212d;
+    uint8  r2130;
+    uint8  r2131;
+    bool8  Pseudo;
     
 #ifdef GFX_MULTI_FORMAT
     uint32 PixelFormat;
@@ -150,19 +168,34 @@ extern uint8 mul_brightness [16][32];
 #define SUB_SCREEN_DEPTH 0
 #define MAIN_SCREEN_DEPTH 32
 
+#if defined(OLD_COLOUR_BLENDING)
 #define COLOR_ADD(C1, C2) \
 GFX.X2 [((((C1) & RGB_REMOVE_LOW_BITS_MASK) + \
 	  ((C2) & RGB_REMOVE_LOW_BITS_MASK)) >> 1) + \
 	((C1) & (C2) & RGB_LOW_BITS_MASK)]
+#else
+#define COLOR_ADD(C1, C2) \
+(GFX.X2 [((((C1) & RGB_REMOVE_LOW_BITS_MASK) + \
+	  ((C2) & RGB_REMOVE_LOW_BITS_MASK)) >> 1) + \
+	 ((C1) & (C2) & RGB_LOW_BITS_MASK)] | \
+ (((C1) ^ (C2)) & RGB_LOW_BITS_MASK))	   
+#endif
 
 #define COLOR_ADD1_2(C1, C2) \
-((((C1) & RGB_REMOVE_LOW_BITS_MASK) + \
+(((((C1) & RGB_REMOVE_LOW_BITS_MASK) + \
           ((C2) & RGB_REMOVE_LOW_BITS_MASK)) >> 1) + \
-         ((C1) & (C2) & RGB_LOW_BITS_MASK)
+         ((C1) & (C2) & RGB_LOW_BITS_MASK) | ALPHA_BITS_MASK)
 
+#if defined(OLD_COLOUR_BLENDING)
 #define COLOR_SUB(C1, C2) \
 GFX.ZERO_OR_X2 [(((C1) | RGB_HI_BITS_MASKx2) - \
 		 ((C2) & RGB_REMOVE_LOW_BITS_MASK)) >> 1]
+#else
+#define COLOR_SUB(C1, C2) \
+(GFX.ZERO_OR_X2 [(((C1) | RGB_HI_BITS_MASKx2) - \
+                  ((C2) & RGB_REMOVE_LOW_BITS_MASK)) >> 1] + \
+((C1) & RGB_LOW_BITS_MASK) - ((C2) & RGB_LOW_BITS_MASK))
+#endif
 
 #define COLOR_SUB1_2(C1, C2) \
 GFX.ZERO [(((C1) | RGB_HI_BITS_MASKx2) - \
