@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "memmap.h"
 #include "ppu.h"
@@ -50,59 +52,43 @@
 #include "soundux.h"
 #include "cheats.h"
 #include "sa1.h"
+#include "netplay.h"
 
 START_EXTERN_C
-char String[513] = {0};
+char String[513];
 
-struct Missing missing = {0};
+struct Missing missing;
 
-struct SICPU ICPU = {
-    0
-};
+struct SICPU ICPU;
 
-struct SCPUState CPU = {
-    0
-};
+struct SCPUState CPU;
 
-struct SRegisters Registers = {
-    0
-};
+struct SRegisters Registers;
 
-struct SAPU APU = {
-    0
-};
+struct SAPU APU;
 
-struct SIAPU IAPU = {
-    0
-};
+struct SIAPU IAPU;
 
-struct SAPURegisters APURegisters = {
-    0
-};
+struct SAPURegisters APURegisters;
 
-struct SSettings Settings = {
-    0
-};
+struct SSettings Settings;
 
-struct SDSP1 DSP1 = {
-    0
-};
+struct SDSP1 DSP1;
 
-struct SSA1Registers SA1Registers = {
-    0
-};
+struct SSA1Registers SA1Registers;
 
-struct SSA1 SA1 = {
-    0
-};
+struct SSA1 SA1;
 
 uint8 *SRAM = NULL;
 uint8 *ROM = NULL;
 uint8 *RegRAM = NULL;
+uint8 *C4RAM = NULL;
 
 long OpAddress = 0;
 
-CMemory Memory = {0};
+CMemory Memory;
+
+struct SSNESGameFixes SNESGameFixes;
 
 uint8 A1 = 0, A2 = 0, A3 = 0, A4 = 0, W1 = 0, W2 = 0, W3 = 0, W4 = 0;
 uint8 Ans8 = 0;
@@ -118,45 +104,38 @@ long Int32 = 0;
 END_EXTERN_C
 
 #ifndef ZSNES_FX
-struct FxInit_s SuperFX = {0};
+struct FxInit_s SuperFX;
 #else
+START_EXTERN_C
 uint8 *SFXPlotTable = NULL;
+END_EXTERN_C
 #endif
 
-struct SPPU PPU = {0};
-struct InternalPPU IPPU = {0};
+struct SPPU PPU;
+struct InternalPPU IPPU;
 
-struct SDMA DMA[8] = {0};
+struct SDMA DMA[8];
 
-uint8 *HDMAMemPointers [8] = { NULL };
-uint8 *HDMABasePointers [8] = { NULL };
+uint8 *HDMAMemPointers [8];
+uint8 *HDMABasePointers [8];
 
-struct SBG BG = {0};
+struct SBG BG;
 
-struct SGFX GFX = {0};
-struct SLineData LineData[240] = { {0} };
-struct SLineMatrixData LineMatrixData [240] = { {0} };
-
-int StartY = 0;
-int EndY = 223;
-struct ClipData *pCurrentClip = NULL;
-uint32 Mode7Mask = 255;
+struct SGFX GFX;
+struct SLineData LineData[240];
+struct SLineMatrixData LineMatrixData [240];
 
+uint8 Mode7Depths [2];
 NormalTileRenderer DrawTilePtr = NULL;
 ClippedTileRenderer DrawClippedTilePtr = NULL;
 NormalTileRenderer DrawHiResTilePtr = NULL;
 ClippedTileRenderer DrawHiResClippedTilePtr = NULL;
 LargePixelRenderer DrawLargePixelPtr = NULL;
-int PixSize = 1;
 
-int OBJList [129] = {0};
-int Sizes [129] = {0};
-int VPositions [129] = {0};
-
-uint32 odd_high[4][16] = { {0} };
-uint32 odd_low[4][16] = { {0} };
-uint32 even_high[4][16] = { {0} };
-uint32 even_low[4][16] = { {0} };
+uint32 odd_high[4][16];
+uint32 odd_low[4][16];
+uint32 even_high[4][16];
+uint32 even_low[4][16];
 
 #ifdef GFX_MULTI_FORMAT
 
@@ -181,23 +160,29 @@ uint32 RGB_HI_BITS_MASKx2 = (RED_HI_BIT_MASK_RGB565 |
 			     GREEN_HI_BIT_MASK_RGB565 |
 			     BLUE_HI_BIT_MASK_RGB565) << 1;
 uint32 RGB_REMOVE_LOW_BITS_MASK = ~RGB_LOW_BITS_MASK;
+uint32 FIRST_COLOR_MASK = FIRST_COLOR_MASK_RGB565;
+uint32 SECOND_COLOR_MASK = SECOND_COLOR_MASK_RGB565;
+uint32 THIRD_COLOR_MASK = THIRD_COLOR_MASK_RGB565;
+uint32 ALPHA_BITS_MASK = ALPHA_BITS_MASK_RGB565;
+uint32 FIRST_THIRD_COLOR_MASK = 0;
+uint32 TWO_LOW_BITS_MASK = 0;
+uint32 HIGH_BITS_SHIFTED_TWO_MASK = 0;
 
 uint32 current_graphic_format = RGB565;
 #endif
 
 uint8 GetBank = 0;
-struct SCheats Cheats [10] = { {0} };
-int num_cheats = 0;
+struct SCheatData Cheat;
 
-volatile SoundStatus so = {0};
-SSoundData SoundData = {0};
-int Echo [24000] = {0};
-int DummyEchoBuffer [SOUND_BUFFER_SIZE] = {0};
-int MixBuffer [SOUND_BUFFER_SIZE] = {0};
-int EchoBuffer [SOUND_BUFFER_SIZE] = {0};
-int FilterTaps [8] = {0};
+volatile SoundStatus so;
+SSoundData SoundData;
+int Echo [24000];
+int DummyEchoBuffer [SOUND_BUFFER_SIZE];
+int MixBuffer [SOUND_BUFFER_SIZE];
+int EchoBuffer [SOUND_BUFFER_SIZE];
+int FilterTaps [8];
 unsigned long Z = 0;
-int Loop [16] = {0};
+int Loop [16];
 
 uint16 SignExtend [2] = {
     0x00, 0xff00
@@ -265,7 +250,7 @@ uint8 Depths[8][4] =
 uint8 BGSizes [2] = {
     8, 16
 };
-uint16 DirectColourMaps [8][256] = { 0 };
+uint16 DirectColourMaps [8][256];
 
 long FilterValues[4][2] =
 {
@@ -307,6 +292,9 @@ uint8 APUROM [64] =
     0xF4,0xC4,0xF4,0xDD,0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF
 };
 
+#ifdef NETPLAY_SUPPORT
+struct SNetPlay NetPlay;
+#endif
 
 // Raw SPC700 instruction cycle lengths
 int32 S9xAPUCycleLengths [256] = 
