@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,11 +38,9 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 .macro GetCarry
 	movb _Carry, %dl
 	shrb %dl
-//	cmpb _Carry, $0
 .endm
 
 .macro GetNotCarry
@@ -98,30 +99,29 @@
 /************* IMMEDIATE8 ****************/
 .Macro Immediate8 K
 	movb (PC), %al
-	incl PC
 #ifdef VAR_CYCLES
 	addl MemSpeed, CYCLES
 #endif
+	incl PC
 .endm
 
 /************* IMMEDIATE16 ****************/
 .macro Immediate16 K
 	movw (PC), %ax
-	addl $2, PC
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
+	addl $2, PC
 .endm
 
 /************* Relative ****************/
 .macro Relative K
+	movsbl (PC), %eax
+	incl PC
 #ifdef VAR_CYCLES
 	addl MemSpeed, CYCLES
 #endif
 	movl PC, %edx
-	movsbl (%edx), %eax
-	incl %edx
-	movl %edx, PC
 	subl PCBase, %edx
 	addl %eax, %edx
 	andl $0xffff, %edx
@@ -129,15 +129,16 @@
 
 /************* RelativeLong ****************/
 .macro RelativeLong K
+	xorl %eax, %eax
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
+#endif
+	movw (PC), %ax
+#ifdef VAR_CYCLES
 	addl $6, CYCLES
 #endif
-	xorl %eax, %eax
+	addl $2, PC
 	movl PC, %edx
-	movw (%edx), %ax
-	addl $2, %edx
-	movl %edx, PC
 	subl PCBase, %edx
 	addl %eax, %edx
 	andl $0xffff, %edx
@@ -145,14 +146,14 @@
 
 /************* AbsoluteIndexedIndirect8 ****************/
 .macro AbsoluteIndexedIndirect8 K
+	xorl %edx, %edx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	xorl %edx, %edx
 	movw XX, %dx
 	addw (PC), %dx
-	addl $2, PC
 	orl  ShiftedPB, %edx
+	addl $2, PC
 	call S9xGetWord
 	movl %eax, %edx
 	andl $0xffff, %edx
@@ -160,10 +161,10 @@
 
 /************* AbsoluteIndirectLong8 ****************/
 .macro AbsoluteIndirectLong8 K
+	movw (PC), %dx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	movw (PC), %dx
 	addl $2, PC
 	andl $0xffff, %edx
 	pushl %edx
@@ -180,10 +181,10 @@
 .endm
 
 .macro AbsoluteIndirect8 K
+	movw (PC), %dx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	movw (PC), %dx
 	addl $2, PC
 	andl $0xffff, %edx
 	call S9xGetWord
@@ -192,23 +193,25 @@
 .endm
 
 .macro Absolute8 K
+	movw (PC), %dx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	movw (PC), %dx
 	addl $2, PC
 	andl $0xffff, %edx
 	orl  ShiftedDB, %edx
 .endm
 
 .macro AbsoluteLong8 K
+	movl (PC), %edx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
+#endif
+	andl $0xffffff, %edx
+#ifdef VAR_CYCLES
 	addl MemSpeed, CYCLES
 #endif
-	movl (PC), %edx
 	add $3, PC
-	andl $0xffffff, %edx
 .endm
 
 .macro Direct8 K
@@ -239,8 +242,8 @@
 #endif
 	xorl %edx, %edx
 	movb (PC), %dl
-	incl PC
 	addw DD, %dx
+	incl PC
 	call S9xGetWord
 	movl ShiftedDB, %edx
 	movw %ax, %dx
@@ -336,48 +339,47 @@
 .endm
 
 .macro AbsoluteIndexedX8 K
+	movl ShiftedDB, %edx
+	xorl %eax, %eax
+	movw (PC), %dx
+	movw XX, %ax
+	addl $2, PC
+	addl %eax, %edx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	movl ShiftedDB, %edx
-	movw (PC), %dx
-	xorl %eax, %eax
-	addl $2, PC
-	movw XX, %ax
-	addl %eax, %edx
 	andl $0xffffff, %edx
 .endm
 
 .macro AbsoluteIndexedY8 K
+	movl ShiftedDB, %edx
+	xorl %eax, %eax
+	movw (PC), %dx
+	movw YY, %ax
+	addl $2, PC
+	addl %eax, %edx
 #ifdef VAR_CYCLES
 	addl MemSpeedx2, CYCLES
 #endif
-	movl ShiftedDB, %edx
-	movw (PC), %dx
-	xorl %eax, %eax
-	addl $2, PC
-	movw YY, %ax
-	addl %eax, %edx
 	andl $0xffffff, %edx
 .endm
 
 .macro AbsoluteLongIndexedX8 K
-#ifdef VAR_CYCLES
-	addl MemSpeedx2, CYCLES
-	addl MemSpeed, CYCLES
-#endif
 	movl (PC), %edx
 	xorl %eax, %eax
 	addl $3, PC
 	movw XX, %ax
+#ifdef VAR_CYCLES
+	addl MemSpeedx2, CYCLES
+#endif
 	addl %eax, %edx
+#ifdef VAR_CYCLES
+	addl MemSpeed, CYCLES
+#endif
 	andl $0xffffff, %edx
 .endm
 
 .macro DirectIndirect8 K
-#ifdef VAR_CYCLES
-	addl MemSpeed, CYCLES
-#endif
 #if 0
 	cmpb $0, DL
 	je .DirectIndirect8\K
@@ -385,6 +387,9 @@
 .DirectIndirect8\K:
 #endif
 	xorl %edx, %edx
+#ifdef VAR_CYCLES
+	addl MemSpeed, CYCLES
+#endif
 	movb (PC), %dl
 	addw DD, %dx
 	incl PC
@@ -394,9 +399,6 @@
 .endm
 
 .macro DirectIndirectLong8 K
-#ifdef VAR_CYCLES
-	addl MemSpeed, CYCLES
-#endif
 #if 0
 	cmpb $0, DL
 	je .DirectIndirectLong8\K
@@ -404,6 +406,9 @@
 .DirectIndirectLong8\K:
 #endif
 	xorl %edx, %edx
+#ifdef VAR_CYCLES
+	addl MemSpeed, CYCLES
+#endif
 	movb (PC), %dl
 	addw DD, %dx
 	incl PC
@@ -420,23 +425,27 @@
 .endm
 
 .macro StackRelative8 K
+	xorl %edx, %edx
 #ifdef VAR_CYCLES
 	addl MemSpeed, CYCLES
+#endif
+	movb (PC), %dl
+#ifdef VAR_CYCLES
 	addl $6, CYCLES
 #endif
-	xorl %edx, %edx
-	movb (PC), %dl
 	addw SS, %dx
 	incl PC
 .endm
 
 .macro StackRelativeIndirectIndexed8 K
+	xorl %edx, %edx
 #ifdef VAR_CYCLES
 	addl MemSpeed, CYCLES
+#endif
+	movb (PC), %dl
+#ifdef VAR_CYCLES
 	addl $12, CYCLES
 #endif	
-	xorl %edx, %edx
-	movb (PC), %dl
 	addw SS, %dx
 	incl PC
 	call S9xGetWord
@@ -591,9 +600,7 @@
 	decl %edx
 	andl $0xffff, %edx
 	call S9xSetWord
-	movl SS, %edx
-	subl $2, %edx
-	movw %dx, SS
+	subw $2, SS
 .endm
 
 .macro PullByte K
@@ -606,10 +613,9 @@
 
 .macro PullWord K
 	movl SS, %edx
-	addw $2, %dx
-	movw %dx, SS
-	decl %edx
+	incl %edx
 	andl $0xffff, %edx
 	call S9xGetWord
+	addw $2, SS
 .endm
 #endif
