@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 .macro Adc8 K
 	testb $Decimal, FLAGS
 	jnz .ADC8Decimal\K
@@ -413,6 +415,9 @@
 	addl $6, CYCLES
 #endif
 	call S9xGetByte
+#ifdef CPU_SHUTDOWN
+	movl $0, WaitAddress
+#endif
 	decb %al
 	SetZN
 	popl %edx
@@ -426,6 +431,9 @@
 	addl $6, CYCLES
 #endif
 	call S9xGetWord
+#ifdef CPU_SHUTDOWN
+	movl $0, WaitAddress
+#endif
 	decw %ax
 	Set16ZN
 	popl %edx
@@ -439,6 +447,9 @@
 	addl $6, CYCLES
 #endif
 	call S9xGetByte
+#ifdef CPU_SHUTDOWN
+	movl $0, WaitAddress
+#endif
 	incb %al
 	SetZN
 	popl %edx
@@ -452,6 +463,9 @@
 	addl $6, CYCLES
 #endif
 	call S9xGetWord
+#ifdef CPU_SHUTDOWN
+	movl $0, WaitAddress
+#endif
 	incw %ax
 	Set16ZN
 	popl %edx
@@ -634,11 +648,12 @@
 .endm
 
 .macro S9xPackStatus K
-	movb _Carry, %al
+//	movb _Carry, %al
 	andb $~(Zero | Negative | Carry | Overflow), FLAGS
-	orb %al, FLAGS
+//	orb %al, FLAGS
 	movb _Negative, %al
 	andb $0x80, %al
+	orb _Carry, %al
 	orb %al, FLAGS
 	movb _Overflow, %al
 	salb $6, %al
@@ -651,25 +666,17 @@
 
 .macro CPUShutdown K
 #ifdef CPU_SHUTDOWN
-	testb $0xff, BranchSkip
-	jnz .NoCPUShutdown\K
 	testb $0xff, Shutdown
 	jz .NoCPUShutdown\K
-	movl PCBase, %eax
-	addl %edx, %eax
-	cmpl %eax, WaitAddress
+	cmpl PC, WaitAddress
 	jne .NoCPUShutdown\K
-	movl WaitCounter, %eax
-	testl %eax, %eax
+	cmpl $0, WaitCounter
 	jnz .CPUShutdownNotYet\K
-	xorl %eax, %eax
-	movl %eax, WaitAddress
-	testb $0xff, SA1Enabled
-	jz .NoSA1\K
-	STORE_REGISTERS
-	call S9xSA1ExecuteDuringSleep
-	LOAD_REGISTERS
-.NoSA1\K:
+	movl Flags, %eax
+	andl $NMI_FLAG | IRQ_PENDING_FLAG, %eax
+	jnz .CPUShutdownNotYet\K
+	
+	movl $0, WaitAddress
 	movl NextEvent, CYCLES
 	testb $0xff, APUExecuting
 	jz .NoCPUShutdown\K
@@ -679,19 +686,19 @@
 .CPUShutdownAPUExecLoop\K:
 #ifdef DEBUGGER
 	testb $2,APUFlags
-	je .CPUShutdownNoAPUS9xTrace\K
+	je .CPUShutdownNoAPUTrace\K
 	STORE_REGISTERS
 	ccall S9xTraceAPU
 	LOAD_REGISTERS
-.CPUShutdownNoAPUS9xTrace\K:
+.CPUShutdownNoAPUTrace\K:
 #endif
 	xorl %eax,%eax
 #ifdef SPC700_C
-	movl APUPC,%edx
+	movl APUPC, %edx
 	SAVE_CYCLES
 	movb (%edx),%al
 #else
-	movb (APUPC),%al
+	movb (APUPC), %al
 #endif
 	movl S9xAPUCycles(,%eax,4), %edx
 	movl S9xApuOpcodes(,%eax,4),%eax
@@ -709,11 +716,10 @@
 	jmp .NoCPUShutdown\K
 
 .CPUShutdownNotYet\K:
-	cmpl $2, %eax
+	cmpl $2, WaitCounter
 	jb .CPUShutdownDecCounter\K
-	movl $1, %eax
-	movl %eax, WaitCounter
-	jmp .NoCPUShutdown\K
+	movl $1, WaitCounter
+	jp .NoCPUShutdown\K
 .CPUShutdownDecCounter\K:
 	decl WaitCounter
 .NoCPUShutdown\K:	
