@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #if defined(__DJGCC) || defined(MICROSOFT_C) || defined(_WINDOWS)
 #define S9xTraceMessage _S9xTraceMessage
 #define S9xGetPPU _S9xGetPPU
@@ -44,6 +46,8 @@
 #define S9xGetDSP _S9xGetDSP
 #define S9xGetCPU _S9xGetCPU
 #define S9xSetCPU _S9xSetCPU
+#define S9xGetC4 _S9xGetC4
+#define S9xSetC4 _S9xSetC4
 #define S9xSyncSpeed _S9xSyncSpeed
 #define S9xTrace _S9xTrace
 #define S9xDoHBlankProcessing _S9xDoHBlankProcessing
@@ -65,6 +69,7 @@
 #define S9xGenerateSound _S9xGenerateSound
 #define S9xUpdateJoypads _S9xUpdateJoypads
 #define S9xMainLoop _S9xMainLoop
+#define S9xDeinterleaveMode2 _S9xDeinterleaveMode2
 #define RenderLine _RenderLine
 #define S9xSetIRQ _S9xSetIRQ
 #define S9xClearIRQ _S9xClearIRQ
@@ -145,6 +150,7 @@
 #define MEMMAP_BLOCKS_PER_BANK (0x10000 / MEMMAP_BLOCK_SIZE)
 #define MEMMAP_SHIFT 12
 #define MEMMAP_MASK (MEMMAP_BLOCK_SIZE - 1)
+#define ONE_CYCLE 6
 
 #define CYCLES %esi
 #define PC %edi
