@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #define Flags CPU + 0
 #define BranchSkip CPU + 4
 #define NMIActive CPU + 5
@@ -54,6 +56,13 @@
 #define MemSpeed CPU + 44
 #define MemSpeedx2 CPU + 48
 #define FastROMSpeed CPU + 52
+#define AutoSaveTimer CPU + 56
+#define SRAMModified CPU + 60
+#define NMITriggerPoint CPU + 64
+#define TriedInterleavedMode2 CPU + 69
+#define BRKTriggered CPU + 68
+#define NMICycleCount CPU + 72
+#define IRQCycleCount CPU + 76
 #define PB Registers + 0
 #define DB Registers + 1
 #define PP Registers + 2
@@ -81,16 +90,17 @@
 #define SRAM Memory + 12
 #define BWRAM Memory + 16
 #define FillRAM Memory + 20
-#define HiROM Memory + 24
-#define LoROM Memory + 25
-#define SRAMMask Memory + 26
-#define SRAMSize Memory + 28
-#define Map Memory + 32
-#define WriteMap Memory + 16416
-#define MemorySpeed Memory + 32800
-#define BlockIsRAM Memory + 36896
-#define BlockIsROM Memory + 40992
-#define ROMFilename Memory + 45144
+#define C4RAM Memory + 24
+#define HiROM Memory + 28
+#define LoROM Memory + 29
+#define SRAMMask Memory + 30
+#define SRAMSize Memory + 32
+#define Map Memory + 36
+#define WriteMap Memory + 16420
+#define MemorySpeed Memory + 32804
+#define BlockIsRAM Memory + 36900
+#define BlockIsROM Memory + 40996
+#define ROMFilename Memory + 53364
 #define APUPCS IAPU + 0
 #define APURAM IAPU + 4
 #define APUExecuting IAPU + 12
@@ -133,13 +143,13 @@
 #define SoundSkipMethod Settings + 2
 #define H_Max Settings + 4
 #define HBlankStart Settings + 8
-#define SPCTo65c816Ratio Settings + 12
+#define CyclesPercentage Settings + 12
 #define DisableIRQ Settings + 16
 #define Paused Settings + 17
-#define BattleToadsNMIHack Settings + 18
-#define PAL Settings + 28
-#define SoundSync Settings + 88
-#define SA1Enabled Settings + 73
+#define PAL Settings + 29
+#define SoundSync Settings + 108
+#define SA1Enabled Settings + 82
+#define SuperFXEnabled Settings + 80
 #define ApuP APURegisters + 0
 #define ApuYA APURegisters + 2
 #define ApuA APURegisters + 2
@@ -263,14 +273,22 @@
 #define SA1WaitingForInterrupt SA1 + 27
 #define SA1PCS SA1 + 32
 #define SA1PCBase SA1 + 36
+#define SA1PCAtOpcodeStart SA1 + 44
+#define SA1WaitAddress SA1 + 48
+#define SA1WaitCounter SA1 + 52
+#define SA1WaitByteAddress1 SA1 + 56
+#define SA1WaitByteAddress2 SA1 + 60
 #define SA1BWRAM SA1 + 40
-#define SA1Map SA1 + 56
-#define SA1WriteMap SA1 + 16440
-#define SA1op1 SA1 + 36920
-#define SA1op2 SA1 + 36922
-#define SA1arithmetic_op SA1 + 36924
-#define SA1sum SA1 + 36928
-#define SA1overflow SA1 + 36936
+#define SA1Map SA1 + 64
+#define SA1WriteMap SA1 + 16448
+#define SA1op1 SA1 + 32832
+#define SA1op2 SA1 + 32834
+#define SA1arithmetic_op SA1 + 32836
+#define SA1sum SA1 + 32840
+#define SA1overflow SA1 + 32848
+#define VirtualBitmapFormat SA1 + 32849
+#define SA1_in_char_dma SA1 + 32850
+#define SA1variable_bit_pos SA1 + 32851
 #define SA1PB SA1Registers + 0
 #define SA1DB SA1Registers + 1
 #define SA1PP SA1Registers + 2
