@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifdef UNZIP_SUPPORT
 /**********************************************************************************************/
 /* Loadzip.CPP                                                                                */
@@ -58,8 +60,8 @@
 #include <stdlib.h>
 
 bool8 LoadZip(const char* zipname,
-	     unsigned long *TotalFileSize,
-	     int *headers)
+	      int32 *TotalFileSize,
+	      int32 *headers)
 {
     *TotalFileSize = 0;
     *headers = 0;
@@ -68,8 +70,6 @@ bool8 LoadZip(const char* zipname,
     if(file == NULL)
 	return (FALSE);
 
-    printf("Reading zipfile: %s\n",zipname);
-
     // find largest file in zip file (under MAX_ROM_SIZE)
     // or a file with extension .1
     char filename[132];
@@ -80,7 +80,6 @@ bool8 LoadZip(const char* zipname,
     {
 	char name[132];
 	unzGetCurrentFileInfo(file, &info, name,128, NULL,0, NULL,0);
-	//printf("zipfile: %s    size: %d\n",name,info.uncompressed_size);
 
 #if 0
 	int calc_size = info.uncompressed_size / 0x2000;
@@ -98,7 +97,7 @@ bool8 LoadZip(const char* zipname,
 	    continue;
 	}
 	
-	if(info.uncompressed_size > filesize)
+	if ((int) info.uncompressed_size > filesize)
 	{
 	    strcpy(filename,name);
 	    filesize = info.uncompressed_size;
@@ -134,7 +133,6 @@ bool8 LoadZip(const char* zipname,
     if( unzOpenCurrentFile(file) != UNZ_OK )
     {
 	unzClose(file);
-	printf("Error in zip file\n");
 	return (FALSE);
     }
 
@@ -146,39 +144,29 @@ bool8 LoadZip(const char* zipname,
 	int calc_size = FileSize / 0x2000;
 	calc_size *= 0x2000;
 	
-	printf("Unzipping %s\n",filename);
-
 	int l = unzReadCurrentFile(file,ptr,FileSize);
 	if(unzCloseCurrentFile(file) == UNZ_CRCERROR)
 	{
 	    unzClose(file);
-	    printf("CRC error in zip file.\n");
 	    return (FALSE);
 	}
 	
 	if(l <= 0 || l != FileSize)
 	{
 	    unzClose(file);
-	    printf("Error in zip file.\n",l);
 	    switch(l)
 	    {
 		case UNZ_ERRNO:
-		    printf("Errno:...\n");
 		    break;
 		case UNZ_EOF:
-		    printf("Unexpected end of file.\n");
 		    break;
 		case UNZ_PARAMERROR:
-		    printf("Parameter error.\n");
 		    break;
 		case UNZ_BADZIPFILE:
-		    printf("Bad zipfile.\n");
 		    break;
 		case UNZ_INTERNALERROR:
-		    printf("Internal error.\n");
 		    break;
 		case UNZ_CRCERROR:
-		    printf("CRC error.\n");
 		    break;
 	    }
 	    return (FALSE);
