@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _memmap_h_
 #define _memmap_h_
 
@@ -43,11 +45,22 @@
 
 #ifdef FAST_LSB_WORD_ACCESS
 #define READ_WORD(s) (*(uint16 *) (s))
+#define READ_DWORD(s) (*(uint32 *) (s))
 #define WRITE_WORD(s, d) (*(uint16 *) (s) = (d)
+#define WRITE_DWORD(s, d) (*(uint32 *) (s) = (d)
 #else
-#define READ_WORD(s) (*(uint8 *) (s) | (*((uint8 *) (s) + 1) << 8))
+#define READ_WORD(s) ( *(uint8 *) (s) |\
+		      (*((uint8 *) (s) + 1) << 8))
+#define READ_DWORD(s) ( *(uint8 *) (s) |\
+		       (*((uint8 *) (s) + 1) << 8) |\
+		       (*((uint8 *) (s) + 2) << 16) |\
+		       (*((uint8 *) (s) + 3) << 24))
 #define WRITE_WORD(s, d) *(uint8 *) (s) = (d), \
                          *((uint8 *) (s) + 1) = (d) >> 8
+#define WRITE_DWORD(s, d) *(uint8 *) (s) = (uint8) (d), \
+                          *((uint8 *) (s) + 1) = (uint8) ((d) >> 8),\
+                          *((uint8 *) (s) + 2) = (uint8) ((d) >> 16),\
+                          *((uint8 *) (s) + 3) = (uint8) ((d) >> 24)
 #endif
 
 #define MEMMAP_BLOCK_SIZE (0x1000)
@@ -55,31 +68,42 @@
 #define MEMMAP_BLOCKS_PER_BANK (0x10000 / MEMMAP_BLOCK_SIZE)
 #define MEMMAP_SHIFT 12
 #define MEMMAP_MASK (MEMMAP_BLOCK_SIZE - 1)
+#define MEMMAP_MAX_SDD1_LOGGED_ENTRIES (0x10000 / 8)
 
 class CMemory {
 public:
     bool8 LoadROM (const char *);
+    void  InitROM (bool8);
     bool8 LoadSRAM (const char *);
     bool8 SaveSRAM (const char *);
     bool8 Init ();
     void  Deinit ();
-
+    void  FreeSDD1Data ();
+    
     void WriteProtectROM ();
     void FixROMSpeed ();
     void MapRAM ();
+    void MapExtraRAM ();
     char *Safe (const char *);
     
     void LoROMMap ();
     void LoROM24MBSMap ();
+    void SRAM512KLoROMMap ();
+    void SRAM1024KLoROMMap ();
+    void SufamiTurboLoROMMap ();
     void HiROMMap ();
     void SuperFXROMMap ();
-    void TalesROMMap ();
+    void TalesROMMap (bool8);
     void AlphaROMMap ();
     void SA1ROMMap ();
+    void BSHiROMMap ();
     bool8 AllASCII (uint8 *b, int size);
     int  ScoreHiROM (bool8 skip_header);
     int  ScoreLoROM (bool8 skip_header);
     void ApplyROMFixes ();
+    void CheckForIPSPatch (const char *rom_filename, bool8 header,
+			   int32 &rom_size);
+    
     const char *TVStandard ();
     const char *Speed ();
     const char *StaticRAMSize ();
@@ -89,11 +113,12 @@ public:
     const char *Size ();
     const char *Headers ();
     const char *ROMID ();
-
+    const char *CompanyID ();
+    
     enum {
 	MAP_PPU, MAP_CPU, MAP_DSP, MAP_LOROM_SRAM, MAP_HIROM_SRAM,
-	MAP_NONE, MAP_DEBUG, MAP_BWRAM, MAP_BWRAM_BITMAP, MAP_BWRAM_BITMAP2,
-	MAP_SA1RAM, MAP_LAST
+	MAP_NONE, MAP_DEBUG, MAP_C4, MAP_BWRAM, MAP_BWRAM_BITMAP,
+	MAP_BWRAM_BITMAP2, MAP_SA1RAM, MAP_LAST
     };
     enum { MAX_ROM_SIZE = 0x600000 };
     
@@ -103,6 +128,7 @@ public:
     uint8 *SRAM;
     uint8 *BWRAM;
     uint8 *FillRAM;
+    uint8 *C4RAM;
     bool8 HiROM;
     bool8 LoROM;
     uint16 SRAMMask;
@@ -114,6 +140,7 @@ public:
     uint8 BlockIsROM [MEMMAP_NUM_BLOCKS];
     char  ROMName [ROM_NAME_LEN];
     char  ROMId [5];
+    char  CompanyId [3];
     uint8 ROMSpeed;
     uint8 ROMType;
     uint8 ROMSize;
@@ -123,6 +150,12 @@ public:
     uint32 CalculatedChecksum;
     uint32 ROMChecksum;
     uint32 ROMComplementChecksum;
+    uint8  *SDD1Index;
+    uint8  *SDD1Data;
+    uint32 SDD1Entries;
+    uint32 SDD1LoggedDataCountPrev;
+    uint32 SDD1LoggedDataCount;
+    uint8  SDD1LoggedData [MEMMAP_MAX_SDD1_LOGGED_ENTRIES];
     char ROMFilename [_MAX_PATH];
 };
 
@@ -131,8 +164,11 @@ extern CMemory Memory;
 extern uint8 *SRAM;
 extern uint8 *ROM;
 extern uint8 *RegRAM;
+void S9xDeinterleaveMode2 ();
 END_EXTERN_C
 
+void S9xAutoSaveSRAM ();
+
 #ifdef NO_INLINE_SET_GET
 uint8 S9xGetByte (uint32 Address);
 uint16 S9xGetWord (uint32 Address);
