@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include <errno.h>
 
 #include <stdlib.h>
@@ -171,12 +173,12 @@ int main (int argc, char **argv)
     Settings.SoundPlaybackRate = 4;
     Settings.Stereo = TRUE;
     Settings.SoundBufferSize = 0;
-    Settings.SPCTo65c816Ratio = 2;
+    Settings.CyclesPercentage = 100;
     Settings.DisableSoundEcho = FALSE;
-    Settings.APUEnabled = TRUE;
-    Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
+    Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
+    Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
     Settings.SkipFrames = AUTO_FRAMERATE;
-    Settings.Shutdown = TRUE;
+    Settings.ShutdownMaster = TRUE;
     Settings.FrameTimePAL = 20000;
     Settings.FrameTimeNTSC = 16667;
     Settings.FrameTime = Settings.FrameTimeNTSC;
@@ -189,10 +191,16 @@ int main (int argc, char **argv)
     Settings.Transparency = FALSE;
     Settings.SixteenBit = FALSE;
     Settings.SupportHiRes = FALSE;
-#ifdef NETPLAY_SUPPORT
-    Settings.Port = NETPLAY_DEFAULT_PORT;
+    Settings.NetPlay = FALSE;
+    Settings.ServerName [0] = 0;
+    Settings.ThreadSound = FALSE;
+    Settings.AutoSaveDelay = 30;
+#ifdef _NETPLAY_SUPPORT
+    Settings.Port = NP_DEFAULT_PORT;
 #endif
-
+    Settings.ApplyCheats = TRUE;
+    Settings.TurboMode = FALSE;
+    Settings.TurboSkipFrames = 15;
     rom_filename = S9xParseArgs (argv, argc);
 
     Settings.Transparency = Settings.ForceTransparency;
@@ -204,6 +212,8 @@ int main (int argc, char **argv)
 
     Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;
 
+    rom_filename = S9xParseArgs (argv, argc);
+
     if (!Memory.Init () || !S9xInitAPU())
 	OutOfMemory ();
 
@@ -395,6 +405,11 @@ Connected to \"%s\" on port %d as player #%d playing \"%s\"\n",
     return (0);
 }
 
+void S9xAutoSaveSRAM ()
+{
+    Memory.SaveSRAM (S9xGetSRAMFilename ());
+}
+
 void S9xExit ()
 {
     S9xDeinitDisplay ();
@@ -481,7 +496,7 @@ const char *S9xGetROMDirectory ()
 	return (roms);
 }
 
-const char *basename (const char *f)
+const char *S9xBasename (const char *f)
 {
     const char *p;
     if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
