@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "65c816.h"
 #include "memmap.h"
@@ -49,32 +51,32 @@
 #endif
 
 #define OFFSET(N,F) \
-fprintf (S9xSTREAM, "#define " #N " CPU + %d\n", &((struct SCPUState *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " CPU + %d\n", (int) &((struct SCPUState *) 0)->F);
 #define OFFSET2(N,F) \
-fprintf (S9xSTREAM, "#define " #N " Registers + %d\n", &((struct SRegisters *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " Registers + %d\n", (int) &((struct SRegisters *) 0)->F);
 #define OFFSET3(F) \
-fprintf (S9xSTREAM, "#define " #F " Memory + %d\n", &((class CMemory *) 0)->F);
+fprintf (S9xSTREAM, "#define " #F " Memory + %d\n", (int) &((class CMemory *) 0)->F);
 #define OFFSET4(N,F) \
-fprintf (S9xSTREAM, "#define " #N " APU + %d\n", &((struct SAPU *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " APU + %d\n", (int) &((struct SAPU *) 0)->F);
 #define OFFSET5(N,F) \
-fprintf (S9xSTREAM, "#define " #N " IAPU + %d\n", &((struct SIAPU *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " IAPU + %d\n", (int) &((struct SIAPU *) 0)->F);
 #define OFFSET6(N,F) \
-fprintf (S9xSTREAM, "#define " #N " ICPU + %d\n", &((struct SICPU *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " ICPU + %d\n", (int) &((struct SICPU *) 0)->F);
 #define OFFSET7(N,F) \
-fprintf (S9xSTREAM, "#define " #N " Settings + %d\n", &((struct SSettings *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " Settings + %d\n", (int) &((struct SSettings *) 0)->F);
 #define OFFSET8(N, F) \
-fprintf (S9xSTREAM, "#define " #N " APURegisters + %d\n", &((struct SAPURegisters *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " APURegisters + %d\n", (int) &((struct SAPURegisters *) 0)->F);
 
 #define OFFSET9(N, F) \
-fprintf (S9xSTREAM, "#define " #N " PPU + %d\n", &((struct SPPU *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " PPU + %d\n", (int) &((struct SPPU *) 0)->F);
 #define OFFSET10(N, F) \
-fprintf (S9xSTREAM, "#define " #N " IPPU + %d\n", &((struct InternalPPU *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " IPPU + %d\n", (int) &((struct InternalPPU *) 0)->F);
 #define OFFSET11(N, F) \
-fprintf (S9xSTREAM, "#define " #N " SA1 + %d\n", &((struct SSA1 *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " SA1 + %d\n", (int) &((struct SSA1 *) 0)->F);
 #define OFFSET12(N, F) \
-fprintf (S9xSTREAM, "#define " #N " SA1Registers + %d\n", &((struct SSA1Registers *) 0)->F);
+fprintf (S9xSTREAM, "#define " #N " SA1Registers + %d\n", (int) &((struct SSA1Registers *) 0)->F);
 
-int main (int argc, char **argv)
+int main (int /*argc*/, char ** /*argv*/)
 {
     OFFSET(Flags,Flags)
     OFFSET(BranchSkip,BranchSkip)
@@ -94,6 +96,13 @@ int main (int argc, char **argv)
     OFFSET(MemSpeed,MemSpeed)
     OFFSET(MemSpeedx2,MemSpeedx2)
     OFFSET(FastROMSpeed,FastROMSpeed)
+    OFFSET(AutoSaveTimer,AutoSaveTimer)
+    OFFSET(SRAMModified,SRAMModified)
+    OFFSET(NMITriggerPoint,NMITriggerPoint)
+    OFFSET(TriedInterleavedMode2,TriedInterleavedMode2)
+    OFFSET(BRKTriggered,BRKTriggered)
+    OFFSET(NMICycleCount,NMICycleCount)
+    OFFSET(IRQCycleCount,IRQCycleCount)
 
     OFFSET2(PB,PB)
     OFFSET2(DB,DB)
@@ -123,6 +132,7 @@ int main (int argc, char **argv)
     OFFSET3(SRAM)
     OFFSET3(BWRAM)
     OFFSET3(FillRAM)
+    OFFSET3(C4RAM)
     OFFSET3(HiROM)
     OFFSET3(LoROM)
     OFFSET3(SRAMMask)
@@ -179,13 +189,13 @@ int main (int argc, char **argv)
     OFFSET7(SoundSkipMethod,SoundSkipMethod)
     OFFSET7(H_Max,H_Max)
     OFFSET7(HBlankStart,HBlankStart)
-    OFFSET7(SPCTo65c816Ratio,SPCTo65c816Ratio)
+    OFFSET7(CyclesPercentage,CyclesPercentage)
     OFFSET7(DisableIRQ,DisableIRQ)
     OFFSET7(Paused,Paused)
-    OFFSET7(BattleToadsNMIHack,BattleToadsNMIHack)
     OFFSET7(PAL,PAL)
     OFFSET7(SoundSync,SoundSync)
     OFFSET7(SA1Enabled,SA1)
+    OFFSET7(SuperFXEnabled,SuperFX)
 
     OFFSET8(ApuP,P)
     OFFSET8(ApuYA,YA.W)
@@ -313,6 +323,11 @@ int main (int argc, char **argv)
     OFFSET11(SA1WaitingForInterrupt,WaitingForInterrupt)
     OFFSET11(SA1PCS,PC)
     OFFSET11(SA1PCBase,PCBase)
+    OFFSET11(SA1PCAtOpcodeStart,PCAtOpcodeStart)
+    OFFSET11(SA1WaitAddress,WaitAddress)
+    OFFSET11(SA1WaitCounter,WaitCounter)
+    OFFSET11(SA1WaitByteAddress1,WaitByteAddress1)
+    OFFSET11(SA1WaitByteAddress2,WaitByteAddress2)
     OFFSET11(SA1BWRAM,BWRAM)
     OFFSET11(SA1Map,Map)
     OFFSET11(SA1WriteMap,WriteMap)
@@ -321,6 +336,9 @@ int main (int argc, char **argv)
     OFFSET11(SA1arithmetic_op,arithmetic_op)
     OFFSET11(SA1sum,sum)
     OFFSET11(SA1overflow,overflow)
+    OFFSET11(VirtualBitmapFormat,VirtualBitmapFormat)
+    OFFSET11(SA1_in_char_dma,in_char_dma)
+    OFFSET11(SA1variable_bit_pos,variable_bit_pos)
 
     OFFSET12(SA1PB,PB)
     OFFSET12(SA1DB,DB)
