@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _PORT_H_
 #define _PORT_H_
 
@@ -54,22 +56,65 @@
 //#define PIXEL_FORMAT RGB565
 #define GFX_MULTI_FORMAT
 
+#if defined(TARGET_OS_MAC) && TARGET_OS_MAC
+
+#include "zlib.h"
+#define ZLIB
+#define EXECUTE_SUPERFX_PER_LINE
+#define SOUND
+#define VAR_CYCLES
+#define CPU_SHUTDOWN
+#define SPC700_SHUTDOWN
+#define PIXEL_FORMAT RGB555
+#define CHECK_SOUND()
+#define M_PI 3.14159265359
+#undef  _MAX_PATH
+
+#undef DEBUGGER // Apple Universal Headers sometimes #define DEBUGGER
+#undef GFX_MULTI_FORMAT
+
+int    strncasecmp(const char *s1, const char *s2, unsigned n);
+int    strcasecmp(const char *s1, const char *s2 );
+
+#endif
+
 #ifndef snes9x_types_defined
 #define snes9x_types_defined
+
+typedef unsigned char bool8;
+
+#ifndef __WIN32__
 typedef unsigned char uint8;
 typedef unsigned short uint16;
-typedef unsigned long uint32;
-typedef unsigned char bool8;
 typedef signed char int8;
 typedef short int16;
 typedef int int32;
-#ifndef _WIN32
+typedef unsigned int uint32;
 typedef long long int64;
+#else // __WIN32__
+
+#ifdef __BORLANDC__
+#include <systypes.h>
 #else
-typedef __int64 int64;
-#endif
+
+typedef unsigned char uint8;
+typedef unsigned short uint16;
+typedef signed char int8;
+typedef short int16;
+
+#ifndef WSAAPI
+// winsock2.h typedefs int32 as well.
+typedef long int32;
 #endif
 
+typedef unsigned int uint32;
+
+#endif // __BORLANDC__
+
+typedef __int64 int64;
+
+#endif // __WIN32__
+#endif // snes9x_types_defined
 #include "pixform.h"
 
 #ifndef TRUE
@@ -96,7 +141,7 @@ typedef __int64 int64;
 #endif
 #endif
 
-#ifndef _WIN32
+#ifndef __WIN32__
 
 #ifndef PATH_MAX
 #define PATH_MAX 1024
@@ -114,7 +159,7 @@ void _makepath (char *path, const char *drive, const char *dir,
 		const char *fname, const char *ext);
 void _splitpath (const char *path, char *drive, char *dir, char *fname,
 		 char *ext);
-#else // _WIN32
+#else // __WIN32__
 #define strcasecmp stricmp
 #define strncasecmp strnicmp
 #endif
@@ -143,7 +188,8 @@ typedef void (*SignalHandler)(int);
 #define SIG_PF SignalHandler
 #endif
 
-#if defined(__i386__) || defined(__i486__) || defined(__i586__) || defined(__WIN32)
+#if defined(__i386__) || defined(__i486__) || defined(__i586__) || \
+    defined(__WIN32__) || defined(__alpha__)
 #define LSB_FIRST
 #define FAST_LSB_WORD_ACCESS
 #else
