@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "memmap.h"
 #include "ppu.h"
@@ -46,6 +48,9 @@
 #include "gfx.h"
 #include "display.h"
 #include "sa1.h"
+#include "netplay.h"
+#include "sdd1.h"
+#include "srtc.h"
 
 #ifndef ZSNES_FX
 #include "fxemu.h"
@@ -56,6 +61,65 @@ EXTERN_C void S9xSuperFXWriteReg (uint8, uint32);
 EXTERN_C uint8 S9xSuperFXReadReg (uint32);
 #endif
 
+void S9xUpdateHTimer ()
+{
+    if (PPU.HTimerEnabled)
+    {
+#ifdef DEBUGGER
+	missing.hirq_pos = PPU.IRQHBeamPos;
+#endif
+	PPU.HTimerPosition = PPU.IRQHBeamPos * Settings.H_Max / SNES_HCOUNTER_MAX;
+	if (PPU.HTimerPosition == Settings.H_Max ||
+	    PPU.HTimerPosition == Settings.HBlankStart)
+	{
+	    PPU.HTimerPosition--;
+	}
+
+	if (!PPU.VTimerEnabled || CPU.V_Counter == PPU.IRQVBeamPos)
+	{
+	    if (PPU.HTimerPosition < CPU.Cycles)
+	    {
+		// Missed the IRQ on this line already
+		if (CPU.WhichEvent == HBLANK_END_EVENT ||
+		    CPU.WhichEvent == HTIMER_AFTER_EVENT)
+		{
+		    CPU.WhichEvent = HBLANK_END_EVENT;
+		    CPU.NextEvent = Settings.H_Max;
+		}
+		else
+		{
+		    CPU.WhichEvent = HBLANK_START_EVENT;
+		    CPU.NextEvent = Settings.HBlankStart;
+		}
+	    }
+	    else
+	    {
+		if (CPU.WhichEvent == HTIMER_BEFORE_EVENT ||
+		    CPU.WhichEvent == HBLANK_START_EVENT)
+		{
+		    if (PPU.HTimerPosition > Settings.HBlankStart)
+		    {
+			// HTimer was to trigger before h-blank start,
+			// now triggers after start of h-blank
+			CPU.NextEvent = Settings.HBlankStart;
+			CPU.WhichEvent = HBLANK_START_EVENT;
+		    }
+		    else
+		    {
+			CPU.NextEvent = PPU.HTimerPosition;
+			CPU.WhichEvent = HTIMER_BEFORE_EVENT;
+		    }
+		}
+		else
+		{
+		    CPU.WhichEvent = HTIMER_AFTER_EVENT;
+		    CPU.NextEvent = PPU.HTimerPosition;
+		}
+	    }
+	}
+    }
+}
+
 void S9xFixColourBrightness ()
 {
     IPPU.XB = mul_brightness [PPU.Brightness];
@@ -118,10 +182,8 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 
     case 0x2102:
 	// Sprite write address (low)
-	//PPU.OAMAddr &= 0xFF00;
-	//PPU.OAMAddr |= Byte;
 	PPU.OAMAddr = Byte;
-	PPU.OAMFlip = 0;
+	PPU.OAMFlip = 2;
 	PPU.OAMReadFlip = 0;
 	PPU.SavedOAMAddr = PPU.OAMAddr;
 	if (PPU.OAMPriorityRotation)
@@ -136,8 +198,6 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
     case 0x2103:
 	// Sprite register write address (high), sprite priority rotation
 	// bit.
-	PPU.OAMAddr &= 0x00FF;
-	PPU.OAMAddr |= (Byte & 1) << 8;
 	if ((PPU.OAMPriorityRotation = (Byte & 0x80) == 0 ? 0 : 1))
 	{
 	    PPU.FirstSprite = PPU.OAMAddr & 0x7f;
@@ -145,9 +205,13 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	    missing.sprite_priority_rotation = 1;
 #endif
 	}
-	// No more wrap around...
-	//if (PPU.OAMAddr >= 0x110)
-	//    PPU.OAMAddr %= 0x110;
+	// Only update the sprite write address top bit if the low byte has
+	// been written to first.
+	if (PPU.OAMFlip & 2)
+	{
+	    PPU.OAMAddr &= 0x00FF;
+	    PPU.OAMAddr |= (Byte & 1) << 8;
+	}
 	PPU.OAMFlip = 0;
 	PPU.OAMReadFlip = 0;
 	PPU.SavedOAMAddr = PPU.OAMAddr;
@@ -197,7 +261,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	{
 	    FLUSH_REDRAW ();
 	    PPU.BG[0].SCSize = Byte & 3;
-	    PPU.BG[0].SCBase = ((Byte >> 2) & 0x3f) << 10;
+	    PPU.BG[0].SCBase = (Byte & 0x7c) << 8;
 	}
 	break;
 
@@ -206,7 +270,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	{
 	    FLUSH_REDRAW ();
 	    PPU.BG[1].SCSize = Byte & 3;
-	    PPU.BG[1].SCBase = ((Byte >> 2) & 0x3f) << 10;
+	    PPU.BG[1].SCBase = (Byte & 0x7c) << 8;
 	}
 	break;
 
@@ -215,7 +279,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	{
 	    FLUSH_REDRAW ();
 	    PPU.BG[2].SCSize = Byte & 3;
-	    PPU.BG[2].SCBase = ((Byte >> 2) & 0x3f) << 10;
+	    PPU.BG[2].SCBase = (Byte & 0x7c) << 8;
 	}
 	break;
 
@@ -224,7 +288,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	{
 	    FLUSH_REDRAW ();
 	    PPU.BG[3].SCSize = Byte & 3;
-	    PPU.BG[3].SCBase = ((Byte >> 2) & 0x3f) << 10;
+	    PPU.BG[3].SCBase = (Byte & 0x7c) << 8;
 	}
 	break;
 
@@ -232,8 +296,8 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	if (Byte != Memory.FillRAM [0x210b])
 	{
 	    FLUSH_REDRAW ();
-	    PPU.BG[0].NameBase    = (Byte & 0xF) << 12;
-	    PPU.BG[1].NameBase    = ((Byte >> 4) &0xF) << 12;
+	    PPU.BG[0].NameBase    = (Byte & 7) << 12;
+	    PPU.BG[1].NameBase    = ((Byte >> 4) & 7) << 12;
 	}
 	break;
 
@@ -241,8 +305,8 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	if (Byte != Memory.FillRAM [0x210c])
 	{
 	    FLUSH_REDRAW ();
-	    PPU.BG[2].NameBase    = (Byte & 0xF) << 12;
-	    PPU.BG[3].NameBase    = ((Byte >> 4) & 0xF) << 12;
+	    PPU.BG[2].NameBase    = (Byte & 7) << 12;
+	    PPU.BG[3].NameBase    = ((Byte >> 4) & 7) << 12;
 	}
 	break;
 
@@ -255,7 +319,6 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	PPU.BG[0].VOffset = ((PPU.BG[0].VOffset >> 8) & 0xff) |
 	    ((uint16) Byte << 8);
 	break;
-
     case 0x210F:
 	PPU.BG[1].HOffset = ((PPU.BG[1].HOffset >> 8) & 0xff) |
 	    ((uint16) Byte << 8);
@@ -341,11 +404,13 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 
     case 0x2118:
 	// VRAM write data (low)
+	IPPU.FirstVRAMRead = TRUE;
 	REGISTER_2118(Byte);
 	break;
 
     case 0x2119:
 	// VRAM write data (high)
+	IPPU.FirstVRAMRead = TRUE;
 	REGISTER_2119(Byte);
 	break;
 
@@ -591,26 +656,24 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	{
 	    FLUSH_REDRAW ();
 	    // Backgrounds 1 - 4, objects and backdrop colour add/sub enable
+#ifdef DEBUGGER
 	    if (Byte & 0x80)
 	    {
-#ifdef DEBUGGER
 		// Subtract
 		if (Memory.FillRAM[0x2130] & 0x02)
 		    missing.subscreen_sub = 1;
 		else
 		    missing.fixed_colour_sub = 1;
-#endif
 	    }
 	    else
 	    {
-#ifdef DEBUGGER
 		// Addition
 		if (Memory.FillRAM[0x2130] & 0x02)
 		    missing.subscreen_add = 1;
 		else
 		    missing.fixed_colour_add = 1;
-#endif
 	    }
+#endif
 	    Memory.FillRAM[0x2131] = Byte;
 	}
 	break;
@@ -660,7 +723,6 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
     case 0x2136:
 	// Matrix 16bit x 8bit multiply result (read-only)
 	return;
-	break; // return;
 
     case 0x2137:
 	// Software latch for horizontal and vertical timers (read-only)
@@ -701,13 +763,17 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
     case 0x2174: case 0x2175: case 0x2176: case 0x2177:
     case 0x2178: case 0x2179: case 0x217a: case 0x217b:
     case 0x217c: case 0x217d: case 0x217e: case 0x217f:
+#ifdef SPCTOOL
+	_SPCInPB (Address & 3, Byte);
+#else	
 //	CPU.Flags |= DEBUG_MODE_FLAG;
 	Memory.FillRAM [Address] = Byte;
 	IAPU.RAM [(Address & 3) + 0xf4] = Byte;
-#ifdef SPC700_SHUTDOWN	
+#ifdef SPC700_SHUTDOWN
 	IAPU.APUExecuting = Settings.APUEnabled;
 	IAPU.WaitCounter++;
 #endif
+#endif // SPCTOOL
 	break;
     case 0x2180:
 	REGISTER_2180(Byte);
@@ -738,90 +804,94 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 	    return;
 	}
 	else
-    if (Address < 0x3000 || Address >= 0x3000 + 768)
-    {
+	// Dai Kaijyu Monogatari II
+	if (Address == 0x2801 && Settings.SRTC)
+	    S9xSetSRTC (Byte, Address);
+	else
+	if (Address < 0x3000 || Address >= 0x3000 + 768)
+	{
 #ifdef DEBUGGER
-	missing.unknownppu_write = Address;
-	if (Settings.TraceUnknownRegisters)
-	{
-	    sprintf (String, "Unknown register write: $%02X->$%04X\n",
-		     Byte, Address);
-	    S9xMessage (S9X_TRACE, S9X_PPU_TRACE, String);
-	}
-#endif
-    }
-    else
-    {
-	if (!Settings.SuperFX)
-	    return;
-	
-#ifdef ZSNES_FX
-	Memory.FillRAM [Address] = Byte;
-	if (Address < 0x3040)
-	    S9xSuperFXWriteReg (Byte, Address);
-#else
-	switch (Address)
-	{
-	case 0x3030:
-	    if ((Memory.FillRAM [0x3030] ^ Byte) & FLG_G)
+	    missing.unknownppu_write = Address;
+	    if (Settings.TraceUnknownRegisters)
 	    {
-		Memory.FillRAM [Address] = Byte;
-		// Go flag has been changed
-		if (Byte & FLG_G)
-		    S9xSuperFXExec ();
+		sprintf (String, "Unknown register write: $%02X->$%04X\n",
+			 Byte, Address);
+		S9xMessage (S9X_TRACE, S9X_PPU_TRACE, String);
+	    }
+#endif
+	}
+	else
+	{
+	    if (!Settings.SuperFX)
+		return;
+	    
+    #ifdef ZSNES_FX
+	    Memory.FillRAM [Address] = Byte;
+	    if (Address < 0x3040)
+		S9xSuperFXWriteReg (Byte, Address);
+    #else
+	    switch (Address)
+	    {
+	    case 0x3030:
+		if ((Memory.FillRAM [0x3030] ^ Byte) & FLG_G)
+		{
+		    Memory.FillRAM [Address] = Byte;
+		    // Go flag has been changed
+		    if (Byte & FLG_G)
+			S9xSuperFXExec ();
+		    else
+			FxFlushCache ();
+		}
 		else
-		    FxFlushCache ();
-	    }
-	    else
+		    Memory.FillRAM [Address] = Byte;
+		break;
+
+	    case 0x3031:
 		Memory.FillRAM [Address] = Byte;
-	    break;
+		break;
+	    case 0x3033:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x3034:
+		Memory.FillRAM [Address] = Byte & 0x7f;
+		break;
+	    case 0x3036:
+		Memory.FillRAM [Address] = Byte & 0x7f;
+		break;
+	    case 0x3037:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x3038:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x3039:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x303a:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x303b:
+		break;
+	    case 0x303f:
+		Memory.FillRAM [Address] = Byte;
+		break;
+	    case 0x301f:
+		Memory.FillRAM [Address] = Byte;
+		Memory.FillRAM [0x3000 + GSU_SFR] |= FLG_G;
+		S9xSuperFXExec ();
+		return;
 
-	case 0x3031:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x3033:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x3034:
-	    Memory.FillRAM [Address] = Byte & 0x7f;
-	    break;
-	case 0x3036:
-	    Memory.FillRAM [Address] = Byte & 0x7f;
-	    break;
-	case 0x3037:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x3038:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x3039:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x303a:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x303b:
-	    break;
-	case 0x303f:
-	    Memory.FillRAM [Address] = Byte;
-	    break;
-	case 0x301f:
-	    Memory.FillRAM [Address] = Byte;
-	    Memory.FillRAM [0x3000 + GSU_SFR] |= FLG_G;
-	    S9xSuperFXExec ();
-	    return;
-
-	default:
-	    Memory.FillRAM[Address] = Byte;
-	    if (Address >= 0x3100)
-	    {
-		FxCacheWriteAccess (Address);
+	    default:
+		Memory.FillRAM[Address] = Byte;
+		if (Address >= 0x3100)
+		{
+		    FxCacheWriteAccess (Address);
+		}
+		break;
 	    }
-	    break;
+    #endif
+	    return;
 	}
-#endif
-	return;
-    }
     }
     Memory.FillRAM[Address] = Byte;
 }
@@ -832,9 +902,9 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 /**********************************************************************************************/
 uint8 S9xGetPPU (uint16 Address)
 {
-    uint8 Byte;
+    uint8 byte = 0;
 
-    if (Address <= 0x2183)
+    if (Address <= 0x2190)
     {
 	switch (Address)
 	{
@@ -940,20 +1010,24 @@ uint8 S9xGetPPU (uint16 Address)
 #endif
 #endif
 	    PPU.HVBeamCounterLatched = 1;
-	    PPU.VBeamPosLatched = CPU.V_Counter;
-	    PPU.HBeamPosLatched = (CPU.Cycles * SNES_HCOUNTER_MAX) / Settings.H_Max;
+	    PPU.VBeamPosLatched = (uint16) CPU.V_Counter;
+	    PPU.HBeamPosLatched = (uint16) ((CPU.Cycles * SNES_HCOUNTER_MAX) / Settings.H_Max);
 	    
-	    CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
+	    // Causes screen flicker for Yoshi's Island if uncommented
+	    //CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
+
+	    if (SNESGameFixes.NeedInit0x2137)
+		PPU.VBeamFlip = 0;   //jyam sword world sfc2 & godzill
 	    return (0);
 	case 0x2138:
 	    // Read OAM (sprite) control data
 	    if (!PPU.OAMReadFlip)
 	    {
-		Byte = PPU.OAMData [PPU.OAMAddr << 1];
+		byte = PPU.OAMData [PPU.OAMAddr << 1];
 	    }
 	    else
 	    {
-		Byte = PPU.OAMData [(PPU.OAMAddr << 1) + 1];
+		byte = PPU.OAMData [(PPU.OAMAddr << 1) + 1];
 		if (++PPU.OAMAddr >= 0x110)
 		    PPU.OAMAddr = 0;
 		
@@ -962,7 +1036,7 @@ uint8 S9xGetPPU (uint16 Address)
 #ifdef DEBUGGER
 	    missing.oam_read = 1;
 #endif
-	    return (Byte);
+	    return (byte);
 
 	case 0x2139:
 	    // Read vram low byte
@@ -970,7 +1044,7 @@ uint8 S9xGetPPU (uint16 Address)
 	    missing.vram_read = 1;
 #endif
 	    if (IPPU.FirstVRAMRead)
-		Byte = Memory.VRAM[PPU.VMA.Address << 1];
+		byte = Memory.VRAM[PPU.VMA.Address << 1];
 	    else
 	    if (PPU.VMA.FullGraphicCount)
 	    {
@@ -979,10 +1053,10 @@ uint8 S9xGetPPU (uint16 Address)
 		uint32 address = (addr & ~PPU.VMA.Mask1) +
 				 (rem >> PPU.VMA.Shift) +
 				 ((rem & (PPU.VMA.FullGraphicCount - 1)) << 3);
-		Byte = Memory.VRAM [((address << 1) - 2) & 0xFFFF];
+		byte = Memory.VRAM [((address << 1) - 2) & 0xFFFF];
 	    }
 	    else
-		Byte = Memory.VRAM[((PPU.VMA.Address << 1) - 2) & 0xffff];
+		byte = Memory.VRAM[((PPU.VMA.Address << 1) - 2) & 0xffff];
 
 	    if (!PPU.VMA.High)
 	    {
@@ -996,7 +1070,7 @@ uint8 S9xGetPPU (uint16 Address)
 	    missing.vram_read = 1;
 #endif
 	    if (IPPU.FirstVRAMRead)
-		Byte = Memory.VRAM[((PPU.VMA.Address << 1) + 1) & 0xffff];
+		byte = Memory.VRAM[((PPU.VMA.Address << 1) + 1) & 0xffff];
 	    else
 	    if (PPU.VMA.FullGraphicCount)
 	    {
@@ -1005,10 +1079,10 @@ uint8 S9xGetPPU (uint16 Address)
 		uint32 address = (addr & ~PPU.VMA.Mask1) +
 				 (rem >> PPU.VMA.Shift) +
 				 ((rem & (PPU.VMA.FullGraphicCount - 1)) << 3);
-		Byte = Memory.VRAM [((address << 1) - 1) & 0xFFFF];
+		byte = Memory.VRAM [((address << 1) - 1) & 0xFFFF];
 	    }
 	    else
-		Byte = Memory.VRAM[((PPU.VMA.Address << 1) - 1) & 0xFFFF];
+		byte = Memory.VRAM[((PPU.VMA.Address << 1) - 1) & 0xFFFF];
 	    if (PPU.VMA.High)
 	    {
 		PPU.VMA.Address += PPU.VMA.Increment;
@@ -1022,12 +1096,12 @@ uint8 S9xGetPPU (uint16 Address)
 	    missing.cgram_read = 1;
 #endif
 	    if (PPU.CGFLIPRead)
-		Byte = PPU.CGDATA [PPU.CGADD++] >> 8;
+		byte = PPU.CGDATA [PPU.CGADD++] >> 8;
 	    else
-		Byte = PPU.CGDATA [PPU.CGADD] & 0xff;
+		byte = PPU.CGDATA [PPU.CGADD] & 0xff;
 
 	    PPU.CGFLIPRead ^= 1;
-	    return (Byte);
+	    return (byte);
 	    
 	case 0x213C:
 	    // Horizontal counter value 0-339
@@ -1035,9 +1109,9 @@ uint8 S9xGetPPU (uint16 Address)
 	    missing.h_counter_read = 1;
 #endif
 	    if (PPU.HBeamFlip)
-		Byte = PPU.HBeamPosLatched >> 8;
+		byte = PPU.HBeamPosLatched >> 8;
 	    else
-		Byte = (uint8)PPU.HBeamPosLatched;
+		byte = (uint8)PPU.HBeamPosLatched;
 	    PPU.HBeamFlip ^= 1;
 	    break;
 	case 0x213D:
@@ -1046,14 +1120,15 @@ uint8 S9xGetPPU (uint16 Address)
 	    missing.v_counter_read = 1;
 #endif
 	    if (PPU.VBeamFlip)
-		Byte = PPU.VBeamPosLatched >> 8;
+		byte = PPU.VBeamPosLatched >> 8;
 	    else
-		Byte = (uint8)PPU.VBeamPosLatched;
+		byte = (uint8)PPU.VBeamPosLatched;
 	    PPU.VBeamFlip ^= 1;
 	    break;
 	case 0x213E:
 	    // PPU time and range over flags
-	    return (0);
+	    return (SNESGameFixes._0x213E_ReturnValue);
+
 	case 0x213F:
 	    // NTSC/PAL and which field flags
 	    PPU.VBeamFlip = PPU.HBeamFlip = 0;
@@ -1075,16 +1150,26 @@ uint8 S9xGetPPU (uint16 Address)
 	case 0x2174: case 0x2175: case 0x2176: case 0x2177:
 	case 0x2178: case 0x2179: case 0x217a: case 0x217b:
 	case 0x217c: case 0x217d: case 0x217e: case 0x217f:
+#ifdef SPCTOOL
+	    return ((uint8) _SPCOutP [Address & 3]);
+#else
     //	CPU.Flags |= DEBUG_MODE_FLAG;
 #ifdef SPC700_SHUTDOWN	
-	IAPU.APUExecuting = Settings.APUEnabled;
-	IAPU.WaitCounter++;
+	    IAPU.APUExecuting = Settings.APUEnabled;
+	    IAPU.WaitCounter++;
 #endif
 	    if (Settings.APUEnabled)
 	    {
 #ifdef CPU_SHUTDOWN
 //		CPU.WaitAddress = CPU.PCAtOpcodeStart;
 #endif	
+		if (SNESGameFixes.APU_OutPorts_ReturnValueFix &&
+		    Address >= 0x2140 && Address <= 0x2143 && !CPU.V_Counter)
+		{
+                    return (uint8)((Address & 1) ? ((rand() & 0xff00) >> 8) : 
+				   (rand() & 0xff));
+		}
+
 		return (APU.OutPorts [Address & 3]);
 	    }
 
@@ -1118,19 +1203,22 @@ uint8 S9xGetPPU (uint16 Address)
 		    return ((r >> 3) & 0xff);
 	    }
 	    return (Memory.FillRAM[Address]);
+#endif // SPCTOOL
 
 	case 0x2180:
 	    // Read WRAM
 #ifdef DEBUGGER
 	    missing.wram_read = 1;
 #endif
-	    Byte = Memory.RAM [PPU.WRAM++];
+	    byte = Memory.RAM [PPU.WRAM++];
 	    PPU.WRAM &= 0x1FFFF;
 	    break;
 	case 0x2181:
 	case 0x2182:
 	case 0x2183:
 	    return (Memory.FillRAM [Address]);
+	case 0x2190:
+	    return (1);
 	}
     }
     else
@@ -1146,6 +1234,12 @@ uint8 S9xGetPPU (uint16 Address)
 	        return (0x20);
 	    case 0x21c3:
 	        return (0);
+	    case 0x2800:
+		// For Dai Kaijyu Monogatari II
+		if (Settings.SRTC)
+		    return (S9xGetSRTC (Address));
+		/*FALL*/
+			    
 	    default:
 #ifdef DEBUGGER
 	        missing.unknownppu_read = Address;
@@ -1155,7 +1249,8 @@ uint8 S9xGetPPU (uint16 Address)
 		    S9xMessage (S9X_TRACE, S9X_PPU_TRACE, String);
 		}
 #endif
-	        return (Memory.FillRAM[Address]);
+		// XXX:
+	        return (0); //Memory.FillRAM[Address]);
 	    }
 	}
 	
@@ -1163,9 +1258,9 @@ uint8 S9xGetPPU (uint16 Address)
 	    return (0x30);
 #ifdef ZSNES_FX
 	if (Address < 0x3040)
-	    Byte = S9xSuperFXReadReg (Address);
+	    byte = S9xSuperFXReadReg (Address);
 	else
-	    Byte = Memory.FillRAM [Address];
+	    byte = Memory.FillRAM [Address];
 
 #ifdef CPU_SHUTDOWN
 	if (Address == 0x3030)
@@ -1174,7 +1269,7 @@ uint8 S9xGetPPU (uint16 Address)
 	if (Address == 0x3031)
 	    CLEAR_IRQ_SOURCE (GSU_IRQ_SOURCE);
 #else
-	Byte = Memory.FillRAM [Address];
+	byte = Memory.FillRAM [Address];
 
 //if (Address != 0x3030 && Address != 0x3031)
 //printf ("%04x\n", Address);
@@ -1188,13 +1283,13 @@ uint8 S9xGetPPU (uint16 Address)
 	if (Address == 0x3031)
 	{
 	    CLEAR_IRQ_SOURCE (GSU_IRQ_SOURCE);
-	    Memory.FillRAM [0x3031] = Byte & 0x7f;
+	    Memory.FillRAM [0x3031] = byte & 0x7f;
 	}
-	return (Byte);
+	return (byte);
 #endif
     }
 
-    return (Byte);
+    return (byte);
 }
 
 /**********************************************************************************************/
@@ -1204,7 +1299,7 @@ uint8 S9xGetPPU (uint16 Address)
 void S9xSetCPU (uint8 byte, uint16 Address)
 {
     int d;
-
+    
     if (Address < 0x4200)
     {
 #ifdef VAR_CYCLES
@@ -1241,46 +1336,66 @@ void S9xSetCPU (uint8 byte, uint16 Address)
     {
     case 0x4200:
 	// NMI, V & H IRQ and joypad reading enable flags
-	if (byte & 0x20)
+	if ((byte & 0x20) && 
+	    (!SNESGameFixes.umiharakawaseFix || PPU.IRQVBeamPos < 209))
 	{
-	    PPU.VTimerEnabled = TRUE;
+	    if (!PPU.VTimerEnabled)
+	    {
 #ifdef DEBUGGER
-	    missing.virq = 1;
-	    missing.virq_pos = PPU.IRQVBeamPos;
+		missing.virq = 1;
+		missing.virq_pos = PPU.IRQVBeamPos;
 #endif
+		PPU.VTimerEnabled = TRUE;
+		if (PPU.HTimerEnabled)
+		    S9xUpdateHTimer ();
+		else
+		if (PPU.IRQVBeamPos == CPU.V_Counter)
+		    S9xSetIRQ (PPU_V_BEAM_IRQ_SOURCE);
+	    }
 	}
 	else
+	{
 	    PPU.VTimerEnabled = FALSE;
+	    if (SNESGameFixes.umiharakawaseFix)
+		byte &= ~0x20;
+	}
+
 	if (byte & 0x10)
 	{
-	    PPU.HTimerEnabled = TRUE;
-	    PPU.HTimerPosition = PPU.IRQHBeamPos * Settings.H_Max / SNES_HCOUNTER_MAX;
-	    if (PPU.HTimerPosition == Settings.H_Max ||
-		PPU.HTimerPosition == Settings.HBlankStart)
+	    if (!PPU.HTimerEnabled)
 	    {
-		PPU.HTimerPosition--;
-	    }
 #ifdef DEBUGGER
-	    missing.hirq = 1;
-	    missing.hirq_pos = PPU.IRQHBeamPos;
+		missing.hirq = 1;
+		missing.hirq_pos = PPU.IRQHBeamPos;
 #endif
+		PPU.HTimerEnabled = TRUE;
+		S9xUpdateHTimer ();
+	    }
 	}
 	else
 	{
+	    // No need to check for HTimer being disabled as the scanline
+	    // event trigger code won't trigger an H-IRQ unless its enabled.
 	    PPU.HTimerEnabled = FALSE;
 	    PPU.HTimerPosition = Settings.H_Max + 1;
 	}
-	CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
+	if (!Settings.DaffyDuck)
+	    CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
 
-	// Needed for Daffy Duck: The Marvin Missions.
-	// ...but causes problems for Breatch of Fire 1 and 2
-	if (Settings.DaffyDuckNMIHack && (byte & 0x80) && 
+	if ((byte & 0x80) && 
 	    !(Memory.FillRAM [0x4200] & 0x80) &&
-	    CPU.V_Counter == PPU.ScreenHeight + FIRST_VISIBLE_LINE)
+	    CPU.V_Counter >= PPU.ScreenHeight + FIRST_VISIBLE_LINE &&
+	    CPU.V_Counter <= PPU.ScreenHeight + 
+		    (SNESGameFixes.alienVSpredetorFix ? 25 : 15) &&   //jyam 15->25 alien vs predetor
+// Panic Bomberman clears the NMI pending flag @ scanline 230 before enabling
+// NMIs again. The NMI routine crashes the CPU if it is called without the NMI
+// pending flag being set...
+	    (Memory.FillRAM [0x4210] & 0x80) &&
+	    !CPU.NMIActive)
 	{
-	    CPU.Flags &= ~DELAYED_NMI_FLAG;
 	    CPU.Flags |= NMI_FLAG;
 	    CPU.NMIActive = TRUE;
+	    CPU.NMICycleCount = CPU.NMITriggerPoint;
 	}
 	break;
     case 0x4201:
@@ -1315,117 +1430,58 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	    break;
 	}
     case 0x4207:
+	d = PPU.IRQHBeamPos;
 	PPU.IRQHBeamPos = (PPU.IRQHBeamPos & 0xFF00) | byte;
-	if (PPU.HTimerEnabled)
-	{
-#ifdef DEBUGGER
-	    missing.hirq_pos = PPU.IRQHBeamPos;
-#endif
-	    PPU.HTimerPosition = PPU.IRQHBeamPos * Settings.H_Max / SNES_HCOUNTER_MAX;
-	    if (PPU.HTimerPosition == Settings.H_Max ||
-		PPU.HTimerPosition == Settings.HBlankStart)
-	    {
-		PPU.HTimerPosition--;
-	    }
-	    if (!PPU.VTimerEnabled ||
-		(PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos))
-	    {
-		if (PPU.HTimerPosition >= CPU.Cycles &&
-		    PPU.HTimerPosition < CPU.NextEvent)
-		{
-		    CPU.NextEvent = PPU.HTimerPosition;
-		    CPU.WhichEvent = HTIMER_EVENT;
-		    break;
-		}
-#if 0
-// Disabled because it causes problems for Ninja Warriors
-		CPU.Flags |= IRQ_PENDING_FLAG;
-		CPU.IRQActive |= PPU_H_BEAM_IRQ_SOURCE;
-#endif
-	    }
-	}
+
+	if (PPU.HTimerEnabled && PPU.IRQHBeamPos != d)
+	    S9xUpdateHTimer ();
 	break;
+
     case 0x4208:
-	PPU.IRQHBeamPos = (PPU.IRQHBeamPos & 0xFF) |
-	    ((byte & 1) << 8);
-	if (PPU.HTimerEnabled)
-	{
-#ifdef DEBUGGER
-	    missing.hirq_pos = PPU.IRQHBeamPos;
-#endif
-	    PPU.HTimerPosition = (PPU.IRQHBeamPos * Settings.H_Max) / SNES_HCOUNTER_MAX;
-	    if (PPU.HTimerPosition == Settings.H_Max ||
-		PPU.HTimerPosition == Settings.HBlankStart)
-	    {
-		PPU.HTimerPosition--;
-	    }
-	    if (!PPU.VTimerEnabled ||
-		(PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos))
-	    {
-		if (PPU.HTimerPosition >= CPU.Cycles &&
-		    PPU.HTimerPosition < CPU.NextEvent)
-		{
-		    CPU.NextEvent = PPU.HTimerPosition;
-		    CPU.WhichEvent = HTIMER_EVENT;
-		    break;
-		}
-#if 0
-// Disabled because it causes problems for Ninja Warriors
-		CPU.Flags |= IRQ_PENDING_FLAG;
-		CPU.IRQActive |= PPU_H_BEAM_IRQ_SOURCE;
-#endif
-	    }
-	}
+	d = PPU.IRQHBeamPos;
+	PPU.IRQHBeamPos = (PPU.IRQHBeamPos & 0xFF) | ((byte & 1) << 8);
+
+	if (PPU.HTimerEnabled && PPU.IRQHBeamPos != d)
+	    S9xUpdateHTimer ();
+
 	break;
+
     case 0x4209:
+	d = PPU.IRQVBeamPos;
 	PPU.IRQVBeamPos = (PPU.IRQVBeamPos & 0xFF00) | byte;
 #ifdef DEBUGGER
 	missing.virq_pos = PPU.IRQVBeamPos;
 #endif
-	if (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos)
+	if (PPU.VTimerEnabled && PPU.IRQVBeamPos != d)
 	{
 	    if (PPU.HTimerEnabled)
+		S9xUpdateHTimer ();
+	    else
 	    {
-		if (PPU.HTimerPosition >= CPU.Cycles &&
-		    PPU.HTimerPosition < CPU.NextEvent)
-		{
-		    CPU.NextEvent = PPU.HTimerPosition;
-		    CPU.WhichEvent = HTIMER_EVENT;
-		    break;
-		}
+		if (PPU.IRQVBeamPos == CPU.V_Counter)
+		    S9xSetIRQ (PPU_V_BEAM_IRQ_SOURCE);
 	    }
-#if 0
-// Disabled for Chuck Rock
-	    CPU.Flags |= IRQ_PENDING_FLAG;
-	    CPU.IRQActive |= PPU_V_BEAM_IRQ_SOURCE;
-#endif
 	}
 	break;
+
     case 0x420A:
-	PPU.IRQVBeamPos = (PPU.IRQVBeamPos & 0xFF) |
-	    ((byte & 1) << 8);
+	d = PPU.IRQVBeamPos;
+	PPU.IRQVBeamPos = (PPU.IRQVBeamPos & 0xFF) | ((byte & 1) << 8);
 #ifdef DEBUGGER
 	missing.virq_pos = PPU.IRQVBeamPos;
 #endif
-	if (PPU.VTimerEnabled && CPU.V_Counter == PPU.IRQVBeamPos)
+	if (PPU.VTimerEnabled && PPU.IRQVBeamPos != d)
 	{
 	    if (PPU.HTimerEnabled)
+		S9xUpdateHTimer ();
+	    else
 	    {
-		if (PPU.HTimerPosition >= CPU.Cycles &&
-		    PPU.HTimerPosition < CPU.NextEvent)
-		{
-		    CPU.NextEvent = PPU.HTimerPosition;
-		    CPU.WhichEvent = HTIMER_EVENT;
-		    break;
-		}
+		if (PPU.IRQVBeamPos == CPU.V_Counter)
+		    S9xSetIRQ (PPU_V_BEAM_IRQ_SOURCE);
 	    }
-#if 0
-// Disabled for Chuck Rock
-	    CPU.Flags |= IRQ_PENDING_FLAG;
-	    CPU.IRQActive |= PPU_V_BEAM_IRQ_SOURCE;
-#endif
 	}
 	break;
+
     case 0x420B:
 #ifdef DEBUGGER
 	missing.dma_this_frame = byte;
@@ -1447,7 +1503,7 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	    S9xDoDMA (6);
 	if ((byte & 0x80) != 0)
 	    S9xDoDMA (7);
-	return;
+	break;
     case 0x420C:
 #ifdef DEBUGGER
 	missing.hdma_this_frame |= byte;
@@ -1465,13 +1521,13 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	{
 	    if (byte & 1)
 	    {
-		CPU.FastROMSpeed = 6;
+		CPU.FastROMSpeed = ONE_CYCLE;
 #ifdef DEBUGGER
 		missing.fast_rom = 1;
 #endif
 	    }
 	    else
-		CPU.FastROMSpeed = 8;
+		CPU.FastROMSpeed = SLOW_ONE_CYCLE;
 
 	    Memory.FixROMSpeed ();
 	}
@@ -1483,7 +1539,6 @@ void S9xSetCPU (uint8 byte, uint16 Address)
     case 0x4210:
 	// NMI ocurred flag (reset on read or write)
 	Memory.FillRAM[0x4210] = 0;
-	CPU.NMIActive = FALSE;
 	return;
     case 0x4211:
 	// IRQ ocurred flag (reset on read or write)
@@ -1655,6 +1710,21 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 	DMA[d].Repeat = !(byte & 0x80);
 	break;
 
+    case 0x4800:
+    case 0x4801:
+    case 0x4802:
+    case 0x4803:
+//printf ("%02x->%04x\n", byte, Address);
+	break;
+
+    case 0x4804:
+    case 0x4805:
+    case 0x4806:
+    case 0x4807:
+//printf ("%02x->%04x\n", byte, Address);
+
+	S9xSetSDD1MemoryMap (Address - 0x4804, byte & 7);
+	break;
     default:
 #ifdef DEBUGGER
 	missing.unknowncpu_write = Address;
@@ -1685,6 +1755,11 @@ uint8 S9xGetCPU (uint16 Address)
 #endif
 	switch (Address)
 	{
+	// Secret of the Evermore
+	case 0x4000:
+	case 0x4001:
+	    return (0x40);
+
 	case 0x4016:
 	{
 	    if (Memory.FillRAM [0x4016] & 1)
@@ -1764,9 +1839,17 @@ uint8 S9xGetCPU (uint16 Address)
     else
     switch (Address)
     {
-//    case 0x4200:
+	// BS Dynami Tracer! needs to be able to check if NMIs are enabled
+	// already, otherwise the game locks up.
+    case 0x4200:
 	// NMI, h & v timers and joypad reading enable
-//	return (Memory.FillRAM[0x4200]);
+	if (SNESGameFixes.Old_Read0x4200)
+	{
+#ifdef CPU_SHUTDOWN
+           CPU.WaitAddress = CPU.PCAtOpcodeStart;
+#endif
+           return (REGISTER_4212());
+	}
     case 0x4201:
 	// I/O port (output - write only?)
     case 0x4202:
@@ -1787,7 +1870,13 @@ uint8 S9xGetCPU (uint16 Address)
 	return (PPU.IRQVBeamPos >> 8);
     case 0x420b:
 	// General purpose DMA enable
-	return (0);
+	// Super Formation Soccer 95 della Serie A UCC Xaqua requires this
+	// register should not always return zero.
+	// .. But Aero 2 waits until this register goes zero..
+	// Just keep toggling the value for now in the hope that it breaks
+	// the game out of its wait loop...
+	Memory.FillRAM [0x420b] = !Memory.FillRAM [0x420b];
+	return (Memory.FillRAM [0x420b]);
     case 0x420c:
 	// H-DMA enable
 	return (IPPU.HDMA);
@@ -1804,15 +1893,13 @@ uint8 S9xGetCPU (uint16 Address)
 #endif	
 	byte = Memory.FillRAM[0x4210];
 	Memory.FillRAM[0x4210] = 0;
-	CPU.NMIActive = FALSE;
 	return (byte);
     case 0x4211:
 	byte = (CPU.IRQActive & (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE)) ? 0x80 : 0;
+	// Super Robot Wars Ex ROM bug requires this.
+	byte |= CPU.Cycles >= Settings.HBlankStart ? 0x40 : 0;
 	CLEAR_IRQ_SOURCE (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE);
-	if (CPU.V_Counter == PPU.IRQVBeamPos)
-	    byte |= 0x40;
 	return (byte);
-    case 0x4200:
     case 0x4212:
 	// V-blank, h-blank and joypads being read flags (read-only)
 #ifdef CPU_SHUTDOWN
@@ -2012,11 +2099,11 @@ void S9xResetPPU ()
     int c;
     for (c = 0; c < 256; c++)
     {
-	PPU.CGDATA [c] = ((c & 7) << 2) | (((c >> 3) & 7) << 7) |
-			 (((c >> 6) & 3) << 13);
-	IPPU.Red [c] = PPU.CGDATA [c] & 0x1f;
-	IPPU.Green [c] = (PPU.CGDATA [c] >> 5) & 0x1f;
-	IPPU.Blue [c] = (PPU.CGDATA [c] >> 10) & 0x1f;
+	IPPU.Red [c] = (c & 7) << 2;
+	IPPU.Green [c] = ((c >> 3) & 7) << 2;
+	IPPU.Blue [c] = ((c >> 6) & 2) << 3;
+	PPU.CGDATA [c] = IPPU.Red [c] | (IPPU.Green [c] << 5) |
+			 (IPPU.Blue [c] << 10);
     }
 
     PPU.FirstSprite = 0;
@@ -2105,8 +2192,8 @@ void S9xResetPPU ()
     IPPU.FirstVRAMRead = FALSE;
     IPPU.LatchedInterlace = FALSE;
     IPPU.DoubleWidthPixels = FALSE;
-    IPPU.RenderedScreenHeight = SNES_WIDTH;
-    IPPU.RenderedScreenWidth = SNES_HEIGHT;
+    IPPU.RenderedScreenWidth = SNES_WIDTH;
+    IPPU.RenderedScreenHeight = SNES_HEIGHT;
     IPPU.XB = NULL;
     for (c = 0; c < 256; c++)
 	IPPU.ScreenColors [c] = c;
@@ -2135,8 +2222,12 @@ void S9xResetPPU ()
     }
     for (c = 0; c < 0x8000; c += 0x100)
 	memset (&Memory.FillRAM [c], c >> 8, 0x100);
-    ZeroMemory (&Memory.FillRAM [0x2100], 0x40);
-    ZeroMemory (&Memory.FillRAM [0x4210], 0x10);
+
+    ZeroMemory (&Memory.FillRAM [0x2100], 0x100);
+    ZeroMemory (&Memory.FillRAM [0x4200], 0x100);
+    ZeroMemory (&Memory.FillRAM [0x4000], 0x100);
+    // For BS Suttehakkun 2...
+    ZeroMemory (&Memory.FillRAM [0x1000], 0x1000);
 }
 
 void S9xProcessMouse (int which1)
@@ -2280,7 +2371,25 @@ void S9xUpdateJoypads ()
     int i;
 
     for (i = 0; i < 5; i++)
+    {
 	IPPU.Joypads [i] = S9xReadJoypad (i);
+	if (IPPU.Joypads [i] & SNES_LEFT_MASK)
+	    IPPU.Joypads [i] &= ~SNES_RIGHT_MASK;
+	if (IPPU.Joypads [i] & SNES_UP_MASK)
+	    IPPU.Joypads [i] &= ~SNES_DOWN_MASK;
+    }
+
+    //touhaiden controller Fix
+    if (SNESGameFixes.TouhaidenControllerFix && 
+        (IPPU.Controller == SNES_JOYPAD || IPPU.Controller == SNES_MULTIPLAYER5))
+    {
+	for (i = 0; i < 5; i++)
+	{
+	    if (IPPU.Joypads [i])
+		IPPU.Joypads [i] |= 0xffff0000;
+	}
+    }
+
     // Read mouse position if enabled
     if (Settings.MouseMaster)
     {
@@ -2346,8 +2455,7 @@ void S9xSuperFXExec ()
 	    if ((GSUStatus & (FLG_G | FLG_IRQ)) == FLG_IRQ)
 	    {
 		// Trigger a GSU IRQ.
-		CPU.IRQActive |= GSU_IRQ_SOURCE;
-		CPU.Flags |= IRQ_PENDING_FLAG;
+		S9xSetIRQ (GSU_IRQ_SOURCE);
 	    }
 	}
     }
@@ -2497,10 +2605,8 @@ printf ("%06x: %d\n", t, FxEmulate (2000000));
     Memory.FillRAM [0x3030] &= ~0x20;
     if (Memory.FillRAM [0x3031] & 0x80)
     {
-	CPU.IRQActive |= GSU_IRQ_SOURCE;
-	CPU.Flags |= IRQ_PENDING_FLAG;
+	S9xSetIRQ (GSU_IRQ_SOURCE);
     }
 #endif
 }
 #endif
-
