@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,13 +38,9 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _PPU_H_
 #define _PPU_H_
 
-#include "gfx.h"
-#include "memmap.h"
-
 #define FIRST_VISIBLE_LINE 1
 
 extern uint8 GetBank;
@@ -79,8 +78,8 @@ struct InternalPPU {
     uint32 FrameCount;
     uint32 RenderedFramesCount;
     uint32 DisplayedRenderedFrameCount;
-    int    SkippedFrames;
-    int    FrameSkip;
+    uint32 SkippedFrames;
+    uint32 FrameSkip;
     uint8  *TileCache [3];
     uint8  *TileCached [3];
     bool8  FirstVRAMRead;
@@ -106,9 +105,9 @@ struct InternalPPU {
 
 struct SOBJ
 {
-    short HPos;
-    uint16  VPos;
-    uint16  Name;
+    short  HPos;
+    uint16 VPos;
+    uint16 Name;
     uint8  VFlip;
     uint8  HFlip;
     uint8  Priority;
@@ -139,31 +138,31 @@ struct SPPU {
 	uint16 SCSize;
     } BG [4];
 
-    bool8 CGFLIP;
+    bool8  CGFLIP;
     uint16 CGDATA [256]; 
-    uint8 FirstSprite;
-    uint8 LastSprite;
+    uint8  FirstSprite;
+    uint8  LastSprite;
     struct SOBJ OBJ [128];
-    uint8 OAMPriorityRotation;
+    uint8  OAMPriorityRotation;
     uint16 OAMAddr;
 
-    uint8 OAMFlip;
+    uint8  OAMFlip;
     uint16 OAMTileAddress;
     uint16 IRQVBeamPos;
     uint16 IRQHBeamPos;
     uint16 VBeamPosLatched;
     uint16 HBeamPosLatched;
 
-    uint8 HBeamFlip;
-    uint8 VBeamFlip;
-    uint8 HVBeamCounterLatched;
+    uint8  HBeamFlip;
+    uint8  VBeamFlip;
+    uint8  HVBeamCounterLatched;
 
-    short MatrixA;
-    short MatrixB;
-    short MatrixC;
-    short MatrixD;
-    short CentreX;
-    short CentreY;
+    short  MatrixA;
+    short  MatrixB;
+    short  MatrixC;
+    short  MatrixD;
+    short  CentreX;
+    short  CentreY;
     uint8  Joypad1ButtonReadPos;
     uint8  Joypad2ButtonReadPos;
 
@@ -171,21 +170,21 @@ struct SPPU {
     uint8  FixedColourRed;
     uint8  FixedColourGreen;
     uint8  FixedColourBlue;
-    uint16  SavedOAMAddr;
-    uint16  ScreenHeight;
+    uint16 SavedOAMAddr;
+    uint16 ScreenHeight;
     uint32 WRAM;
     uint8  BG_Forced;
     bool8  ForcedBlanking;
     bool8  OBJThroughMain;
     bool8  OBJThroughSub;
     uint8  OBJSizeSelect;
-    uint16  OBJNameBase;
+    uint16 OBJNameBase;
     bool8  OBJAddition;
     uint8  OAMReadFlip;
     uint8  OAMData [512 + 32];
     bool8  VTimerEnabled;
     bool8  HTimerEnabled;
-    short HTimerPosition;
+    short  HTimerPosition;
     uint8  Mosaic;
     bool8  BGMosaic [4];
     bool8  Mode7HFlip;
@@ -221,16 +220,16 @@ struct SDMA {
     uint8  TransferMode;
 
     uint8  ABank;
-    uint16  AAddress;
-    uint16  Address;
+    uint16 AAddress;
+    uint16 Address;
     uint8  BAddress;
 
     // General DMA only:
-    uint16  TransferBytes;
+    uint16 TransferBytes;
 
     // H-DMA only:
     bool8  HDMAIndirectAddressing;
-    uint16  IndirectAddress;
+    uint16 IndirectAddress;
     uint8  IndirectBank;
     uint8  Repeat;
     uint8  LineCount;
@@ -250,12 +249,21 @@ uint8 S9xGetPPU (uint16 Address);
 void S9xSetCPU (uint8 Byte, uint16 Address);
 uint8 S9xGetCPU (uint16 Address);
 
+void S9xInitC4 ();
+void S9xSetC4 (uint8 Byte, uint16 Address);
+uint8 S9xGetC4 (uint16 Address);
+void S9xSetC4RAM (uint8 Byte, uint16 Address);
+uint8 S9xGetC4RAM (uint16 Address);
+
 extern struct SPPU PPU;
 extern struct SDMA DMA [8];
 extern struct InternalPPU IPPU;
 END_EXTERN_C
 
-inline uint8 REGISTER_4212()
+#include "gfx.h"
+#include "memmap.h"
+
+STATIC inline uint8 REGISTER_4212()
 {
     GetBank = 0;
     if (CPU.V_Counter >= PPU.ScreenHeight + FIRST_VISIBLE_LINE &&
@@ -263,112 +271,92 @@ inline uint8 REGISTER_4212()
 	GetBank = 1;
 
     GetBank |= CPU.Cycles >= Settings.HBlankStart ? 0x40 : 0;
-    if (CPU.V_Counter < FIRST_VISIBLE_LINE || 
-	CPU.V_Counter >= PPU.ScreenHeight + FIRST_VISIBLE_LINE)
+    if (CPU.V_Counter >= PPU.ScreenHeight + FIRST_VISIBLE_LINE)
 	GetBank |= 0x80; /* XXX: 0x80 or 0xc0 ? */
+
     return (GetBank);
 }
 
-inline void FLUSH_REDRAW ()
+STATIC inline void FLUSH_REDRAW ()
 {
     if (IPPU.PreviousLine != IPPU.CurrentLine)
 	S9xUpdateScreen ();
 }
 
-inline void REGISTER_2104 (uint8 Byte)
+STATIC inline void REGISTER_2104 (uint8 byte)
 {
     if (PPU.OAMAddr >= 0x110)
-	return;//PPU.OAMAddr %= 0x110;
+	return;
 	
-    if (!PPU.OAMFlip)
+    int addr = (PPU.OAMAddr << 1) + (PPU.OAMFlip & 1);
+    
+    if (byte != PPU.OAMData [addr])
     {
-	if (Byte != PPU.OAMData [PPU.OAMAddr << 1])
+	FLUSH_REDRAW ();
+	PPU.OAMData [addr] = byte;
+	IPPU.OBJChanged = TRUE;
+	if (addr & 0x200)
 	{
-	    FLUSH_REDRAW ();
-	    PPU.OAMData [PPU.OAMAddr << 1] = Byte;
-	    IPPU.OBJChanged = TRUE;
-	    if ((PPU.OAMAddr & 0x100) == 0)
+	    // X position high bit, and sprite size (x4)
+	    struct SOBJ *pObj = &PPU.OBJ [(addr & 0x1f) * 4];
+
+	    pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(byte >> 0) & 1];
+	    pObj++->Size = byte & 2;
+	    pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(byte >> 2) & 1];
+	    pObj++->Size = byte & 8;
+	    pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(byte >> 4) & 1];
+	    pObj++->Size = byte & 32;
+	    pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(byte >> 6) & 1];
+	    pObj->Size = byte & 128;
+	}
+	else
+	{
+	    if (addr & 1)
 	    {
-		if (!(PPU.OAMAddr & 1))
+		if (addr & 2)
 		{
-		    // X position (low)
-		    PPU.OBJ[PPU.OAMAddr >> 1].HPos &= 0xFF00;
-		    PPU.OBJ[PPU.OAMAddr >> 1].HPos |= Byte;
+		    addr = PPU.OAMAddr >> 1;
+		    // Tile within group, priority, h and v flip.
+		    PPU.OBJ[addr].Name &= 0xFF;
+		    PPU.OBJ[addr].Name |= ((uint16) (byte & 1)) << 8;
+		    PPU.OBJ[addr].Palette = (byte >> 1) & 7;
+		    PPU.OBJ[addr].Priority = (byte >> 4) & 3;
+		    PPU.OBJ[addr].HFlip = (byte >> 6) & 1;
+		    PPU.OBJ[addr].VFlip = (byte >> 7) & 1;
 		}
 		else
-		{
-		    // Tile group
-		    PPU.OBJ[PPU.OAMAddr >> 1].Name &= 0x100;
-		    PPU.OBJ[PPU.OAMAddr >> 1].Name |= Byte;
-		}
-	    }
-	    else
-	    {
-		// X position high bit, and sprite size (x4)
-		struct SOBJ *pObj = &PPU.OBJ [(PPU.OAMAddr & 0x0F) * 8];
-
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 0) & 1];
-		pObj++->Size = Byte & 2;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 2) & 1];
-		pObj++->Size = Byte & 8;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 4) & 1];
-		pObj++->Size = Byte & 32;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 6) & 1];
-		pObj->Size = Byte & 128;
-	    }
-	}
-    }
-    else
-    {
-	if (Byte != PPU.OAMData [(PPU.OAMAddr << 1) + 1])
-	{
-	    FLUSH_REDRAW ();
-	    PPU.OAMData [(PPU.OAMAddr << 1) + 1] = Byte;
-	    IPPU.OBJChanged = TRUE;
-	    
-	    if ((PPU.OAMAddr & 0x100) == 0)
-	    {
-		if (!(PPU.OAMAddr & 1))
 		{
 		    // Sprite Y position
-		    PPU.OBJ[PPU.OAMAddr >> 1].VPos = Byte;
-		}
-		else
-		{
-		    // Tile within group, priority, h and v flip.
-		    PPU.OBJ[PPU.OAMAddr >> 1].Name &= 0xFF;
-		    PPU.OBJ[PPU.OAMAddr >> 1].Name |= ((uint16) (Byte & 1)) << 8;
-		    PPU.OBJ[PPU.OAMAddr >> 1].Palette = (Byte >> 1) & 7;
-		    PPU.OBJ[PPU.OAMAddr >> 1].Priority = (Byte >> 4) & 3;
-		    PPU.OBJ[PPU.OAMAddr >> 1].HFlip = (Byte >> 6) & 1;
-		    PPU.OBJ[PPU.OAMAddr >> 1].VFlip = (Byte >> 7) & 1;
+		    PPU.OBJ[PPU.OAMAddr >> 1].VPos = byte;
 		}
 	    }
 	    else
 	    {
-		// X position high bit, and sprite size (x4)
-		struct SOBJ *pObj = &PPU.OBJ [((PPU.OAMAddr & 0x0F) * 8) + 4];
-
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 0) & 1];
-		pObj++->Size = Byte & 2;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 2) & 1];
-		pObj++->Size = Byte & 8;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 4) & 1];
-		pObj++->Size = Byte & 32;
-		pObj->HPos = (pObj->HPos & 0xFF) | SignExtend[(Byte >> 6) & 1];
-		pObj->Size = Byte & 128;
+		if (addr & 2)
+		{
+		    // Tile group
+		    
+		    PPU.OBJ[addr = PPU.OAMAddr >> 1].Name &= 0x100;
+		    PPU.OBJ[addr].Name |= byte;
+		}
+		else
+		{
+		    // X position (low)
+		    PPU.OBJ[addr = PPU.OAMAddr >> 1].HPos &= 0xFF00;
+		    PPU.OBJ[addr].HPos |= byte;
+		}
 	    }
 	}
     }
-
-    if ((PPU.OAMFlip ^= 1) == 0)
+    PPU.OAMFlip ^= 1;
+    if (!(PPU.OAMFlip & 1))
 	PPU.OAMAddr++;
-    Memory.FillRAM [0x2104] = Byte;
+
+    Memory.FillRAM [0x2104] = byte;
 }
 
-inline void REGISTER_2118 (uint8 Byte)
+STATIC inline void REGISTER_2118 (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 address;
     if (PPU.VMA.FullGraphicCount)
     {
@@ -397,12 +385,11 @@ inline void REGISTER_2118 (uint8 Byte)
 #endif	
 	PPU.VMA.Address += PPU.VMA.Increment;
     }
-    Memory.FillRAM [0x2118] = Byte;
+//    Memory.FillRAM [0x2118] = Byte;
 }
 
-inline void REGISTER_2118_tile (uint8 Byte)
+STATIC inline void REGISTER_2118_tile (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 address;
     uint32 rem = PPU.VMA.Address & PPU.VMA.Mask1;
     address = (((PPU.VMA.Address & ~PPU.VMA.Mask1) +
@@ -414,12 +401,11 @@ inline void REGISTER_2118_tile (uint8 Byte)
     IPPU.TileCached [TILE_8BIT][address >> 6] = FALSE;
     if (!PPU.VMA.High)
 	PPU.VMA.Address += PPU.VMA.Increment;
-    Memory.FillRAM [0x2118] = Byte;
+//    Memory.FillRAM [0x2118] = Byte;
 }
 
-inline void REGISTER_2118_linear (uint8 Byte)
+STATIC inline void REGISTER_2118_linear (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 address;
     Memory.VRAM[address = (PPU.VMA.Address << 1) & 0xFFFF] = Byte;
     IPPU.TileCached [TILE_2BIT][address >> 4] = FALSE;
@@ -427,12 +413,11 @@ inline void REGISTER_2118_linear (uint8 Byte)
     IPPU.TileCached [TILE_8BIT][address >> 6] = FALSE;
     if (!PPU.VMA.High)
 	PPU.VMA.Address += PPU.VMA.Increment;
-    Memory.FillRAM [0x2118] = Byte;
+//    Memory.FillRAM [0x2118] = Byte;
 }
 
-inline void REGISTER_2119 (uint8 Byte)
+STATIC inline void REGISTER_2119 (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 address;
     if (PPU.VMA.FullGraphicCount)
     {
@@ -461,12 +446,11 @@ inline void REGISTER_2119 (uint8 Byte)
 #endif	
 	PPU.VMA.Address += PPU.VMA.Increment;
     }
-    Memory.FillRAM [0x2119] = Byte;
+//    Memory.FillRAM [0x2119] = Byte;
 }
 
-inline void REGISTER_2119_tile (uint8 Byte)
+STATIC inline void REGISTER_2119_tile (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 rem = PPU.VMA.Address & PPU.VMA.Mask1;
     uint32 address = ((((PPU.VMA.Address & ~PPU.VMA.Mask1) +
 		    (rem >> PPU.VMA.Shift) +
@@ -477,12 +461,11 @@ inline void REGISTER_2119_tile (uint8 Byte)
     IPPU.TileCached [TILE_8BIT][address >> 6] = FALSE;
     if (PPU.VMA.High)
 	PPU.VMA.Address += PPU.VMA.Increment;
-    Memory.FillRAM [0x2119] = Byte;
+//    Memory.FillRAM [0x2119] = Byte;
 }
 
-inline void REGISTER_2119_linear (uint8 Byte)
+STATIC inline void REGISTER_2119_linear (uint8 Byte)
 {
-    IPPU.FirstVRAMRead = TRUE;
     uint32 address;
     Memory.VRAM[address = ((PPU.VMA.Address << 1) + 1) & 0xFFFF] = Byte;
     IPPU.TileCached [TILE_2BIT][address >> 4] = FALSE;
@@ -490,10 +473,10 @@ inline void REGISTER_2119_linear (uint8 Byte)
     IPPU.TileCached [TILE_8BIT][address >> 6] = FALSE;
     if (PPU.VMA.High)
 	PPU.VMA.Address += PPU.VMA.Increment;
-    Memory.FillRAM [0x2119] = Byte;
+//    Memory.FillRAM [0x2119] = Byte;
 }
 
-inline void REGISTER_2122(uint8 Byte)
+STATIC inline void REGISTER_2122(uint8 Byte)
 {
     // CG-RAM (palette) write
 
@@ -537,10 +520,10 @@ inline void REGISTER_2122(uint8 Byte)
 	}
     }
     PPU.CGFLIP ^= 1;
-    Memory.FillRAM [0x2122] = Byte;
+//    Memory.FillRAM [0x2122] = Byte;
 }
 
-inline void REGISTER_2180(uint8 Byte)
+STATIC inline void REGISTER_2180(uint8 Byte)
 {
     Memory.RAM[PPU.WRAM++] = Byte;
     PPU.WRAM &= 0x1FFFF;
