@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,13 +38,16 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "ppu.h"
 #include "cpuexec.h"
 
 #include "sa1.h"
 
+static void S9xSA1CharConv2 ();
+static void S9xSA1DMA ();
+static void S9xSA1ReadVariableLengthData (bool8 inc, bool8 no_shift);
+
 void S9xSA1Init ()
 {
     SA1.NMIActive = FALSE;
@@ -89,6 +95,7 @@ void S9xSA1Reset ()
     SA1.S9xOpcodes = S9xSA1OpcodesM1X1;
 
     S9xSA1UnpackStatus();
+    S9xSA1FixCycles ();
     SA1.Executing = TRUE;
     SA1.BWRAM = Memory.SRAM;
     Memory.FillRAM [0x2225] = 0;
@@ -140,36 +147,18 @@ void S9xFixSA1AfterSnapshotLoad ()
 
 uint8 S9xSA1GetByte (uint32 address)
 {
-#ifdef CPU_SHUTDOWN
-    int block;
-    uint8 *Getaddress = SA1.Map [block = ((address >> MEMMAP_SHIFT) & MEMMAP_MASK)];
-    if (Getaddress >= (uint8 *) CMemory::MAP_LAST)
-    {
-        if (SA1.BlockIsRAM [block])
-	    SA1.WaitAddress = SA1.PCAtOpcodeStart;
+    uint8 *GetAddress = SA1.Map [(address >> MEMMAP_SHIFT) & MEMMAP_MASK];
+    if (GetAddress >= (uint8 *) CMemory::MAP_LAST)
+	return (*(GetAddress + (address & 0xffff)));
 
-        return (*(Getaddress + (address & 0xffff)));
-    }
-#else
-    uint8 *Getaddress = SA1.Map [(address >> MEMMAP_SHIFT) & MEMMAP_MASK];
-    if (Getaddress >= (uint8 *) CMemory::MAP_LAST)
-	return (*(Getaddress + (address & 0xffff)));
-#endif
-
-    switch ((int) Getaddress)
+    switch ((int) GetAddress)
     {
     case CMemory::MAP_PPU:
 	return (S9xGetSA1 (address & 0xffff));
     case CMemory::MAP_LOROM_SRAM:
     case CMemory::MAP_SA1RAM:
-#ifdef CPU_SHUTDOWN
-	SA1.WaitAddress = SA1.PCAtOpcodeStart;
-#endif
 	return (*(Memory.SRAM + (address & 0xffff)));
     case CMemory::MAP_BWRAM:
-#ifdef CPU_SHUTDOWN
-	SA1.WaitAddress = SA1.PCAtOpcodeStart;
-#endif
 	return (*(SA1.BWRAM + ((address & 0x7fff) - 0x6000)));
     case CMemory::MAP_BWRAM_BITMAP:
 	address -= 0x600000;
@@ -187,7 +176,7 @@ uint8 S9xSA1GetByte (uint32 address)
     case CMemory::MAP_DEBUG:
     default:
 #ifdef DEBUGGER
-	printf ("R(B) %06x\n", address);
+//	printf ("R(B) %06x\n", address);
 #endif
 
 	return (0);
@@ -205,9 +194,6 @@ void S9xSA1SetByte (uint8 byte, uint32 address)
 
     if (Setaddress >= (uint8 *) CMemory::MAP_LAST)
     {
-#ifdef CPU_SHUTDOWN
-	SA1.WaitAddress = NULL;
-#endif
 	*(Setaddress + (address & 0xffff)) = byte;
 	return;
     }
@@ -266,15 +252,15 @@ void S9xSA1SetWord (uint16 Word, uint32 address)
 
 void S9xSA1SetPCBase (uint32 address)
 {
-    uint8 *Getaddress = SA1.Map [(address >> MEMMAP_SHIFT) & MEMMAP_MASK];
-    if (Getaddress >= (uint8 *) CMemory::MAP_LAST)
+    uint8 *GetAddress = SA1.Map [(address >> MEMMAP_SHIFT) & MEMMAP_MASK];
+    if (GetAddress >= (uint8 *) CMemory::MAP_LAST)
     {
-	SA1.PCBase = Getaddress;
-	SA1.PC = Getaddress + (address & 0xffff);
+	SA1.PCBase = GetAddress;
+	SA1.PC = GetAddress + (address & 0xffff);
 	return;
     }
 
-    switch ((int) Getaddress)
+    switch ((int) GetAddress)
     {
     case CMemory::MAP_PPU:
 	SA1.PCBase = Memory.FillRAM - 0x2000;
@@ -363,22 +349,15 @@ void S9xSetSA1MemMap (uint32 which1, uint8 map)
 
 uint8 S9xGetSA1 (uint32 address)
 {
+//	printf ("R: %04x\n", address);
     switch (address)
     {
     case 0x2300:
-#ifdef CPU_SHUTDOWN
-//	CPU.WaitAddress = CPU.PCAtOpcodeStart;
-	SA1.Executing = !SA1.Waiting;
-#endif
-	return ((Memory.FillRAM [0x2209] & 0x5f) | 
-		 (CPU.IRQActive & (SA1_IRQ_SOURCE | SA1_DMA_IRQ_SOURCE)));
+	return ((uint8) ((Memory.FillRAM [0x2209] & 0x5f) | 
+		 (CPU.IRQActive & (SA1_IRQ_SOURCE | SA1_DMA_IRQ_SOURCE))));
     case 0x2301:
-#ifdef CPU_SHUTDOWN
-	SA1.WaitAddress = SA1.PCAtOpcodeStart;
-#endif
 	return ((Memory.FillRAM [0x2200] & 0xf) |
-		(SA1.IRQActive & (SNES_IRQ_SOURCE | TIMER_IRQ_SOURCE |
-				  DMA_IRQ_SOURCE)));
+		(Memory.FillRAM [0x2301] & 0xf0));
     case 0x2306:
 	return ((uint8)  SA1.sum);
     case 0x2307:
@@ -389,9 +368,20 @@ uint8 S9xGetSA1 (uint32 address)
 	return ((uint8) (SA1.sum >> 24));
     case 0x230a:
 	return ((uint8) (SA1.sum >> 32));
+    case 0x230c:
+	return (Memory.FillRAM [0x230c]);
+    case 0x230d:
+    {
+	uint8 byte = Memory.FillRAM [0x230d];
 
+	if (Memory.FillRAM [0x2258] & 0x80)
+	{
+	    S9xSA1ReadVariableLengthData (TRUE, FALSE);
+	}
+	return (byte);
+    }
     default:	
-//	printf ("R: %04x\n", address);
+	printf ("R: %04x\n", address);
 	break;
     }
     return (Memory.FillRAM [address]);
@@ -399,17 +389,17 @@ uint8 S9xGetSA1 (uint32 address)
 
 void S9xSetSA1 (uint8 byte, uint32 address)
 {
-    static char *DMAsource [4] = { "ROM", "BW-RAM", "I-RAM", "???" };
-
 //printf ("W: %02x -> %04x\n", byte, address);
     switch (address)
     {
     case 0x2200:
 	SA1.Waiting = (byte & 0x60) != 0;
-	SA1.Executing = !SA1.Waiting;
+//	SA1.Executing = !SA1.Waiting && SA1.S9xOpcodes;
 
 	if (!(byte & 0x20) && (Memory.FillRAM [0x2200] & 0x20))
+	{
 	    S9xSA1Reset ();
+	}
 	if (byte & 0x80)
 	{
 	    Memory.FillRAM [0x2301] |= 0x80;
@@ -417,17 +407,17 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 	    {
 		SA1.Flags |= IRQ_PENDING_FLAG;
 		SA1.IRQActive |= SNES_IRQ_SOURCE;
-		SA1.Executing = !SA1.Waiting;
+		SA1.Executing = !SA1.Waiting && SA1.S9xOpcodes;
 	    }
 	}
 	if (byte & 0x10)
 	{
 	    Memory.FillRAM [0x2301] |= 0x10;
-	    if (Memory.FillRAM [0x220a] & 0x10)
-	    {
 #ifdef DEBUGGER
 		printf ("###SA1 NMI\n");
 #endif
+	    if (Memory.FillRAM [0x220a] & 0x10)
+	    {
 	    }
 	}
 	break;
@@ -493,27 +483,27 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 	{
 	    SA1.Flags |= IRQ_PENDING_FLAG;
 	    SA1.IRQActive |= SNES_IRQ_SOURCE;
-	    SA1.Executing = !SA1.Waiting;
+//	    SA1.Executing = !SA1.Waiting;
 	}
 	if (((byte ^ Memory.FillRAM [0x220a]) & 0x40) &&
 	    (Memory.FillRAM [0x2301] & byte & 0x40))
 	{
 	    SA1.Flags |= IRQ_PENDING_FLAG;
 	    SA1.IRQActive |= TIMER_IRQ_SOURCE;
-	    SA1.Executing = !SA1.Waiting;
+//	    SA1.Executing = !SA1.Waiting;
 	}
 	if (((byte ^ Memory.FillRAM [0x220a]) & 0x20) &&
 	    (Memory.FillRAM [0x2301] & byte & 0x20))
 	{
 	    SA1.Flags |= IRQ_PENDING_FLAG;
 	    SA1.IRQActive |= DMA_IRQ_SOURCE;
-	    SA1.Executing = !SA1.Waiting;
+//	    SA1.Executing = !SA1.Waiting;
 	}
 	if (((byte ^ Memory.FillRAM [0x220a]) & 0x10) &&
 	    (Memory.FillRAM [0x2301] & byte & 0x10))
 	{
 #ifdef DEBUGGER
-		printf ("###SA1 NMI\n");
+	    printf ("###SA1 NMI\n");
 #endif
 	}
 	break;
@@ -563,19 +553,19 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 #endif
 	break;
     case 0x2211:
-//	printf ("Timer reset\n");
+	printf ("Timer reset\n");
 	break;
     case 0x2212:
-//	printf ("H-Timer %04x\n", byte | (Memory.FillRAM [0x2213] << 8));
+	printf ("H-Timer %04x\n", byte | (Memory.FillRAM [0x2213] << 8));
 	break;
     case 0x2213:
-//	printf ("H-Timer %04x\n", (byte << 8) | Memory.FillRAM [0x2212]);
+	printf ("H-Timer %04x\n", (byte << 8) | Memory.FillRAM [0x2212]);
 	break;
     case 0x2214:
-//	printf ("V-Timer %04x\n", byte | (Memory.FillRAM [0x2215] << 8));
+	printf ("V-Timer %04x\n", byte | (Memory.FillRAM [0x2215] << 8));
 	break;
     case 0x2215:
-//	printf ("V-Timer %04x\n", (byte << 8) | Memory.FillRAM [0x2214]);
+	printf ("V-Timer %04x\n", (byte << 8) | Memory.FillRAM [0x2214]);
 	break;
     case 0x2220:
     case 0x2221:
@@ -639,37 +629,36 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 #endif
 	break;
     case 0x2235:
+	Memory.FillRAM [address] = byte;
+	break;
     case 0x2236:
+	Memory.FillRAM [address] = byte;
+	if ((Memory.FillRAM [0x2230] & 0xa4) == 0x80)
+	{
+	    // Normal DMA to I-RAM
+	    S9xSA1DMA ();
+	}
+	else
+	if ((Memory.FillRAM [0x2230] & 0xb0) == 0xb0)
+	{
+	    Memory.FillRAM [0x2300] |= 0x20;
+	    if (Memory.FillRAM [0x2201] & 0x20)
+		S9xSetIRQ (SA1_DMA_IRQ_SOURCE);
+	    SA1.in_char_dma = TRUE;
+	}
+	break;
     case 0x2237:
 	Memory.FillRAM [address] = byte;
+	if ((Memory.FillRAM [0x2230] & 0xa4) == 0x84)
+	{
+	    // Normal DMA to BW-RAM
+	    S9xSA1DMA ();
+	}
 #if 0
 	printf ("DMA dest address %06x\n", 
 		Memory.FillRAM [0x2235] | (Memory.FillRAM [0x2236] << 8) |
 		(Memory.FillRAM [0x2237] << 16));
 #endif
-	// XXX:
-	if (address == 0x2236)
-	{
-	    if ((Memory.FillRAM [0x2230] & 0xb0) == 0xb0)
-	    {
-		Memory.FillRAM [0x2300] |= 0x20;
-		if (Memory.FillRAM [0x2201] & 0x20)
-		    S9xSetIRQ (SA1_DMA_IRQ_SOURCE);
-		SA1.in_char_dma = TRUE;
-	    }
-	    else
-	    {
-		Memory.FillRAM [0x2301] |= 0x20;
-		
-		if (Memory.FillRAM [0x220a] & 0x20)
-		{
-		    SA1.Flags |= IRQ_PENDING_FLAG;
-		    SA1.IRQActive |= DMA_IRQ_SOURCE;
-		    SA1.Executing = !SA1.Waiting;
-		}
-	    }
-
-	}
 	break;
     case 0x2238:
     case 0x2239:
@@ -684,6 +673,34 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 	//printf ("virtual VRAM depth %d\n", (byte & 0x80) ? 2 : 4);
 	break;
 
+    case 0x2240:    case 0x2241:    case 0x2242:    case 0x2243:
+    case 0x2244:    case 0x2245:    case 0x2246:    case 0x2247:
+    case 0x2248:    case 0x2249:    case 0x224a:    case 0x224b:
+    case 0x224c:    case 0x224d:    case 0x224e:
+#if 0
+	if (!(SA1.Flags & TRACE_FLAG))
+	{
+	    TraceSA1 ();
+	    Trace ();
+	}
+#endif
+	Memory.FillRAM [address] = byte;
+	break;
+
+    case 0x224f:
+	Memory.FillRAM [address] = byte;
+	if ((Memory.FillRAM [0x2230] & 0xb0) == 0xa0)
+	{
+	    // Char conversion 2 DMA enabled
+	    memmove (&Memory.ROM [CMemory::MAX_ROM_SIZE - 0x10000] + SA1.in_char_dma * 16,
+		     &Memory.FillRAM [0x2240], 16);
+	    SA1.in_char_dma = (SA1.in_char_dma + 1) & 7;
+	    if ((SA1.in_char_dma & 3) == 0)
+	    {
+		S9xSA1CharConv2 ();
+	    }
+	}
+	break;
     case 0x2250:
 	if (byte & 2)
 	    SA1.sum = 0;
@@ -723,6 +740,18 @@ void S9xSetSA1 (uint8 byte, uint32 address)
 	    break;
 	}
 	break;
+    case 0x2258:    // Variable bit-field length/auto inc/start.
+	Memory.FillRAM [0x2258] = byte;
+	S9xSA1ReadVariableLengthData (TRUE, FALSE);
+	return;
+    case 0x2259:
+    case 0x225a:
+    case 0x225b:    // Variable bit-field start address
+	Memory.FillRAM [address] = byte;
+	// XXX: ???
+	SA1.variable_bit_pos = 0;
+	S9xSA1ReadVariableLengthData (FALSE, TRUE);
+	return;
     default:
 //	printf ("W: %02x->%04x\n", byte, address);
 	break;
@@ -730,3 +759,134 @@ void S9xSetSA1 (uint8 byte, uint32 address)
     if (address >= 0x2200 && address <= 0x22ff)
 	Memory.FillRAM [address] = byte;
 }
+
+static void S9xSA1CharConv2 ()
+{
+    uint32 dest = Memory.FillRAM [0x2235] | (Memory.FillRAM [0x2236] << 8);
+    uint32 offset = (SA1.in_char_dma & 7) ? 0 : 1;
+    int depth = (Memory.FillRAM [0x2231] & 3) == 0 ? 8 :
+		(Memory.FillRAM [0x2231] & 3) == 1 ? 4 : 2;
+    int bytes_per_char = 8 * depth;
+    uint8 *p = &Memory.FillRAM [0x3000] + dest + offset * bytes_per_char;
+    uint8 *q = &Memory.ROM [CMemory::MAX_ROM_SIZE - 0x10000] + offset * 64;
+
+    switch (depth)
+    {
+    case 2:
+	break;
+    case 4:
+	break;
+    case 8:
+	for (int l = 0; l < 8; l++, q += 8)
+	{
+	    for (int b = 0; b < 8; b++)
+	    {
+		uint8 r = *(q + b);
+		*(p +  0) = (*(p +  0) << 1) | ((r >> 0) & 1);
+		*(p +  1) = (*(p +  1) << 1) | ((r >> 1) & 1);
+		*(p + 16) = (*(p + 16) << 1) | ((r >> 2) & 1);
+		*(p + 17) = (*(p + 17) << 1) | ((r >> 3) & 1);
+		*(p + 32) = (*(p + 32) << 1) | ((r >> 4) & 1);
+		*(p + 33) = (*(p + 33) << 1) | ((r >> 5) & 1);
+		*(p + 48) = (*(p + 48) << 1) | ((r >> 6) & 1);
+		*(p + 49) = (*(p + 49) << 1) | ((r >> 7) & 1);
+	    }
+	    p += 2;
+	}
+	break;
+    }
+}
+
+static void S9xSA1DMA ()
+{
+    uint32 src =  Memory.FillRAM [0x2232] |
+	         (Memory.FillRAM [0x2233] << 8) |
+		 (Memory.FillRAM [0x2234] << 16);
+    uint32 dst =  Memory.FillRAM [0x2235] |
+	         (Memory.FillRAM [0x2236] << 8) |
+		 (Memory.FillRAM [0x2237] << 16);
+    uint32 len =  Memory.FillRAM [0x2238] |
+		 (Memory.FillRAM [0x2239] << 8);
+
+    uint8 *s;
+    uint8 *d;
+
+    switch (Memory.FillRAM [0x2230] & 3)
+    {
+    case 0: // ROM
+	s = SA1.Map [(src >> MEMMAP_SHIFT) & MEMMAP_MASK];
+	if (s >= (uint8 *) CMemory::MAP_LAST)
+	    s += (src & 0xffff);
+	else
+	    s = Memory.ROM + (src & 0xffff);
+	break;
+    case 1: // BW-RAM
+	src &= Memory.SRAMMask;
+	len &= Memory.SRAMMask;
+	s = Memory.SRAM + src;
+	break;
+    default:
+    case 2:
+	src &= 0x3ff;
+	len &= 0x3ff;
+	s = &Memory.FillRAM [0x3000] + src;
+	break;
+    }
+
+    if (Memory.FillRAM [0x2230] & 4)
+    {
+	dst &= Memory.SRAMMask;
+	len &= Memory.SRAMMask;
+	d = Memory.SRAM + dst;
+    }
+    else
+    {
+	dst &= 0x3ff;
+	len &= 0x3ff;
+	d = &Memory.FillRAM [0x3000] + dst;
+    }
+    memmove (d, s, len);
+    Memory.FillRAM [0x2301] |= 0x20;
+    
+    if (Memory.FillRAM [0x220a] & 0x20)
+    {
+	SA1.Flags |= IRQ_PENDING_FLAG;
+	SA1.IRQActive |= DMA_IRQ_SOURCE;
+//	SA1.Executing = !SA1.Waiting;
+    }
+}
+
+void S9xSA1ReadVariableLengthData (bool8 inc, bool8 no_shift)
+{
+    uint32 addr =  Memory.FillRAM [0x2259] |
+		  (Memory.FillRAM [0x225a] << 8) |
+		  (Memory.FillRAM [0x225b] << 16);
+    uint8 shift = Memory.FillRAM [0x2258] & 15;
+
+    if (no_shift)
+	shift = 0;
+    else
+    if (shift == 0)
+	shift = 16;
+
+    uint8 s = shift + SA1.variable_bit_pos;
+
+    if (s >= 16)
+    {
+	addr += (s >> 4) << 1;
+	s &= 15;
+    }
+    uint32 data = S9xSA1GetWord (addr) |
+		  (S9xSA1GetWord (addr + 2) << 16);
+
+    data >>= s;
+    Memory.FillRAM [0x230c] = (uint8) data;
+    Memory.FillRAM [0x230d] = (uint8) (data >> 8);
+    if (inc)
+    {
+	SA1.variable_bit_pos = (SA1.variable_bit_pos + shift) & 15;
+	Memory.FillRAM [0x2259] = (uint8) addr;
+	Memory.FillRAM [0x225a] = (uint8) (addr >> 8);
+	Memory.FillRAM [0x225b] = (uint8) (addr >> 16);
+    }
+}
