@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _sa1_h_
 #define _sa1_h_
 
@@ -75,12 +77,13 @@ struct SSA1 {
     uint8   *PCAtOpcodeStart;
     uint8   *WaitAddress;
     uint32  WaitCounter;
+    uint8   *WaitByteAddress1;
+    uint8   *WaitByteAddress2;
 //    long    Cycles;
 //    long    NextEvent;
 //    long    V_Counter;
     uint8   *Map [MEMMAP_NUM_BLOCKS];
     uint8   *WriteMap [MEMMAP_NUM_BLOCKS];
-    bool8   BlockIsRAM [MEMMAP_NUM_BLOCKS];
     int16   op1;
     int16   op2;
     int     arithmetic_op;
@@ -88,6 +91,7 @@ struct SSA1 {
     bool8   overflow;
     uint8   VirtualBitmapFormat;
     bool8   in_char_dma;
+    uint8   variable_bit_pos;
 };
 
 #define SA1CheckZero() (SA1._Zero == 0)
