@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,12 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
-/**********************************************************************************************/
-/* Snes96.CPP                                                                                 */
-/* This file contains all the initialization code of the windows part of Snes96               */
-/**********************************************************************************************/
-
 #include <stdlib.h>
 #include <stdio.h>
 
@@ -49,10 +46,6 @@
 #include "display.h"
 #include "cheats.h"
 
-/**********************************************************************************************/
-/* Globals                                                                                    */
-/**********************************************************************************************/
-
 #ifdef DEBUGGER
 extern FILE *trace;
 #endif
@@ -186,12 +179,12 @@ char *S9xParseArgs (char **argv, int argc)
 	    if (strcasecmp (argv [i], "-so") == 0 ||
 		     strcasecmp (argv [i], "-sound") == 0)
 	    {
-		Settings.APUEnabled = TRUE;
+		Settings.NextAPUEnabled = TRUE;
 	    }
 	    else if (strcasecmp (argv [i], "-ns") == 0 ||
 		     strcasecmp (argv [i], "-nosound") == 0)
 	    {
-		Settings.APUEnabled = FALSE;
+		Settings.NextAPUEnabled = FALSE;
 	    }
 	    else if (strcasecmp (argv [i], "-soundskip") == 0 ||
 		     strcasecmp (argv [i], "-sk") == 0)
@@ -205,7 +198,8 @@ char *S9xParseArgs (char **argv, int argc)
 		     strcasecmp (argv [i], "-ratio") == 0)
 	    {
 		if (i + 1 < argc)
-		    Settings.SPCTo65c816Ratio = atoi (argv [++i]);
+		{
+		}
 		else
 		    S9xUsage ();
 	    }
@@ -216,7 +210,7 @@ char *S9xParseArgs (char **argv, int argc)
 		{
 		    int p = atoi (argv [++i]);
 		    if (p > 0 && p < 200)
-			Settings.H_Max = (SNES_CYCLES_PER_SCANLINE * p) / 100;
+			Settings.CyclesPercentage = p;
 		}
 		else
 		    S9xUsage ();
@@ -234,12 +228,12 @@ char *S9xParseArgs (char **argv, int argc)
 	    else if (strcasecmp (argv [i], "-n") == 0 ||
 		     strcasecmp (argv [i], "-nospeedhacks") == 0)
 	    {
-		Settings.Shutdown = FALSE;
+		Settings.ShutdownMaster = FALSE;
 	    }
 	    else if (strcasecmp (argv [i], "-sh") == 0 ||
 		     strcasecmp (argv [i], "-speedhacks") == 0)
 	    {
-		Settings.Shutdown = TRUE;
+		Settings.ShutdownMaster = TRUE;
 	    }
 	    else if (strcasecmp (argv [i], "-p") == 0 ||
 		     strcasecmp (argv [i], "-pal") == 0)
@@ -278,6 +272,10 @@ char *S9xParseArgs (char **argv, int argc)
 	    {
 		Settings.ForceNoHeader = TRUE;
 	    }
+	    else if (strcasecmp (argv [i], "-bs") == 0)
+	    {
+		Settings.BS = TRUE;
+	    }
 #ifdef DEBUGGER
 	    else if (strcasecmp (argv [i], "-d") == 0 ||
 		     strcasecmp (argv [i], "-debug") == 0)
@@ -356,6 +354,10 @@ char *S9xParseArgs (char **argv, int argc)
 	    {
 		Settings.DisableGraphicWindows = FALSE;
 	    }
+	    else if (strcasecmp (argv [i], "-im7") == 0)
+	    {
+		Settings.Mode7Interpolate = TRUE;
+	    }
 	    else if (strcasecmp (argv [i], "-gg") == 0 ||
 		     strcasecmp (argv [i], "-gamegenie") == 0)
 	    {
@@ -365,7 +367,7 @@ char *S9xParseArgs (char **argv, int argc)
 		    uint8 byte;
 		    const char *error;
 		    if ((error = S9xGameGenieToRaw (argv [++i], address, byte)) == NULL)
-			S9xAddCheat (address, TRUE, FALSE, 1, byte);
+			S9xAddCheat (TRUE, FALSE, address, byte);
 		    else
 			S9xMessage (S9X_ERROR, S9X_GAME_GENIE_CODE_ERROR,
 				    error);
@@ -382,7 +384,7 @@ char *S9xParseArgs (char **argv, int argc)
 		    uint8 byte;
 		    const char *error;
 		    if ((error = S9xProActionReplayToRaw (argv [++i], address, byte)) == NULL)
-			S9xAddCheat (address, TRUE, FALSE, 1, byte);
+			S9xAddCheat (TRUE, FALSE, address, byte);
 		    else
 			S9xMessage (S9X_ERROR, S9X_ACTION_REPLY_CODE_ERROR,
 				    error);
@@ -401,9 +403,11 @@ char *S9xParseArgs (char **argv, int argc)
 		    uint8 num_bytes;
 		    const char *error;
 		    if ((error = S9xGoldFingerToRaw (argv [++i], address, sram,
-						  num_bytes, bytes)) == NULL)
-			S9xAddCheat (address, FALSE, sram, num_bytes, bytes [0],
-				  bytes [1], bytes [2]);
+						     num_bytes, bytes)) == NULL)
+		    {
+			for (int c = 0; c < num_bytes; c++)
+			    S9xAddCheat (TRUE, FALSE, address + c, bytes [c]);
+		    }
 		    else
 			S9xMessage (S9X_ERROR, S9X_GOLD_FINGER_CODE_ERROR,
 				    error);
@@ -419,8 +423,14 @@ char *S9xParseArgs (char **argv, int argc)
 		    double ft;
 		    if (sscanf (argv [++i], "%lf", &ft) == 1)
 		    {
+#ifdef __WIN32__
 			Settings.FrameTimePAL = (int32) (ft * 1000);
 			Settings.FrameTimeNTSC = (int32) (ft * 1000);
+#else
+			Settings.FrameTimePAL = (int32) ft;
+			Settings.FrameTimeNTSC = (int32) ft;
+#endif
+
 		    }
 		}
 		else
@@ -446,11 +456,12 @@ char *S9xParseArgs (char **argv, int argc)
 	    {
 		Settings.Stereo = TRUE;
 		Settings.APUEnabled = TRUE;
+		Settings.NextAPUEnabled = TRUE;
 	    }
 	    else if (strcasecmp (argv [i], "-mono") == 0)
 	    {
 		Settings.Stereo = FALSE;
-		Settings.APUEnabled = TRUE;
+		Settings.NextAPUEnabled = TRUE;
 	    }
 	    else if (strcasecmp (argv [i], "-envx") == 0 ||
 		     strcasecmp (argv [i], "-ex") == 0)
@@ -494,6 +505,15 @@ char *S9xParseArgs (char **argv, int argc)
 		Settings.ThreadSound = TRUE;
 	    }
 #endif
+	    else if (strcasecmp (argv [i], "-alt") == 0 ||
+		     strcasecmp (argv [i], "-altsampledecode") == 0)
+	    {
+		Settings.AltSampleDecode = 1;
+	    }
+	    else if (strcasecmp (argv [i], "-fix") == 0)
+	    {
+		Settings.FixFrequency = 1;
+	    }
 	    else if (strcasecmp (argv [i], "-nosuperfx") == 0 ||
 		     strcasecmp (argv [i], "-nosfx") == 0)
 		Settings.ForceNoSuperFX = TRUE;
@@ -618,14 +638,14 @@ void S9xParseCheatsFile (const char *rom_filename)
 	    if (((error = S9xGameGenieToRaw (buf, address, byte)) == NULL) ||
 		((error = S9xProActionReplayToRaw (buf, address, byte)) == NULL))
 	    {
-		S9xAddCheat (address, TRUE, FALSE, 1, byte);
+		S9xAddCheat (TRUE, FALSE, address, byte);
 	    }
 	    else
 	    if ((error = S9xGoldFingerToRaw (buf, address, sram,
 					     num_bytes, bytes)) == NULL)
 	    {
-		S9xAddCheat (address, FALSE, sram, num_bytes, bytes [0],
-			     bytes [1], bytes [2]);
+		for (int c = 0; c < num_bytes; c++)
+		    S9xAddCheat (TRUE, FALSE, address + c, bytes [c]);
 	    }
 	    else
 		S9xMessage (S9X_ERROR, S9X_GAME_GENIE_CODE_ERROR, error);
