@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,17 +38,16 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _SNES9X_H_
 #define _SNES9X_H_
 
-#define VERSION "1.26"
+#define VERSION "1.39"
 
 #include <stdio.h>
 #include <stdlib.h>
 
-#ifdef _WIN32
-#include "..\Snes9X.h"
+#ifdef __WIN32__
+#include "..\wsnes9x.h"
 #include "..\zlib\zlib.h"
 #endif
 
@@ -60,7 +62,7 @@
 #define ROM_NAME_LEN 23
 
 #ifdef ZLIB
-#ifndef _WIN32
+#ifndef __WIN32__
 #include "zlib.h"
 #endif
 #define STREAM gzFile
@@ -86,20 +88,33 @@
 
 #define SNES_MAX_NTSC_VCOUNTER  262
 #define SNES_MAX_PAL_VCOUNTER   312
-#define SNES_HCOUNTER_MAX	341
+#define SNES_HCOUNTER_MAX	342
 #define SPC700_TO_65C816_RATIO	2
 #define AUTO_FRAMERATE		200
 
-#define SNES_SCANLINE_TIME (63.49e-6)
+// NTSC master clock signal 21.47727MHz
+// PPU: master clock / 4
+// 1 / PPU clock * 342 -> 63.695us
+// 63.695us / (1 / 3.579545MHz) -> 228 cycles per scanline
+// From Earth Worm Jim: APU executes an average of 65.14285714 cycles per
+// scanline giving an APU clock speed of 1.022731096MHz
+
+// PAL master clock signal 21.28137MHz
+// PPU: master clock / 4
+// 1 / PPU clock * 342 -> 64.281us
+// 64.281us / (1 / 3.546895MHz) -> 228 cycles per scanline.
+
+#define SNES_SCANLINE_TIME (63.695e-6)
+#define SNES_CLOCK_SPEED (3579545)
+
+#define SNES_CLOCK_LEN (1.0 / SNES_CLOCK_SPEED)
 
 #ifdef VAR_CYCLES
-#define SNES_CYCLES_PER_SCANLINE ((uint16) (63.5e-6 / (1 / 3580000.0)) * 6)
+#define SNES_CYCLES_PER_SCANLINE ((uint32) ((SNES_SCANLINE_TIME / SNES_CLOCK_LEN) * 6 + 0.5))
 #else
-#define SNES_CYCLES_PER_SCANLINE ((uint16) (63.5e-6 / (1 / 3580000.0)))
+#define SNES_CYCLES_PER_SCANLINE ((uint32) (SNES_SCANLINE_TIME / SNES_CLOCK_LEN + 0.5))
 #endif
 
-#define SCANLINE_FREQUENCY (15748.0315)
-
 #define SNES_TR_MASK	    (1 << 4)
 #define SNES_TL_MASK	    (1 << 5)
 #define SNES_X_MASK	    (1 << 6)
@@ -137,9 +152,11 @@ enum {
 
 #ifdef VAR_CYCLES
 #define ONE_CYCLE 6
+#define SLOW_ONE_CYCLE 8
 #define TWO_CYCLES 12
 #else
 #define ONE_CYCLE 1
+#define SLOW_ONE_CYCLE 1
 #define TWO_CYCLES 2
 #endif
 
@@ -162,12 +179,20 @@ struct SCPUState{
     long   MemSpeed;
     long   MemSpeedx2;
     long   FastROMSpeed;
+    uint32 AutoSaveTimer;
+    bool8  SRAMModified;
+    uint32 NMITriggerPoint;
+    bool8  BRKTriggered;
+    bool8  TriedInterleavedMode2;
+    uint32 NMICycleCount;
+    uint32 IRQCycleCount;
 };
 
 #define HBLANK_START_EVENT 0
 #define HBLANK_END_EVENT 1
-#define HTIMER_EVENT 2
-#define NO_EVENT 3
+#define HTIMER_BEFORE_EVENT 2
+#define HTIMER_AFTER_EVENT 3
+#define NO_EVENT 4
 
 struct SSettings{
     // CPU options
@@ -176,10 +201,11 @@ struct SSettings{
     uint8  SoundSkipMethod;
     long   H_Max;
     long   HBlankStart;
-    uint32 SPCTo65c816Ratio;
+    long   CyclesPercentage;
     bool8  DisableIRQ;
-    uint8  Paused;
-    bool8  BattleToadsNMIHack;
+    bool8  Paused;
+    bool8  ForcedPause;
+    bool8  StopEmulation;
 
     // Tracing options
     bool8  TraceDMA;
@@ -217,23 +243,34 @@ struct SSettings{
     bool8  ForceNoDSP1;
     bool8  ForceSA1;
     bool8  ForceNoSA1;
+    bool8  ForceC4;
+    bool8  ForceNoC4;
+    bool8  ForceSDD1;
+    bool8  ForceNoSDD1;
     bool8  MultiPlayer5;
     bool8  Mouse;
     bool8  SuperScope;
+    bool8  SRTC;
     uint32 ControllerOption;
     
+    bool8  ShutdownMaster;
     bool8  MultiPlayer5Master;
     bool8  SuperScopeMaster;
     bool8  MouseMaster;
     bool8  SuperFX;
     bool8  DSP1Master;
     bool8  SA1;
+    bool8  C4;
+    bool8  SDD1;
 
     // Sound options
+    uint32 SoundPlaybackRate;
     bool8  TraceSoundDSP;
-    uint8  SoundPlaybackRate;
     bool8  Stereo;
+    bool8  ReverseStereo;
+    bool8  SixteenBitSound;
     int    SoundBufferSize;
+    int    SoundMixInterval;
     bool8  SoundEnvelopeHeightReading;
     bool8  DisableSoundEcho;
     bool8  DisableSampleCaching;
@@ -241,11 +278,16 @@ struct SSettings{
     bool8  SoundSync;
     bool8  InterpolatedSound;
     bool8  ThreadSound;
+    bool8  Mute;
+    bool8  NextAPUEnabled;
+    uint8  AltSampleDecode;
+    bool8  FixFrequency;
     
     // Graphics options
     bool8  SixteenBit;
     bool8  Transparency;
     bool8  SupportHiRes;
+    bool8  Mode7Interpolate;
 
     // SNES graphics options
     bool8  BGLayering;
@@ -257,24 +299,68 @@ struct SSettings{
 
     // Others
     bool8  NetPlay;
+    bool8  NetPlayServer;
     char   ServerName [128];
     int    Port;
-#ifdef USE_GLIDE
     bool8  GlideEnable;
-#endif
-    bool8  DaffyDuckNMIHack;
+    bool8  OpenGLEnable;
+    int32  AutoSaveDelay; // Time in seconds before S-RAM auto-saved if modified.
+    bool8  ApplyCheats;
+    bool8  TurboMode;
+    uint32 TurboSkipFrames;
+    uint32 AutoMaxSkipFrames;
+    
+// Fixes for individual games
+    uint32 StrikeGunnerOffsetHack;
+    bool8  ChuckRock;
     bool8  StarfoxHack;
     bool8  WinterGold;
-    bool8  Seiken3;
+    bool8  Dezaemon;
+    bool8  WrestlemaniaArcade;
+    bool8  BS;	// Japanese Satellite System games.
+    bool8  DaffyDuck;
+    uint8  APURAMInitialValue;
+    
+#ifdef __WIN32__
+    int    SoundDriver;
+#endif
+};
+
+struct SSNESGameFixes
+{
+    uint8 NeedInit0x2137;
+    uint8 umiharakawaseFix;
+    uint8 alienVSpredetorFix;
+    uint8 APU_OutPorts_ReturnValueFix;
+    uint8 Old_Read0x4200;
+    uint8 _0x213E_ReturnValue;
+    uint8 TouhaidenControllerFix;
+    uint8 SoundEnvelopeHeightReading2;
+    uint8 SRAMInitialValue;
 };
 
 START_EXTERN_C
 extern struct SSettings Settings;
 extern struct SCPUState CPU;
+extern struct SSNESGameFixes SNESGameFixes;
 extern char String [513];
 
 void S9xExit ();
 void S9xMessage (int type, int number, const char *message);
+void S9xLoadSDD1Data ();
 END_EXTERN_C
 
+enum {
+    PAUSE_NETPLAY_CONNECT = (1 << 0),
+    PAUSE_TOGGLE_FULL_SCREEN = (1 << 1),
+    PAUSE_EXIT = (1 << 2),
+    PAUSE_MENU = (1 << 3),
+    PAUSE_INACTIVE_WINDOW = (1 << 4),
+    PAUSE_WINDOW_ICONISED = (1 << 5),
+    PAUSE_RESTORE_GUI = (1 << 6),
+    PAUSE_FREEZE_FILE = (1 << 7)
+};
+void S9xSetPause (uint32 mask);
+void S9xClearPause (uint32 mask);
+
 #endif
