@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifdef __DJGPP__
 #include <allegro.h>
 #undef TRUE
@@ -54,6 +56,13 @@ else \
 if ((v) > 32767) \
     (v) = 32767
 
+#define CLIP16_latch(v,l) \
+if ((v) < -32768) \
+{ (v) = -32768; (l)++; }\
+else \
+if ((v) > 32767) \
+{ (v) = 32767; (l)++; }
+
 #define CLIP24(v) \
 if ((v) < -8388608) \
     (v) = -8388608; \
@@ -86,12 +95,6 @@ extern int Loop [16];
 extern long FilterValues[4][2];
 extern int NoiseFreq [32];
 
-static char* envelope [] = 
-{
-    "silent", "attack", "decay", "sustain", "release", "gain",
-    "inc_lin", "inc_bent", "dec_lin", "dec_exp"
-};
-
 #undef ABS
 #define ABS(a) ((a) < 0 ? -(a) : (a))
 
@@ -103,9 +106,13 @@ static char* envelope [] =
 #define VOL_DIV16 0x0080
 #define ENVX_SHIFT 24
 
+extern "C" void DecodeBlockAsm (int8 *, int16 *, int32 *, int32 *);
+extern "C" void DecodeBlockAsm2 (int8 *, int16 *, int32 *, int32 *);
+
 // F is channel's current frequency and M is the 16-bit modulation waveform
 // from the previous channel multiplied by the current envelope volume level.
-#define PITCH_MOD(F,M) ((F) * ((((M) & 0x7fffff) >> 16) + 1) >> 7)
+#define PITCH_MOD(F,M) ((F) * ((((unsigned long) (M)) + 0x800000) >> 16) >> 7)
+//#define PITCH_MOD(F,M) ((F) * ((((M) & 0x7fffff) >> 14) + 1) >> 8)
 
 #define LAST_SAMPLE 0xffffff
 #define JUST_PLAYED_LAST_SAMPLE(c) ((c)->sample_pointer >= LAST_SAMPLE)
@@ -141,7 +148,7 @@ void S9xSetEnvRate (Channel *ch, unsigned long rate, int direction, int target)
 {
     ch->envx_target = target;
 
-    if (rate == ~0)
+    if (rate == ~0UL)
     {
 	ch->direction = 0;
 	rate = 0;
@@ -205,8 +212,8 @@ void S9xSetMasterVolume (short volume_left, short volume_right)
 	    volume_left = (ABS (volume_right) + ABS (volume_left)) / 2;
 	SoundData.master_volume_left = volume_left;
 	SoundData.master_volume_right = volume_right;
-	SoundData.master_volume [0] = volume_left;
-	SoundData.master_volume [1] = volume_right;
+	SoundData.master_volume [Settings.ReverseStereo] = volume_left;
+	SoundData.master_volume [1 ^ Settings.ReverseStereo] = volume_right;
     }
 }
 
@@ -216,8 +223,8 @@ void S9xSetEchoVolume (short volume_left, short volume_right)
 	volume_left = (ABS (volume_right) + ABS (volume_left)) / 2;
     SoundData.echo_volume_left = volume_left;
     SoundData.echo_volume_right = volume_right;
-    SoundData.echo_volume [0] = volume_left;
-    SoundData.echo_volume [1] = volume_right;
+    SoundData.echo_volume [Settings.ReverseStereo] = volume_left;
+    SoundData.echo_volume [1 ^ Settings.ReverseStereo] = volume_right;
 }
 
 void S9xSetEchoEnable (uint8 byte)
@@ -304,11 +311,13 @@ void S9xFixSoundAfterSnapshotLoad ()
 	SoundData.channels [i].envxx = SoundData.channels [i].envx << ENVX_SHIFT;
 	SoundData.channels [i].next_sample = 0;
 	SoundData.channels [i].interpolate = 0;
+	SoundData.channels [i].previous [0] = (int32) SoundData.channels [i].previous16 [0];
+	SoundData.channels [i].previous [1] = (int32) SoundData.channels [i].previous16 [1];
     }
-    SoundData.master_volume [0] = SoundData.master_volume_left;
-    SoundData.master_volume [1] = SoundData.master_volume_right;
-    SoundData.echo_volume [0] = SoundData.echo_volume_left;
-    SoundData.echo_volume [1] = SoundData.echo_volume_right;
+    SoundData.master_volume [Settings.ReverseStereo] = SoundData.master_volume_left;
+    SoundData.master_volume [1 ^ Settings.ReverseStereo] = SoundData.master_volume_right;
+    SoundData.echo_volume [Settings.ReverseStereo] = SoundData.echo_volume_left;
+    SoundData.echo_volume [1 ^ Settings.ReverseStereo] = SoundData.echo_volume_right;
     IAPU.Scanline = 0;
 }
 
@@ -368,20 +377,27 @@ void S9xSetEnvelopeHeight (int channel, int level)
 
 int S9xGetEnvelopeHeight (int channel)
 {
-    if (Settings.SoundEnvelopeHeightReading &&
-	SoundData.channels[channel].state != SOUND_SILENT &&
-	SoundData.channels[channel].state != SOUND_GAIN)
+    if ((Settings.SoundEnvelopeHeightReading ||
+	 SNESGameFixes.SoundEnvelopeHeightReading2) &&
+        SoundData.channels[channel].state != SOUND_SILENT &&
+        SoundData.channels[channel].state != SOUND_GAIN)
     {
-	return (SoundData.channels[channel].envx);
+        return (SoundData.channels[channel].envx);
     }
+
+    //siren fix from XPP
+    if (SNESGameFixes.SoundEnvelopeHeightReading2 &&
+        SoundData.channels[channel].state != SOUND_SILENT)
+    {
+        return (SoundData.channels[channel].envx);
+    }
+
     return (0);
 }
 
 #if 1
-void S9xSetSoundSample (int channel, uint16 sample_number) 
+void S9xSetSoundSample (int, uint16) 
 {
-    Channel *ch = &SoundData.channels[channel];
-//    ch->sample_number = sample_number;
 }
 #else
 void S9xSetSoundSample (int channel, uint16 sample_number)
@@ -414,6 +430,11 @@ void S9xSetSoundFrequency (int channel, int hertz)
 	    hertz = NoiseFreq [APU.DSP [APU_FLG] & 0x1f];
 	SoundData.channels[channel].frequency = (int)
 	    (((int64) hertz * FIXED_POINT) / so.playback_rate);
+	if (Settings.FixFrequency)
+	{
+	    SoundData.channels[channel].frequency = 
+		(unsigned long) ((double)  SoundData.channels[channel].frequency * 0.980);
+	}
     }
 }
 
@@ -435,14 +456,283 @@ bool8 S9xSetSoundMute (bool8 mute)
     return (old);
 }
 
-void DecodeBlock (Channel *ch)
+void AltDecodeBlock (Channel *ch)
 {
-    long out;
+    if (ch->block_pointer >= 0x10000 - 9)
+    {
+	ch->last_block = TRUE;
+	ch->loop = FALSE;
+	ch->block = ch->decoded;
+	memset ((void *) ch->decoded, 0, sizeof (int16) * 16);
+	return;
+    }
+    signed char *compressed = (signed char *) &IAPU.RAM [ch->block_pointer];
+
+    unsigned char filter = *compressed;
+    if ((ch->last_block = filter & 1))
+	ch->loop = (filter & 2) != 0;
+
+#if (defined (__i386__) || defined (__i486__) ||\
+     defined (__i586__) || defined (__WIN32__) || defined (__DJGPP))
+    int16 *raw = ch->block = ch->decoded;
+
+    if (Settings.AltSampleDecode == 1)
+	DecodeBlockAsm (compressed, raw, &ch->previous [0], &ch->previous [1]);
+    else
+    	DecodeBlockAsm2 (compressed, raw, &ch->previous [0], &ch->previous [1]);
+#else
+    int32 out;
+    unsigned char shift;
+    signed char sample1, sample2;
+    unsigned int i;
+
+    compressed++;
+    signed short *raw = ch->block = ch->decoded;
+    
+    int32 prev0 = ch->previous [0];
+    int32 prev1 = ch->previous [1];
+    shift = filter >> 4;
+
+    switch ((filter >> 2) & 3)
+    {
+    case 0:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    sample2 >>= 4;
+	    sample1 >>= 4;
+	    *raw++ = ((int32) sample1 << shift);
+	    *raw++ = ((int32) sample2 << shift);
+	}
+	prev1 = *(raw - 2);
+	prev0 = *(raw - 1);
+	break;
+    case 1:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    sample2 >>= 4;
+	    sample1 >>= 4;
+	    prev0 = (int16) prev0;
+	    *raw++ = prev1 = ((int32) sample1 << shift) + prev0 - (prev0 >> 4);
+	    prev1 = (int16) prev1;
+	    *raw++ = prev0 = ((int32) sample2 << shift) + prev1 - (prev1 >> 4);
+	}
+	break;
+    case 2:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    sample2 >>= 4;
+	    sample1 >>= 4;
+	    
+	    out = (sample1 << shift) - prev1 + (prev1 >> 4);
+	    prev1 = (int16) prev0;
+	    prev0 &= ~3;
+	    *raw++ = prev0 = out + (prev0 << 1) - (prev0 >> 5) - 
+			     (prev0 >> 4);
+
+	    out = (sample2 << shift) - prev1 + (prev1 >> 4);
+	    prev1 = (int16) prev0;
+	    prev0 &= ~3;
+	    *raw++ = prev0 = out + (prev0 << 1) - (prev0 >> 5) -
+			     (prev0 >> 4);
+	}
+	break;
+    case 3:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    sample2 >>= 4;
+	    sample1 >>= 4;
+	    out = (sample1 << shift);
+
+	    out = out - prev1 + (prev1 >> 3) + (prev1 >> 4);
+	    prev1 = (int16) prev0;
+	    prev0 &= ~3;
+	    *raw++ = prev0 = out + (prev0 << 1) - (prev0 >> 3) - 
+			     (prev0 >> 4) - (prev1 >> 6);
+
+	    out = (sample2 << shift);
+	    out = out - prev1 + (prev1 >> 3) + (prev1 >> 4);
+	    prev1 = (int16) prev0;
+	    prev0 &= ~3;
+	    *raw++ = prev0 = out + (prev0 << 1) - (prev0 >> 3) - 
+			     (prev0 >> 4) - (prev1 >> 6);
+	}
+	break;
+    }
+    ch->previous [0] = prev0;
+    ch->previous [1] = prev1;
+#endif
+
+    ch->block_pointer += 9;
+}
+
+void AltDecodeBlock2 (Channel *ch)
+{
+    int32 out;
     unsigned char filter;
     unsigned char shift;
     signed char sample1, sample2;
     unsigned char i;
 
+    if (ch->block_pointer > 0x10000 - 9)
+    {
+	ch->last_block = TRUE;
+	ch->loop = FALSE;
+	ch->block = ch->decoded;
+	memset ((void *) ch->decoded, 0, sizeof (int16) * 16);
+	return;
+    }
+
+    signed char *compressed = (signed char *) &IAPU.RAM [ch->block_pointer];
+
+    filter = *compressed;
+    if ((ch->last_block = filter & 1))
+    ch->loop = (filter & 2) != 0;
+
+    compressed++;
+    signed short *raw = ch->block = ch->decoded;
+    
+    shift = filter >> 4;
+    int32 prev0 = ch->previous [0];
+    int32 prev1 = ch->previous [1];
+
+    if(shift > 12)
+	shift -= 4;
+
+    switch ((filter >> 2) & 3)
+    {
+    case 0:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    //Sample 2 = Bottom Nibble, Sign Extended.
+	    sample2 >>= 4;
+	    //Sample 1 = Top Nibble, shifted down and Sign Extended.
+	    sample1 >>= 4;
+
+	    out = (int32)(sample1 << shift);
+
+	    prev1 = prev0;
+	    prev0 = out;
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+
+	    out = (int32)(sample2 << shift);
+
+	    prev1 = prev0;
+	    prev0 = out;
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+	}
+	break;
+    case 1:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    //Sample 2 = Bottom Nibble, Sign Extended.
+	    sample2 >>= 4;
+	    //Sample 1 = Top Nibble, shifted down and Sign Extended.
+	    sample1 >>= 4;
+	    out = (int32)(sample1 << shift);
+	    out += (int32)((double)prev0 * 15/16);
+
+	    prev1 = prev0;
+	    prev0 = out;
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+
+	    out = (int32)(sample2 << shift);
+	    out += (int32)((double)prev0 * 15/16);
+
+	    prev1 = prev0;
+	    prev0 = out;
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+	}
+	break;
+    case 2:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    //Sample 2 = Bottom Nibble, Sign Extended.
+	    sample2 >>= 4;
+	    //Sample 1 = Top Nibble, shifted down and Sign Extended.
+	    sample1 >>= 4;
+
+	    out = ((sample1 << shift) * 256 + (prev0 & ~0x2) * 488 - prev1 * 240) >> 8;
+
+	    prev1 = prev0;
+	    prev0 = (int16)out;
+	    *raw++ = (int16)out;
+
+	    out = ((sample2 << shift) * 256 + (prev0 & ~0x2) * 488 - prev1 * 240) >> 8;
+
+	    prev1 = prev0;
+	    prev0 = (int16)out;
+	    *raw++ = (int16)out;
+	}
+	break;
+
+    case 3:
+	for (i = 8; i != 0; i--)
+	{
+	    sample1 = *compressed++;
+	    sample2 = sample1 << 4;
+	    //Sample 2 = Bottom Nibble, Sign Extended.
+	    sample2 >>= 4;
+	    //Sample 1 = Top Nibble, shifted down and Sign Extended.
+	    sample1 >>= 4;
+	    out = (int32)(sample1 << shift);
+	    out += (int32)((double)prev0 * 115/64 - (double)prev1 * 13/16);
+
+	    prev1 = prev0;
+	    prev0 = out;
+
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+
+	    out = (int32)(sample2 << shift);
+	    out += (int32)((double)prev0 * 115/64 - (double)prev1 * 13/16);
+
+	    prev1 = prev0;
+	    prev0 = out;
+
+	    CLIP16(out);
+	    *raw++ = (int16)out;
+	}
+	break;
+    }
+    ch->previous [0] = prev0;
+    ch->previous [1] = prev1;
+    ch->block_pointer += 9;
+}
+
+void DecodeBlock (Channel *ch)
+{
+    int32 out;
+    unsigned char filter;
+    unsigned char shift;
+    signed char sample1, sample2;
+    unsigned char i;
+
+    if (Settings.AltSampleDecode)
+    {
+	if (Settings.AltSampleDecode < 3)
+	    AltDecodeBlock (ch);
+	else
+	    AltDecodeBlock2 (ch);
+        return;
+    }
     if (ch->block_pointer > 0x10000 - 9)
     {
 	ch->last_block = TRUE;
@@ -453,7 +743,7 @@ void DecodeBlock (Channel *ch)
     signed char *compressed = (signed char *) &IAPU.RAM [ch->block_pointer];
 
     filter = *compressed;
-    if (ch->last_block = filter & 1)
+    if ((ch->last_block = filter & 1))
 	ch->loop = (filter & 2) != 0;
 
     // If enabled, results in 'tick' sound on some samples that repeat by
@@ -475,9 +765,10 @@ void DecodeBlock (Channel *ch)
 
 	shift = filter >> 4;
 	filter = ((filter >> 2) & 3);
-	long prev0 = ch->previous [0];
-	long prev1 = ch->previous [1];
-	long *f = (long *) &FilterValues[filter];
+	int32 prev0 = ch->previous [0];
+	int32 prev1 = ch->previous [1];
+	int32 f0 = FilterValues[filter][0];
+	int32 f1 = FilterValues[filter][1];
 
 	for (i = 8; i != 0; i--)
 	{
@@ -487,40 +778,41 @@ void DecodeBlock (Channel *ch)
 	    sample2 >>= 4;
 	    //Sample 1 = Top Nibble, shifted down and Sign Extended.
 	    sample1 >>= 4;
-	    out = (sample1 << shift) + 
-		   (((prev0 * f[0]) + (prev1 * f[1])) / 256);
+	    out = (sample1 << shift);
+	    out += (prev0 * f0 + prev1 * f1) / 256;
 
-	    CLIP16(out);
+	    CLIP16 (out);
 	    prev1 = prev0;
 	    prev0 = out;
 	    *raw++ = (signed short) out;
 
-	    out = (sample2 << shift) +
-		   (((prev0 * f[0]) + (prev1 * f[1])) / 256);
+	    out = (sample2 << shift);
+	    out += (prev0 * f0 + prev1 * f1) / 256;
 
-	    CLIP16(out);
+	    CLIP16 (out);
 	    prev1 = prev0;
 	    prev0 = out;
 	    *raw++ = (signed short) out;
 	}
-	ch->previous [0] = (short) prev0;
-	ch->previous [1] = (short) prev1;
+	ch->previous [0] = prev0;
+	ch->previous [1] = prev1;
+
 	if (!Settings.DisableSampleCaching)
+	{
 	    memcpy (IAPU.CachedSamples + (ch->block_pointer << 2),
 		    (uint8 *) ch->decoded, 32);
+	}
     }
-
-    if ((ch->block_pointer += 9) >= 0x10000 - 9)
-    {
-	ch->last_block = TRUE;
-	ch->loop = FALSE;
-	ch->block_pointer -= 0x10000 - 9;
-    }
+    ch->block_pointer += 9;
 }
 
 void MixStereo (int sample_count)
 {
+#if defined(TARGET_OS_MAC) && TARGET_OS_MAC
+    static int wave[SOUND_BUFFER_SIZE];
+#else
     int wave[SOUND_BUFFER_SIZE];
+#endif
     int pitch_mod = SoundData.pitch_mod & ~APU.DSP[APU_NON];
 
     for (uint32 J = 0; J < NUM_CHANNELS; J++) 
@@ -532,6 +824,8 @@ void MixStereo (int sample_count)
 	if (ch->state == SOUND_SILENT || !(so.sound_switch & (1 << J)))
 	    continue;
 
+//	freq0 = (unsigned long) ((double) freq0 * 0.985);
+
 	bool8 mod = pitch_mod & (1 << J);
 
 	if (ch->needs_decode) 
@@ -757,7 +1051,19 @@ void MixStereo (int sample_count)
 		else
 		    ch->next_sample = ch->block [ch->sample_pointer];
 
-		if (ch->type == SOUND_NOISE) 
+		if (ch->type == SOUND_SAMPLE)
+		{
+		    if (Settings.InterpolatedSound && freq < FIXED_POINT && !mod)
+		    {
+			ch->interpolate = ((ch->next_sample - ch->sample) * 
+					   (long) freq) / (long) FIXED_POINT;
+			ch->sample = (int16) (ch->sample + (((ch->next_sample - ch->sample) * 
+					   (long) (ch->count)) / (long) FIXED_POINT));
+		    }		  
+		    else
+			ch->interpolate = 0;
+		}
+		else
 		{
 		    for (;VL > 0; VL--)
 			if ((so.noise_gen <<= 1) & 0x80000000L)
@@ -765,14 +1071,7 @@ void MixStereo (int sample_count)
 		    ch->sample = (so.noise_gen << 17) >> 17;
 		    ch->interpolate = 0;
 		}
-		else
-		{
-		    if (Settings.InterpolatedSound && freq < FIXED_POINT && !mod)
-			ch->interpolate = ((ch->next_sample - ch->sample) * 
-					   (long) freq) / (long) FIXED_POINT;
-		    else
-			ch->interpolate = 0;
-		}
+
 		VL = (ch->sample * ch-> left_vol_level) / 128;
 		VR = (ch->sample * ch->right_vol_level) / 128;
             }
@@ -783,19 +1082,19 @@ void MixStereo (int sample_count)
 		    int32 s = (int32) ch->sample + ch->interpolate;
 		    
 		    CLIP16(s);
-		    ch->sample = s;
+		    ch->sample = (int16) s;
 		    VL = (ch->sample * ch-> left_vol_level) / 128;
 		    VR = (ch->sample * ch->right_vol_level) / 128;
 		}
 	    }
 
-	    MixBuffer [I  ] += VL;
-	    MixBuffer [I+1] += VR;
-	    ch->echo_buf_ptr [I  ] += VL;
-	    ch->echo_buf_ptr [I+1] += VR;
-
 	    if (pitch_mod & (1 << (J + 1)))
 		wave [I / 2] = ch->sample * ch->envx;
+
+	    MixBuffer [I      ^ Settings.ReverseStereo] += VL;
+	    MixBuffer [I + (1 ^ Settings.ReverseStereo)] += VR;
+	    ch->echo_buf_ptr [I      ^ Settings.ReverseStereo] += VL;
+	    ch->echo_buf_ptr [I + (1 ^ Settings.ReverseStereo)] += VR;
         }
 stereo_exit: ;
     }
@@ -807,7 +1106,11 @@ END_OF_FUNCTION(MixStereo);
 
 void MixMono (int sample_count)
 {
+#if defined(TARGET_OS_MAC) && TARGET_OS_MAC
+    static int wave[SOUND_BUFFER_SIZE];
+#else
     int wave[SOUND_BUFFER_SIZE];
+#endif
     int pitch_mod = SoundData.pitch_mod & (~APU.DSP[APU_NON]);
 
     for (uint32 J = 0; J < NUM_CHANNELS; J++) 
@@ -818,6 +1121,8 @@ void MixMono (int sample_count)
 	if (ch->state == SOUND_SILENT || !(so.sound_switch & (1 << J)))
 	    continue;
 
+//	freq0 = (unsigned long) ((double) freq0 * 0.985);
+
 	bool8 mod = pitch_mod & (1 << J);
 
 	if (ch->needs_decode) 
@@ -1039,7 +1344,19 @@ void MixMono (int sample_count)
 		else
 		    ch->next_sample = ch->block [ch->sample_pointer];
 
-		if (ch->type == SOUND_NOISE) 
+		if (ch->type == SOUND_SAMPLE)
+		{
+		    if (Settings.InterpolatedSound && freq < FIXED_POINT && !mod)
+		    {
+			ch->interpolate = ((ch->next_sample - ch->sample) * 
+					   (long) freq) / (long) FIXED_POINT;
+			ch->sample = (int16) (ch->sample + (((ch->next_sample - ch->sample) * 
+					   (long) (ch->count)) / (long) FIXED_POINT));
+		    }		  
+		    else
+			ch->interpolate = 0;
+		}
+		else
 		{
 		    for (;V > 0; V--)
 			if ((so.noise_gen <<= 1) & 0x80000000L)
@@ -1047,14 +1364,6 @@ void MixMono (int sample_count)
 		    ch->sample = (so.noise_gen << 17) >> 17;
 		    ch->interpolate = 0;
 		}
-		else
-		{
-		    if (Settings.InterpolatedSound && freq < FIXED_POINT && !mod)
-			ch->interpolate = ((ch->next_sample - ch->sample) * 
-					   (long) freq) / (long) FIXED_POINT;
-		    else
-			ch->interpolate = 0;
-		}
 		V = (ch->sample * ch-> left_vol_level) / 128;
             }
 	    else
@@ -1064,7 +1373,7 @@ void MixMono (int sample_count)
 		    int32 s = (int32) ch->sample + ch->interpolate;
 
 		    CLIP16(s);
-		    ch->sample = s;
+		    ch->sample = (int16) s;
 		    V = (ch->sample * ch-> left_vol_level) / 128;
 		}
 	    }
@@ -1453,6 +1762,30 @@ void S9xResetSound (bool8 full)
     so.samples_mixed_so_far = 0;
     so.play_position = 0;
     so.err_counter = 0;
+
+    if (full)
+    {
+	SoundData.master_volume_left = 0;
+	SoundData.master_volume_right = 0;
+	SoundData.echo_volume_left = 0;
+	SoundData.echo_volume_right = 0;
+	SoundData.echo_enable = 0;
+	SoundData.echo_write_enabled = 0;
+	SoundData.echo_channel_enable = 0;
+	SoundData.pitch_mod = 0;
+	SoundData.dummy[0] = 0;
+	SoundData.dummy[1] = 0;
+	SoundData.dummy[2] = 0;
+	SoundData.master_volume[0] = 0;
+	SoundData.master_volume[1] = 0;
+	SoundData.echo_volume[0] = 0;
+	SoundData.echo_volume[1] = 0;
+	SoundData.noise_hertz = 0;
+    }
+
+    SoundData.master_volume_left = 127;
+    SoundData.master_volume_right = 127;
+    SoundData.master_volume [0] = SoundData.master_volume [1] = 127;
     if (so.playback_rate)
 	so.err_rate = (uint32) (FIXED_POINT * SNES_SCANLINE_TIME / (1.0 / so.playback_rate));
     else
@@ -1463,7 +1796,7 @@ void S9xResetSound (bool8 full)
 void S9xSetPlaybackRate (uint32 playback_rate)
 {
     so.playback_rate = playback_rate;
-    so.err_rate = (uint32) (SNES_SCANLINE_TIME * FIXED_POINT / (1 / (double) so.playback_rate));
+    so.err_rate = (uint32) (SNES_SCANLINE_TIME * FIXED_POINT / (1.0 / (double) so.playback_rate));
     S9xSetEchoDelay (APU.DSP [APU_EDL] & 0xf);
     for (int i = 0; i < 8; i++)
 	S9xSetSoundFrequency (i, SoundData.channels [i].hertz);
@@ -1476,12 +1809,12 @@ bool8 S9xInitSound (int mode, bool8 stereo, int buffer_size)
 
     so.playback_rate = 0;
     so.buffer_size = 0;
-    so.stereo = FALSE;
-    so.sixteen_bit = FALSE;
+    so.stereo = stereo;
+    so.sixteen_bit = Settings.SixteenBitSound;
     so.encoded = FALSE;
     
     S9xResetSound (TRUE);
-    
+
     if (!(mode & 7))
 	return (1);
 
@@ -1499,10 +1832,6 @@ bool8 S9xInitSound (int mode, bool8 stereo, int buffer_size)
 bool8 S9xSetSoundMode (int channel, int mode)
 {
     Channel *ch = &SoundData.channels[channel];
-    static char *modes [] = 
-	{"none", "adsr", "decay", "sustain", "release", "gain", "inc_lin",
-	 "inc_bent", "dec_lin", "dec_exp"
-	};
 
     switch (mode)
     {
@@ -1570,6 +1899,11 @@ void S9xPlaySample (int channel)
 		    APU.DSP [APU_ADSR2 + (channel << 4)]);
 
     ch->sample_number = APU.DSP [APU_SRCN + channel * 0x10];
+    if (APU.DSP [APU_NON] & (1 << channel))
+	ch->type = SOUND_NOISE;
+    else
+	ch->type = SOUND_SAMPLE;
+
     S9xSetSoundFrequency (channel, ch->hertz);
     ch->loop = FALSE;
     ch->needs_decode = TRUE;
