@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,11 +38,10 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _SOUND_H_
 #define _SOUND_H_
 
-enum { SOUND_SAMPLE = 0, SOUND_NOISE };
+enum { SOUND_SAMPLE = 0, SOUND_NOISE, SOUND_EXTRA_NOISE, SOUND_MUTE };
 enum { SOUND_SILENT, SOUND_ATTACK, SOUND_DECAY, SOUND_SUSTAIN,
        SOUND_RELEASE, SOUND_GAIN, SOUND_INCREASE_LINEAR,
        SOUND_INCREASE_BENT_LINE, SOUND_DECREASE_LINEAR,
@@ -83,8 +85,8 @@ typedef struct {
 #ifdef __sgi
     ALport al_port;
 #endif /* __sgi */
-    uint32 samples_mixed_so_far;
-    uint32 play_position;
+    int32  samples_mixed_so_far;
+    int32  play_position;
     uint32 err_counter;
     uint32 err_rate;
 } SoundStatus;
@@ -96,9 +98,9 @@ typedef struct {
     int type;
     short volume_left;
     short volume_right;
-    int hertz;
-    int frequency;
-    int count;
+    uint32 hertz;
+    uint32 frequency;
+    uint32 count;
     bool8 loop;
     int envx;
     short left_vol_level;
@@ -114,7 +116,7 @@ typedef struct {
     unsigned long sustain_level;
     signed short sample;
     signed short decoded [16];
-    signed short previous [2];
+    signed short previous16 [2];
     signed short *block;
     uint16 sample_number;
     bool8 last_block;
@@ -126,6 +128,7 @@ typedef struct {
     int32 envxx;
     signed short next_sample;
     int32 interpolate;
+    int32 previous [2];
     // Just incase they are needed in the future, for snapshot compatibility.
     uint32 dummy [8];
 } Channel;
