@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 #include "spc700.h"
 #include "memmap.h"
@@ -109,7 +111,7 @@ void STOP (char *s)
 #endif
 
     sprintf (String, "Sound CPU in unknown state executing %s at %04X\n%s\n", s, IAPU.PC - IAPU.RAM, buffer);
-    S9xMessage (S9X_ERROR, S9X_APU_STOPED, String);
+    S9xMessage (S9X_ERROR, S9X_APU_STOPPED, String);
     APU.TimerEnabled[0] = APU.TimerEnabled[1] = APU.TimerEnabled[2] = FALSE;
     IAPU.APUExecuting = FALSE;
 
@@ -194,7 +196,7 @@ APUSetZN8 ((uint8) Int16);
 #else
 #define PushW(w)\
     *(IAPU.RAM + 0xff + APURegisters.S) = w;\
-    *(IAPU.RAM + 0x100 + APURegisters.S) = w >> 8;\
+    *(IAPU.RAM + 0x100 + APURegisters.S) = (w >> 8);\
     APURegisters.S -= 2;
 #define PopW(w)\
     APURegisters.S += 2; \
@@ -294,6 +296,7 @@ void ApuF1 () { TCALL (15) }
 void Apu3F () // CALL absolute
 {
     Absolute ();
+    // 0xB6f for Star Fox 2
     PushW (IAPU.PC + 3 - IAPU.RAM);
     IAPU.PC = IAPU.RAM + IAPU.Address;
 }
@@ -306,7 +309,7 @@ void Apu4F () // PCALL $XX
 }
 
 #define SET(b) \
-S9xAPUSetByteZ (S9xAPUGetByteZ (OP1 ) | (1 << (b)), OP1); \
+S9xAPUSetByteZ ((uint8) (S9xAPUGetByteZ (OP1 ) | (1 << (b))), OP1); \
 IAPU.PC += 2
 
 void Apu02 ()
@@ -350,7 +353,7 @@ void ApuE2 ()
 }
 
 #define CLR(b) \
-S9xAPUSetByteZ (S9xAPUGetByteZ (OP1) & ~(1 << (b)), OP1); \
+S9xAPUSetByteZ ((uint8) (S9xAPUGetByteZ (OP1) & ~(1 << (b))), OP1); \
 IAPU.PC += 2;
 
 void Apu12 ()
@@ -1877,7 +1880,6 @@ void Apu7F ()
     IAPU.PC = IAPU.RAM + APURegisters.PC;
 }
 
-// HERE
 void Apu84 ()
 {
 // ADC A,dp
@@ -2148,8 +2150,21 @@ void ApuAF ()
 void ApuBE ()
 {
 // DAS
-    // XXX:
-    IAPU.PC++;
+   // Implemented by Jonathan Gevaryahu (using the DAA instruction code)
+   W1 = APURegisters.YA.B.A & 0xf;
+   W2 = APURegisters.YA.B.A >> 4;
+   APUClearCarry ();
+   if (W1 > 9)
+   {
+       W1 -= 6;
+   }
+   if (W2 > 9)    // This should never happen....
+   {
+       W2 -= 6;
+       APUSetCarry ();
+   }
+   APURegisters.YA.B.A = W1 | (W2 << 4);
+   IAPU.PC++;
 }
 
 void ApuBF ()
@@ -2303,12 +2318,11 @@ void ApuDF ()
     APUClearCarry ();
     if (W1 > 9)
     {
-	W1 -= 10;
-	W2++;
+	W1 -= 6;
     }
     if (W2 > 9)
     {
-	W2 -= 10;
+	W2 -= 6;
 	APUSetCarry ();
     }
     APURegisters.YA.B.A = W1 | (W2 << 4);
