@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,10 +38,16 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _SPC700_H_
 #define _SPC700_H_
 
+#ifdef SPCTOOL
+#define NO_CHANNEL_STRUCT
+#include "spctool/dsp.h"
+#include "spctool/spc700.h"
+#include "spctool/soundmod.h"
+#endif
+
 #define Carry       1
 #define Zero        2
 #define Interrupt   4
@@ -95,10 +104,29 @@ struct SAPURegisters{
 EXTERN_C struct SAPURegisters APURegisters;
 
 // Needed by ILLUSION OF GAIA
-#define ONE_APU_CYCLE 14
+//#define ONE_APU_CYCLE 14
+#define ONE_APU_CYCLE 21
 
 // Needed by all games written by the software company called Human
-#define ONE_APU_CYCLE_HUMAN 17
+//#define ONE_APU_CYCLE_HUMAN 17
+#define ONE_APU_CYCLE_HUMAN 21
+
+// 1.953us := 1.024065.54MHz
+
+#ifdef SPCTOOL
+EXTERN_C int32 ESPC (int32);
+
+#define APU_EXECUTE() \
+{ \
+    int32 l = (CPU.Cycles - APU.Cycles) / 14; \
+    if (l > 0) \
+    { \
+        l -= _EmuSPC(l); \
+        APU.Cycles += l * 14; \
+    } \
+}
+
+#else
 
 #ifdef DEBUGGER
 #define APU_EXECUTE1() \
@@ -122,5 +150,6 @@ if (IAPU.APUExecuting) \
     while (APU.Cycles <= CPU.Cycles) \
 	APU_EXECUTE1(); \
 }
+#endif
 
 #endif
