@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include "snes9x.h"
 
 #include "memmap.h"
@@ -67,42 +69,42 @@ uint8 ConvertTile (uint8 *pCache, uint32 TileAddr)
 	    uint32 p2 = 0;
 	    register uint8 pix;
 
-	    if (pix = *(tp + 0))
+	    if ((pix = *(tp + 0)))
 	    {
 		p1 |= odd_high[0][pix >> 4];
 		p2 |= odd_low[0][pix & 0xf];
 	    }
-	    if (pix = *(tp + 1))
+	    if ((pix = *(tp + 1)))
 	    {
 		p1 |= even_high[0][pix >> 4];
 		p2 |= even_low[0][pix & 0xf];
 	    }
-	    if (pix = *(tp + 16))
+	    if ((pix = *(tp + 16)))
 	    {
 		p1 |= odd_high[1][pix >> 4];
 		p2 |= odd_low[1][pix & 0xf];
 	    }
-	    if (pix = *(tp + 17))
+	    if ((pix = *(tp + 17)))
 	    {
 		p1 |= even_high[1][pix >> 4];
 		p2 |= even_low[1][pix & 0xf];
 	    }
-	    if (pix = *(tp + 32))
+	    if ((pix = *(tp + 32)))
 	    {
 		p1 |= odd_high[2][pix >> 4];
 		p2 |= odd_low[2][pix & 0xf];
 	    }
-	    if (pix = *(tp + 33))
+	    if ((pix = *(tp + 33)))
 	    {
 		p1 |= even_high[2][pix >> 4];
 		p2 |= even_low[2][pix & 0xf];
 	    }
-	    if (pix = *(tp + 48))
+	    if ((pix = *(tp + 48)))
 	    {
 		p1 |= odd_high[3][pix >> 4];
 		p2 |= odd_low[3][pix & 0xf];
 	    }
-	    if (pix = *(tp + 49))
+	    if ((pix = *(tp + 49)))
 	    {
 		p1 |= even_high[3][pix >> 4];
 		p2 |= even_low[3][pix & 0xf];
@@ -119,22 +121,22 @@ uint8 ConvertTile (uint8 *pCache, uint32 TileAddr)
 	    uint32 p1 = 0;
 	    uint32 p2 = 0;
 	    register uint8 pix;
-	    if (pix = *(tp + 0))
+	    if ((pix = *(tp + 0)))
 	    {
 		p1 |= odd_high[0][pix >> 4];
 		p2 |= odd_low[0][pix & 0xf];
 	    }
-	    if (pix = *(tp + 1))
+	    if ((pix = *(tp + 1)))
 	    {
 		p1 |= even_high[0][pix >> 4];
 		p2 |= even_low[0][pix & 0xf];
 	    }
-	    if (pix = *(tp + 16))
+	    if ((pix = *(tp + 16)))
 	    {
 		p1 |= odd_high[1][pix >> 4];
 		p2 |= odd_low[1][pix & 0xf];
 	    }
-	    if (pix = *(tp + 17))
+	    if ((pix = *(tp + 17)))
 	    {
 		p1 |= even_high[1][pix >> 4];
 		p2 |= even_low[1][pix & 0xf];
@@ -151,12 +153,12 @@ uint8 ConvertTile (uint8 *pCache, uint32 TileAddr)
 	    uint32 p1 = 0;
 	    uint32 p2 = 0;
 	    register uint8 pix;
-	    if (pix = *(tp + 0))
+	    if ((pix = *(tp + 0)))
 	    {
 		p1 |= odd_high[0][pix >> 4];
 		p2 |= odd_low[0][pix & 0xf];
 	    }
-	    if (pix = *(tp + 1))
+	    if ((pix = *(tp + 1)))
 	    {
 		p1 |= even_high[0][pix >> 4];
 		p2 |= even_low[0][pix & 0xf];
@@ -646,7 +648,7 @@ inline void WRITE_4PIXELS16_ADD1_2 (uint32 Offset, uint8 *Pixels)
 	{ \
 	    if (SubDepth [N] != 1) \
 		Screen [N] = (uint16) (COLOR_ADD1_2 (GFX.ScreenColors [Pixel], \
-					   Screen [GFX.Delta + N])); \
+						     Screen [GFX.Delta + N])); \
 	    else \
 		Screen [N] = COLOR_ADD (GFX.ScreenColors [Pixel], \
 					GFX.FixedColour); \
@@ -678,7 +680,7 @@ inline void WRITE_4PIXELS16_FLIPPED_ADD1_2 (uint32 Offset, uint8 *Pixels)
 	{ \
 	    if (SubDepth [N] != 1) \
 		Screen [N] = (uint16) (COLOR_ADD1_2 (GFX.ScreenColors [Pixel], \
-					   Screen [GFX.Delta + N])); \
+						     Screen [GFX.Delta + N])); \
 	    else \
 		Screen [N] = COLOR_ADD (GFX.ScreenColors [Pixel], \
 					GFX.FixedColour); \
@@ -824,6 +826,7 @@ inline void WRITE_4PIXELS16_FLIPPED_SUB1_2 (uint32 Offset, uint8 *Pixels)
 #undef FN
 }
 
+
 void DrawTile16Add (uint32 Tile, uint32 Offset, uint32 StartLine,
 		    uint32 LineCount)
 {
@@ -904,6 +907,152 @@ void DrawClippedTile16Sub1_2 (uint32 Tile, uint32 Offset,
     RENDER_CLIPPED_TILE(WRITE_4PIXELS16_SUB1_2, WRITE_4PIXELS16_FLIPPED_SUB1_2, 4)
 }
 
+inline void WRITE_4PIXELS16_ADDF1_2 (uint32 Offset, uint8 *Pixels)
+{
+    uint32 Pixel;
+    uint16 *Screen = (uint16 *) GFX.S + Offset;
+    uint8  *Depth = GFX.ZBuffer + Offset;
+    uint8  *SubDepth = GFX.SubZBuffer + Offset;
+
+#define FN(N) \
+    if (GFX.Z1 > Depth [N] && (Pixel = Pixels[N])) \
+    { \
+	if (SubDepth [N] == 1) \
+	    Screen [N] = (uint16) (COLOR_ADD1_2 (GFX.ScreenColors [Pixel], \
+						 GFX.FixedColour)); \
+	else \
+	    Screen [N] = GFX.ScreenColors [Pixel];\
+	Depth [N] = GFX.Z2; \
+    }
+
+    FN(0)
+    FN(1)
+    FN(2)
+    FN(3)
+
+#undef FN
+}
+
+inline void WRITE_4PIXELS16_FLIPPED_ADDF1_2 (uint32 Offset, uint8 *Pixels)
+{
+    uint32 Pixel;
+    uint16 *Screen = (uint16 *) GFX.S + Offset;
+    uint8  *Depth = GFX.ZBuffer + Offset;
+    uint8  *SubDepth = GFX.SubZBuffer + Offset;
+
+#define FN(N) \
+    if (GFX.Z1 > Depth [N] && (Pixel = Pixels[3 - N])) \
+    { \
+	if (SubDepth [N] == 1) \
+	    Screen [N] = (uint16) (COLOR_ADD1_2 (GFX.ScreenColors [Pixel], \
+						 GFX.FixedColour)); \
+	else \
+	    Screen [N] = GFX.ScreenColors [Pixel];\
+	Depth [N] = GFX.Z2; \
+    }
+
+    FN(0)
+    FN(1)
+    FN(2)
+    FN(3)
+
+#undef FN
+}
+
+inline void WRITE_4PIXELS16_SUBF1_2 (uint32 Offset, uint8 *Pixels)
+{
+    uint32 Pixel;
+    uint16 *Screen = (uint16 *) GFX.S + Offset;
+    uint8  *Depth = GFX.ZBuffer + Offset;
+    uint8  *SubDepth = GFX.SubZBuffer + Offset;
+
+#define FN(N) \
+    if (GFX.Z1 > Depth [N] && (Pixel = Pixels[N])) \
+    { \
+	if (SubDepth [N] == 1) \
+	    Screen [N] = (uint16) COLOR_SUB1_2 (GFX.ScreenColors [Pixel], \
+						GFX.FixedColour); \
+	else \
+	    Screen [N] = GFX.ScreenColors [Pixel]; \
+	Depth [N] = GFX.Z2; \
+    }
+
+    FN(0)
+    FN(1)
+    FN(2)
+    FN(3)
+
+#undef FN
+}
+
+inline void WRITE_4PIXELS16_FLIPPED_SUBF1_2 (uint32 Offset, uint8 *Pixels)
+{
+    uint32 Pixel;
+    uint16 *Screen = (uint16 *) GFX.S + Offset;
+    uint8  *Depth = GFX.ZBuffer + Offset;
+    uint8  *SubDepth = GFX.SubZBuffer + Offset;
+
+#define FN(N) \
+    if (GFX.Z1 > Depth [N] && (Pixel = Pixels[3 - N])) \
+    { \
+	if (SubDepth [N] == 1) \
+	    Screen [N] = (uint16) COLOR_SUB1_2 (GFX.ScreenColors [Pixel], \
+						GFX.FixedColour); \
+	else \
+	    Screen [N] = GFX.ScreenColors [Pixel]; \
+	Depth [N] = GFX.Z2; \
+    }
+
+    FN(0)
+    FN(1)
+    FN(2)
+    FN(3)
+
+#undef FN
+}
+
+void DrawTile16FixedAdd1_2 (uint32 Tile, uint32 Offset, uint32 StartLine,
+			    uint32 LineCount)
+{
+    TILE_PREAMBLE
+    register uint8 *bp;
+
+    RENDER_TILE(WRITE_4PIXELS16_ADDF1_2, WRITE_4PIXELS16_FLIPPED_ADDF1_2, 4)
+}
+
+void DrawClippedTile16FixedAdd1_2 (uint32 Tile, uint32 Offset,
+				   uint32 StartPixel, uint32 Width,
+				   uint32 StartLine, uint32 LineCount)
+{
+    TILE_PREAMBLE
+    register uint8 *bp;
+
+    TILE_CLIP_PREAMBLE
+    RENDER_CLIPPED_TILE(WRITE_4PIXELS16_ADDF1_2, 
+			WRITE_4PIXELS16_FLIPPED_ADDF1_2, 4)
+}
+
+void DrawTile16FixedSub1_2 (uint32 Tile, uint32 Offset, uint32 StartLine,
+			    uint32 LineCount)
+{
+    TILE_PREAMBLE
+    register uint8 *bp;
+
+    RENDER_TILE(WRITE_4PIXELS16_SUBF1_2, WRITE_4PIXELS16_FLIPPED_SUBF1_2, 4)
+}
+
+void DrawClippedTile16FixedSub1_2 (uint32 Tile, uint32 Offset,
+				   uint32 StartPixel, uint32 Width,
+				   uint32 StartLine, uint32 LineCount)
+{
+    TILE_PREAMBLE
+    register uint8 *bp;
+
+    TILE_CLIP_PREAMBLE
+    RENDER_CLIPPED_TILE(WRITE_4PIXELS16_SUBF1_2, 
+			WRITE_4PIXELS16_FLIPPED_SUBF1_2, 4)
+}
+
 void DrawLargePixel16Add (uint32 Tile, uint32 Offset,
 			  uint32 StartPixel, uint32 Pixels,
 			  uint32 StartLine, uint32 LineCount)
@@ -1072,8 +1221,3 @@ void DrawClippedTile3dfx (uint32 Tile, uint32 Offset,
 }
 #endif
 #endif
-
-
-
-
-
