@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifndef _TILE_H_
 #define _TILE_H_
 
@@ -52,12 +54,11 @@
     pCache = &BG.Buffer[(TileNumber = (TileAddr >> BG.TileShift)) << 6]; \
 \
     if (!BG.Buffered [TileNumber]) \
-    { \
 	BG.Buffered[TileNumber] = ConvertTile (pCache, TileAddr); \
-    } \
 \
     if (BG.Buffered [TileNumber] == BLANK_TILE) \
 	return; \
+\
     register uint32 l; \
     if (BG.DirectColourMode) \
     { \
@@ -149,9 +150,9 @@
 	bp = pCache + StartLine; \
 	for (l = LineCount; l != 0; l--, bp += 8, Offset += GFX.PPL) \
 	{ \
-	    if (dd = (*(uint32 *) bp) & d1) \
+	    if ((dd = (*(uint32 *) bp) & d1)) \
 		NORMAL (Offset, (uint8 *) &dd); \
-	    if (dd = (*(uint32 *) (bp + 4)) & d2) \
+	    if ((dd = (*(uint32 *) (bp + 4)) & d2)) \
 		NORMAL (Offset + N, (uint8 *) &dd); \
 	} \
     } \
@@ -163,9 +164,9 @@
 	SWAP_DWORD (d2); \
 	for (l = LineCount; l != 0; l--, bp += 8, Offset += GFX.PPL) \
 	{ \
-	    if (dd = *(uint32 *) (bp + 4) & d1) \
+	    if ((dd = *(uint32 *) (bp + 4) & d1)) \
 		FLIPPED (Offset, (uint8 *) &dd); \
-	    if (dd = *(uint32 *) bp & d2) \
+	    if ((dd = *(uint32 *) bp & d2)) \
 		FLIPPED (Offset + N, (uint8 *) &dd); \
 	} \
     } \
@@ -177,9 +178,9 @@
 	SWAP_DWORD (d2); \
 	for (l = LineCount; l != 0; l--, bp -= 8, Offset += GFX.PPL) \
 	{ \
-	    if (dd = *(uint32 *) (bp + 4) & d1) \
+	    if ((dd = *(uint32 *) (bp + 4) & d1)) \
 		FLIPPED (Offset, (uint8 *) &dd); \
-	    if (dd = *(uint32 *) bp & d2) \
+	    if ((dd = *(uint32 *) bp & d2)) \
 		FLIPPED (Offset + N, (uint8 *) &dd); \
 	} \
     } \
@@ -188,9 +189,9 @@
 	bp = pCache + 56 - StartLine; \
 	for (l = LineCount; l != 0; l--, bp -= 8, Offset += GFX.PPL) \
 	{ \
-	    if (dd = (*(uint32 *) bp) & d1) \
+	    if ((dd = (*(uint32 *) bp) & d1)) \
 		NORMAL (Offset, (uint8 *) &dd); \
-	    if (dd = (*(uint32 *) (bp + 4)) & d2) \
+	    if ((dd = (*(uint32 *) (bp + 4)) & d2)) \
 		NORMAL (Offset + N, (uint8 *) &dd); \
 	} \
     }
@@ -198,7 +199,7 @@
 #define RENDER_TILE_LARGE(PIXEL, FUNCTION) \
     if (!(Tile & (V_FLIP | H_FLIP))) \
     { \
-	if (pixel = *(pCache + StartLine + StartPixel)) \
+	if ((pixel = *(pCache + StartLine + StartPixel))) \
 	{ \
 	    pixel = PIXEL; \
 	    for (l = LineCount; l != 0; l--, sp += GFX.PPL, Depth += GFX.PPL) \
@@ -216,7 +217,7 @@
     if (!(Tile & V_FLIP)) \
     { \
 	StartPixel = 7 - StartPixel; \
-	if (pixel = *(pCache + StartLine + StartPixel)) \
+	if ((pixel = *(pCache + StartLine + StartPixel))) \
 	{ \
 	    pixel = PIXEL; \
 	    for (l = LineCount; l != 0; l--, sp += GFX.PPL, Depth += GFX.PPL) \
@@ -234,7 +235,7 @@
     if (Tile & H_FLIP) \
     { \
 	StartPixel = 7 - StartPixel; \
-	if (pixel = *(pCache + 56 - StartLine + StartPixel)) \
+	if ((pixel = *(pCache + 56 - StartLine + StartPixel))) \
 	{ \
 	    pixel = PIXEL; \
 	    for (l = LineCount; l != 0; l--, sp += GFX.PPL, Depth += GFX.PPL) \
@@ -250,7 +251,7 @@
     } \
     else \
     { \
-	if (pixel = *(pCache + 56 - StartLine + StartPixel)) \
+	if ((pixel = *(pCache + 56 - StartLine + StartPixel))) \
 	{ \
 	    pixel = PIXEL; \
 	    for (l = LineCount; l != 0; l--, sp += GFX.PPL, Depth += GFX.PPL) \
