@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #ifdef __linux
 #include <string.h>
 #include <ctype.h>
@@ -76,6 +78,8 @@ static vga_modeinfo *info = NULL;
 static uint32 video_page_size = 64 * 1024;
 static uint32 selected_video_page = ~0;
 
+uint8 snes9x_clear_change_log = 0;
+
 extern uint32 joypads [5];
 
 #define ATTRCON_ADDR	0x3c0
@@ -190,7 +194,7 @@ static struct sigaction oldsig1handler;
 static struct sigaction sig2handler;
 static struct sigaction oldsig2handler;
 
-void Sig1HandlerFunction(int n)
+void Sig1HandlerFunction(int)
 {
     extern void StopTimer ();
     StopTimer ();
@@ -201,7 +205,7 @@ void Sig1HandlerFunction(int n)
     raise(SIGUSR1);
 }
 
-void Sig2HandlerFunction(int n)
+void Sig2HandlerFunction(int)
 {
     restore_modex = TRUE;
 
@@ -211,7 +215,7 @@ void Sig2HandlerFunction(int n)
     raise(SIGUSR2);
 }
 
-void S9xInitDisplay (int argc, char **argv)
+void S9xInitDisplay (int /*argc*/, char ** /*argv*/)
 {
     if (vga_init() < 0)
     {
@@ -581,7 +585,7 @@ if (KEY_WASPRESSED(k)) b &= ~v;
     memcpy (prev_keystate, keystate, sizeof (prev_keystate));
 }
 
-void S9xSetTitle (const char *title)
+void S9xSetTitle (const char * /*title*/)
 {
 }
 
@@ -723,8 +727,6 @@ const char *S9xSelectFilename (const char *def, const char *dir1,
     return (NULL);
 }
 
-#include <asm/io.h>
-
 void outReg(Register r)
 {
     switch (r.port)
@@ -793,7 +795,7 @@ void S9xParseDisplayArg (char **argv, int &ind, int)
 	 strcasecmp (argv [ind], "-mode") == 0) && argv [ind + 1])
     {
 	mode = atoi (argv [++ind]);
-	if (mode >= sizeof (modes) / sizeof (modes [0]))
+	if (mode >= (int) (sizeof (modes) / sizeof (modes [0])))
 	    mode = 0;
     }
     else
@@ -823,17 +825,19 @@ void S9xExtraUsage ()
 -scale      Scale SNES screen to fit S-VGA screen\n");
 }
 
-bool8 S9xReadMousePosition (int which1, int &x, int &y, uint32 &buttons)
+bool8 S9xReadMousePosition (int /* which1 */, int &/* x */, int & /* y */,
+			    uint32 & /* buttons */)
 {
     return (FALSE);
 }
 
-bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
+bool8 S9xReadSuperScopePosition (int & /* x */, int & /* y */, 
+				 uint32 & /* buttons */)
 {
     return (FALSE);
 }
 
-void S9xMessage (int type, int number, const char *message)
+void S9xMessage (int /* type */, int /* number */, const char *message)
 {
     fprintf (stderr, "%s\n", message);
 }
