@@ -1,19 +1,22 @@
 /*
  * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  *
- * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
- *                                      Jerremy Koot (jkoot@snes9x.com)
+ * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
+ *                           Jerremy Koot (jkoot@snes9x.com)
  *
  * Super FX C emulator code 
- * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
- *                                Gary Henderson.
+ * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
+ *                           Gary Henderson.
  * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
  *
  * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
+ * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
+ * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
+ *
  * DOS port code contains the works of other authors. See headers in
  * individual files.
  *
- * Snes9x homepage: www.snes9x.com
+ * Snes9x homepage: http://www.snes9x.com
  *
  * Permission to use, copy, modify and distribute Snes9x in both binary and
  * source form, for non-commercial purposes, is hereby granted without fee,
@@ -35,7 +38,6 @@
  * Super NES and Super Nintendo Entertainment System are trademarks of
  * Nintendo Co., Limited and its subsidiary companies.
  */
-
 #include <signal.h>
 #include <errno.h>
 
@@ -47,7 +49,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <ctype.h>
-
+#include <dirent.h>
 
 #if defined(__linux) || defined(__sun)
 #undef USE_THREADS
@@ -93,8 +95,15 @@ typedef void (*SIG_PF)();
 #include "display.h"
 #include "gfx.h"
 #include "soundux.h"
+#include "spc700.h"
 
-#ifdef NETPLAY_SUPPORT
+#ifdef SPCTOOL
+#include "spctool/spc700.h"
+extern "C" void TraceSPC (unsigned char *PC, unsigned short YA, unsigned char X,
+			  SPCFlags PS, unsigned char *SP);
+#endif
+
+#ifdef _NETPLAY_SUPPORT
 #include "netplay.h"
 #endif
 
@@ -295,6 +304,8 @@ void S9xParseArg (char **argv, int &i, int argc)
 	S9xParseDisplayArg (argv, i, argc);
 }
 
+#include "cheats.h"
+
 int main (int argc, char **argv)
 {
     if (argc < S9xMinCommandLineArgs ())
@@ -311,12 +322,12 @@ int main (int argc, char **argv)
     Settings.SoundPlaybackRate = 4;
     Settings.Stereo = TRUE;
     Settings.SoundBufferSize = 0;
-    Settings.SPCTo65c816Ratio = 2;
+    Settings.CyclesPercentage = 100;
     Settings.DisableSoundEcho = FALSE;
-    Settings.APUEnabled = TRUE;
-    Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
+    Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
+    Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
     Settings.SkipFrames = AUTO_FRAMERATE;
-    Settings.Shutdown = TRUE;
+    Settings.ShutdownMaster = TRUE;
     Settings.FrameTimePAL = 20000;
     Settings.FrameTimeNTSC = 16667;
     Settings.FrameTime = Settings.FrameTimeNTSC;
@@ -332,10 +343,13 @@ int main (int argc, char **argv)
     Settings.NetPlay = FALSE;
     Settings.ServerName [0] = 0;
     Settings.ThreadSound = FALSE;
-#ifdef NETPLAY_SUPPORT
-    Settings.Port = NETPLAY_DEFAULT_PORT;
+    Settings.AutoSaveDelay = 30;
+#ifdef _NETPLAY_SUPPORT
+    Settings.Port = NP_DEFAULT_PORT;
 #endif
-
+    Settings.ApplyCheats = TRUE;
+    Settings.TurboMode = FALSE;
+    Settings.TurboSkipFrames = 15;
     rom_filename = S9xParseArgs (argv, argc);
 
     Settings.Transparency = Settings.ForceTransparency;
@@ -352,6 +366,7 @@ int main (int argc, char **argv)
 
     (void) S9xInitSound (Settings.SoundPlaybackRate, Settings.Stereo,
 			 Settings.SoundBufferSize);
+
     if (!Settings.APUEnabled)
 	S9xSetSoundMute (TRUE);
 
@@ -365,11 +380,11 @@ int main (int argc, char **argv)
     {
 	if (!Memory.LoadROM (rom_filename))
 	{
-	    char dir [_MAX_DIR];
-	    char drive [_MAX_DRIVE];
-	    char name [_MAX_FNAME];
-	    char ext [_MAX_EXT];
-	    char fname [_MAX_PATH];
+	    char dir [_MAX_DIR + 1];
+	    char drive [_MAX_DRIVE + 1];
+	    char name [_MAX_FNAME + 1];
+	    char ext [_MAX_EXT + 1];
+	    char fname [_MAX_PATH + 1];
 
 	    _splitpath (rom_filename, drive, dir, name, ext);
 	    _makepath (fname, drive, dir, name, ext);
@@ -390,7 +405,8 @@ int main (int argc, char **argv)
 		exit (1);
 	    }
 	}
-	Memory.LoadSRAM (S9xGetSRAMFilename ());
+	Memory.LoadSRAM (S9xGetFilename (".srm"));
+	S9xLoadCheatFile (S9xGetFilename (".cht"));
     }
     else
     {
@@ -429,7 +445,7 @@ int main (int argc, char **argv)
 	
     S9xTextMode ();
 
-#ifdef NETPLAY_SUPPORT
+#ifdef _NETPLAY_SUPPORT
     if (strlen (Settings.ServerName) == 0)
     {
 	char *server = getenv ("S9XSERVER");
@@ -491,26 +507,29 @@ Connected to \"%s\" on port %d as player #%d playing \"%s\"\n",
 	FILE *fs = fopen ("test.bin", "r");
 	if (fs)
 	{
-	    int bytes = fread (IAPU.RAM, 1, 1024 * 64, fs);
+	    memset (IAPU.RAM, 0, 1024 * 64);
+	    int bytes = fread (IAPU.RAM + 1024, 1, 13, fs);
+	    bytes = fread (IAPU.RAM + 1024, 1, 1024 * 63, fs);
 	    fclose (fs);
-	    int len = 0;
-	    char buffer [200];
-	    for (int i = 0; i < bytes; i += len)
-	    {
-		len = S9xAPUOPrint (buffer, i);
-		printf ("%s\n", buffer);
-	    }
-	}
-    }
+#ifdef SPCTOOL
+	    _FixSPC (1024, 0, 0, 0, 0, 0xff);
+#else
+	    IAPU.PC = IAPU.RAM + 1024;
 #endif
-
-#if 0
-    if (Settings.APUEnabled)
-    {
-	// Needed for Goal
-	for (int z = 0; z < 0x2000; z++)
-	{
-	    APU_EXECUTE1();
+	    APU.Flags ^= TRACE_FLAG;
+	    extern FILE *apu_trace;
+	    if (APU.Flags & TRACE_FLAG)
+	    {
+#ifdef SPCTOOL
+		printf ("ENABLED\n");
+		_SetSPCDbg (TraceSPC);                   //Install debug handler
+#endif
+		if (apu_trace == NULL)
+		    apu_trace = fopen ("aputrace.log", "wb");
+	    }
+	    CPU.Cycles = 1024 * 10;
+	    APU_EXECUTE ();
+	    exit (0);
 	}
     }
 #endif
@@ -561,15 +580,21 @@ Connected to \"%s\" on port %d as player #%d playing \"%s\"\n",
     return (0);
 }
 
+void S9xAutoSaveSRAM ()
+{
+    Memory.SaveSRAM (S9xGetFilename (".srm"));
+}
+
 void S9xExit ()
 {
     S9xSetSoundMute (TRUE);
     S9xDeinitDisplay ();
-    Memory.SaveSRAM (S9xGetSRAMFilename ());
+    Memory.SaveSRAM (S9xGetFilename (".srm"));
+    S9xSaveCheatFile (S9xGetFilename (".cht"));
     Memory.Deinit ();
     S9xDeinitAPU ();
 
-#ifdef NETPLAY_SUPPORT
+#ifdef _NETPLAY_SUPPORT
     if (Settings.NetPlay)
 	S9xNetPlayDisconnect ();
 #endif
@@ -733,19 +758,19 @@ const char *S9xGetSnapshotDirectory ()
     return (filename);
 }
 
-const char *S9xGetSRAMFilename ()
+const char *S9xGetFilename (const char *ex)
 {
-    static char filename [PATH_MAX];
-    char drive [_MAX_DRIVE];
-    char dir [_MAX_DIR];
-    char fname [_MAX_FNAME];
-    char ext [_MAX_EXT];
+    static char filename [PATH_MAX + 1];
+    char drive [_MAX_DRIVE + 1];
+    char dir [_MAX_DIR + 1];
+    char fname [_MAX_FNAME + 1];
+    char ext [_MAX_EXT + 1];
 
     _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
     strcpy (filename, S9xGetSnapshotDirectory ());
     strcat (filename, SLASH_STR);
     strcat (filename, fname);
-    strcat (filename, ".srm");
+    strcat (filename, ex);
 
     return (filename);
 }
@@ -761,8 +786,7 @@ const char *S9xGetROMDirectory ()
 	return (roms);
 }
 
-#if !defined(__GLIBC__) || (__GLIBC__ < 2)
-const char *basename (const char *f)
+const char *S9xBasename (const char *f)
 {
     const char *p;
     if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
@@ -770,15 +794,14 @@ const char *basename (const char *f)
 
     return (f);
 }
-#endif
 
 const char *S9xChooseFilename (bool8 read_only)
 {
-    char def [PATH_MAX];
-    char title [PATH_MAX];
-    char drive [_MAX_DRIVE];
-    char dir [_MAX_DIR];
-    char ext [_MAX_EXT];
+    char def [PATH_MAX + 1];
+    char title [PATH_MAX + 1];
+    char drive [_MAX_DRIVE + 1];
+    char dir [_MAX_DIR + 1];
+    char ext [_MAX_EXT + 1];
 
     _splitpath (Memory.ROMFilename, drive, dir, def, ext);
     strcat (def, ".s96");
@@ -794,10 +817,10 @@ const char *S9xChooseFilename (bool8 read_only)
 
 bool8 S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM *file)
 {
-    char filename [PATH_MAX];
-    char drive [_MAX_DRIVE];
-    char dir [_MAX_DIR];
-    char ext [_MAX_EXT];
+    char filename [PATH_MAX + 1];
+    char drive [_MAX_DRIVE + 1];
+    char dir [_MAX_DIR + 1];
+    char ext [_MAX_EXT + 1];
 
     _splitpath (fname, drive, dir, filename, ext);
 
@@ -957,12 +980,11 @@ void _splitpath (const char *path, char *drive, char *dir, char *fname,
 
 void S9xToggleSoundChannel (int c)
 {
-    static int channel_enable = 255;
     if (c == 8)
-	channel_enable = 255;
+	so.sound_switch = 255;
     else
-	channel_enable ^= 1 << c;
-    S9xSetSoundControl (channel_enable);
+	so.sound_switch ^= 1 << c;
+    S9xSetSoundControl (so.sound_switch);
 }
 
 static void SoundTrigger ()
@@ -1022,7 +1044,7 @@ void InitTimer ()
 
 void S9xSyncSpeed ()
 {
-#ifdef NETPLAY_SUPPORT
+#ifdef _NETPLAY_SUPPORT
     if (Settings.NetPlay)
     {
 	// XXX: Send joypad position update to server
@@ -1054,15 +1076,17 @@ void S9xSyncSpeed ()
     }
     else
 #endif
+#if 0
     if (Settings.SoundSync == 2)
     {
 	IPPU.RenderThisFrame = TRUE;
 	IPPU.SkippedFrames = 0;
     }
     else
-    if (Settings.SkipFrames == AUTO_FRAMERATE)
+#endif
+    if (!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE)
     {
-	static struct timeval next1 = {0};
+	static struct timeval next1 = {0, 0};
 	struct timeval now;
 
 #ifdef __sgi
@@ -1113,7 +1137,8 @@ void S9xSyncSpeed ()
     }
     else
     {
-	if (++IPPU.FrameSkip >= (int) Settings.SkipFrames)
+	if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
+						    : Settings.SkipFrames))
 	{
 	    IPPU.FrameSkip = 0;
 	    IPPU.SkippedFrames = 0;
@@ -1410,7 +1435,7 @@ static int BufferSizes [8] =
 
 bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
 {
-    int I, J, K;
+    int J, K;
 
     if ((so.sound_fd = open ("/dev/dsp", O_WRONLY)) < 0)
     {
@@ -1473,6 +1498,8 @@ bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
 	return (FALSE);
     }
 
+    S9xSetPlaybackRate (so.playback_rate);
+
     if (buffer_size == 0)
 	buffer_size = BufferSizes [mode & 7];
 
@@ -1615,14 +1642,16 @@ void *S9xProcessSound (void *)
 {
 #ifdef __linux
     audio_buf_info info;
-    count_info count;
 
     if (!Settings.ThreadSound &&
 	(ioctl (so.sound_fd, SNDCTL_DSP_GETOSPACE, &info) == -1 ||
 	 info.bytes < so.buffer_size))
+    {
 	return (NULL);
-
+    }
 #ifdef MMAP_SOUND
+    count_info count;
+
     if (ioctl (so.sound_fd, SNDCTL_DSP_GETOPTR, &count) < 0)
     {
 	printf ("F"); fflush (stdout);
@@ -1748,7 +1777,7 @@ void *S9xProcessSound (void *)
 
 uint32 S9xReadJoypad (int which1)
 {
-#ifdef NETPLAY_SUPPORT
+#ifdef _NETPLAY_SUPPORT
     if (Settings.NetPlay)
 	return (S9xNetPlayGetJoypad (which1));
 #endif
@@ -1847,7 +1876,7 @@ void S9xParseConfigFile ()
        Settings.SkipFrames = get_config_int("Graphics", "Frameskip", AUTO_FRAMERATE);
     else
        Settings.SkipFrames = AUTO_FRAMERATE;
-    Settings.Shutdown = get_config_int("Graphics", "Shutdown", TRUE);
+    Settings.ShutdownMaster = get_config_int("Graphics", "Shutdown", TRUE);
     Settings.FrameTimePAL = get_config_int("Graphics", "FrameTimePAL", 20000);
     Settings.FrameTimeNTSC = get_config_int("Graphics", "FrameTimeNTSC", 16667);
     Settings.FrameTime = Settings.FrameTimeNTSC;
@@ -1897,3 +1926,122 @@ void S9xParseConfigFile ()
     }
 }
 #endif
+
+static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
+{
+    return (*(uint32 *) p1 - *(uint32 *) p2);
+}
+
+void S9xLoadSDD1Data ()
+{
+    char filename [_MAX_PATH + 1];
+    char index [_MAX_PATH + 1];
+    char data [_MAX_PATH + 1];
+    char patch [_MAX_PATH + 1];
+
+    Memory.FreeSDD1Data ();
+
+    strcpy (filename, S9xGetSnapshotDirectory ());
+
+    if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0)
+	strcat (filename, "/socnsdd1");
+    else
+	strcat (filename, "/sfa2sdd1");
+
+    DIR *dir = opendir (filename);
+
+    index [0] = 0;
+    data [0] = 0;
+    patch [0] = 0;
+
+    if (dir)
+    {
+	struct dirent *d;
+	
+	while ((d = readdir (dir)))
+	{
+	    if (strcasecmp (d->d_name, "SDD1GFX.IDX") == 0)
+	    {
+		strcpy (index, filename);
+		strcat (index, "/");
+		strcat (index, d->d_name);
+	    }
+	    else
+	    if (strcasecmp (d->d_name, "SDD1GFX.DAT") == 0)
+	    {
+		strcpy (data, filename);
+		strcat (data, "/");
+		strcat (data, d->d_name);
+	    }
+	    if (strcasecmp (d->d_name, "SDD1GFX.PAT") == 0)
+	    {
+		strcpy (patch, filename);
+		strcat (patch, "/");
+		strcat (patch, d->d_name);
+	    }
+	}
+	closedir (dir);
+
+	if (strlen (index) && strlen (data))
+	{
+	    FILE *fs = fopen (index, "rb");
+	    int len = 0;
+
+	    if (fs)
+	    {
+		// Index is stored as a sequence of entries, each entry being
+		// 12 bytes consisting of:
+		// 4 byte key: (24bit address & 0xfffff * 16) | translated block
+		// 4 byte ROM offset
+		// 4 byte length
+		fseek (fs, 0, SEEK_END);
+		len = ftell (fs);
+		rewind (fs);
+		Memory.SDD1Index = (uint8 *) malloc (len);
+		fread (Memory.SDD1Index, 1, len, fs);
+		fclose (fs);
+		Memory.SDD1Entries = len / 12;
+
+		if (!(fs = fopen (data, "rb")))
+		{
+		    free ((char *) Memory.SDD1Index);
+		    Memory.SDD1Index = NULL;
+		    Memory.SDD1Entries = 0;
+		}
+		else
+		{
+		    fseek (fs, 0, SEEK_END);
+		    len = ftell (fs);
+		    rewind (fs);
+		    Memory.SDD1Data = (uint8 *) malloc (len);
+		    fread (Memory.SDD1Data, 1, len, fs);
+		    fclose (fs);
+
+		    if (strlen (patch) > 0 &&
+			(fs = fopen (patch, "rb")))
+		    {
+			fclose (fs);
+		    }
+#ifdef MSB_FIRST
+		    // Swap the byte order of the 32-bit value triplets on
+		    // MSBFirst machines.
+		    uint8 *ptr = Memory.SDD1Index;
+		    for (int i = 0; i < Memory.SDD1Entries; i++, ptr += 12)
+		    {
+			SWAP_DWORD ((*(uint32 *) (ptr + 0)));
+			SWAP_DWORD ((*(uint32 *) (ptr + 4)));
+			SWAP_DWORD ((*(uint32 *) (ptr + 8)));
+		    }
+#endif
+		    qsort (Memory.SDD1Index, Memory.SDD1Entries, 12,
+			   S9xCompareSDD1IndexEntries);
+		}
+	    }
+	}
+	else
+	{
+	    printf ("Decompressed data pack not found in '%s'.\n", filename);
+	}
+    }
+}
+
